/* --------------------------------------------------------------------------
 * File: clspst.c
 * Description: Column generation procedure for the CLSPST - Disaggregated
 * CAPACITATED MULTI ITEM LOT SIZING PROBLEM WITH SETUP TIMES (WITHOUT BACKLOGGING OR LINKED LOT-SIZES)
 * --------------------------------------------------------------------------
 * Authors: Pablo Gonzalez Brevis [P.Gonzalez-Brevis@sms.ed.ac.uk]
 *          Pedro Munari [pedro.munari@gmail.com]
 * Last update: Oct/2012
 * --------------------------------------------------------------------------
 *  ON TOP OF HOPDM/PDCGM
 *  Written by:    Jacek Gondzio
 *                 Department of Mathematics and Statistics,
 *                 The University of Edinburgh,
 *                 Mayfield Road, King's Buildings.
 *                 EH9 3JZ Edinburgh, Scotland.
 *  on leave from: Systems Research Institute,
 *                 Polish Academy of Sciences,
 *                 Newelska 6, 01-447 Warsaw, Poland.
 ----------------------------------------------------------------------------*/ 

/* Bringing in the standard C library files */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>

/* Bringing in pdcgm and interfaces */
#include "../../pdcgm/pdcgm.h"

/* Bring wagner-whitin algorithm */
#include "include/wagnerwhitin.h"

#define SAVE_TO_FILE

/* Tolerances and bounds */
#define CLSPST_DELTA 1.E-6

/* Data structure for the instance data */
typedef struct
{
  int n;  /* number of products */
  int t;  /* number of periods */

  double C;   /* capacity per period C */
  double *h;  /* holding cost h(j,t) (product j in in period t) */
  double *f;  /* setup cost f(j,t) (product j in in period t) */
  double *d;  /* demand d(j,t) (product j in period t) */
  double *a;  /* production coefficient a(j) (product j)*/
  double *b;  /* setup time -> loss of production capacity b(j,t) (prodcut j in period t)*/
  double *buffer_double;  /* auxiliar array of doubles to be used in the oracle */
 
} T_instance_data;

/*----------------------------------------------------
READ AN INSTANCE FROM THE FILE SPECIFIED AS PARAMETER 
------------------------------------------------------*/
int read_clspst(
    char file_name[],  /* name of the file that describes the instance */
    T_instance_data *instance) /* pointer to the instance data */
{
    /* 
        For the CLSPST instances obtained from:
        Trigeiro, W.W., L.J. Thomas, J.O. McClain (1989): Capacitated lotsizing with setup times, Management Science 35, 353-366
    */
  
    /* Statement of variables */
    FILE *myfile;
    char *ret, current_line[101];
    int i, j, aux, m;

    /* Open the file */
    myfile = fopen(file_name, "r");
    if(myfile == NULL) return 0;

    /* Read the values n and t (number of items and periods) */
    instance->n = 0;
    instance->t = 0;
    aux = fscanf(myfile, "%d %d", &(instance->n), &(instance->t));
    if( (instance->n <= 0) || (instance->t <= 0) ) return 0;

    /* Memory allocation */
    instance->h = (double *) malloc(sizeof(double) * (instance->n));
    if(instance->h == NULL) return -1;
    instance->d = (double *) malloc(sizeof(double) * (instance->n)*(instance->t));
    if(instance->d == NULL) return -1;
    instance->a = (double *) malloc(sizeof(double) * (instance->n));
    if(instance->a == NULL) return -1;
    instance->f = (double *) malloc(sizeof(double) * (instance->n));
    if(instance->f == NULL) return -1;
    instance->b = (double *) malloc(sizeof(double) * (instance->n));
    if(instance->b == NULL) return -1;
  
    /* Read number of machines */
    aux = fscanf(myfile, "%i", &m);

    /* Read the capacity of the machine */
    aux = fscanf(myfile, "%lf", &(instance->C));
    //instance->C *= 1.1;
  
    /* Read the production coefficient, holding cost, setup time and setup cost per product */
    for (i=0; i < instance->n; i++)
    {
        aux = fscanf(myfile, "%lf", &(instance->a[i]));
        aux = fscanf(myfile, "%lf", &(instance->h[i]));
        aux = fscanf(myfile, "%lf", &(instance->b[i]));
        aux = fscanf(myfile, "%lf", &(instance->f[i]));
    }
  
    /* Read the demand per period and product*/
    for (j=0; j < instance->t; j++)
    {
        for (i=0; i < instance->n; i++)
        {
            aux = fscanf(myfile, "%lf", &(instance->d[i*(instance->t) + j]));
        }
    }
    printf("File '%s' has been succesfully read\n", file_name);
    
    /* Close the file */
    fclose(myfile); 

    /* Return successfully */
    return 1;   
}

/* DEFINE INITIAL COLUMNS TO THE RMP */
int set_initial_columns(
    PDCGM  *PDCGM_env,         /* PDCGM environment */
    T_instance_data *instance) /* Pointer to the instance data */
{
  
    /* Statement of variables */
    int i, j, k, p, t, nnz, max_m, max_n, max_nz;
    PDCGM_SMatrix_CW *M;  /* Matrix with the generated column(s) */

    /* Set initial dimensions */
    M = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    if(M == NULL) return -1;
    max_m = instance->t + instance->n + 1;  /* number of rows in the RMP (number of periods + convexity constraints (number of items) + obj function) */
    max_n = 1;  /* number of columns in the RMP (one artificial column only) */
    max_nz = max_m;
    PDCGM_set_SMatrix_CW(M, max_m, max_n, max_nz);

    /* Set the matrix counters */
    M->m = max_m;
    M->n = max_n;
    M->nz= 0;  /* number of nonzeros */
  
    /* Initialize the data structure */ 
    M->clpnts[0] = M->nz+1; /* Fortran index */
  
    /* Big-M cost for the articial column */
    M->coeff[M->nz] = 1.0e+8; 
    M->rwnmbs[M->nz] = instance->n + instance->t + 1; /* Fortran index */
    (M->nz)++;
    
    /* Set coefficient 1.0 for all the convexity constraints */
    for(j=0; j < instance->n; j++)
    {
        /* Set the coefficient corresponding to the dual variable v_j (convexity constraints in the RMP)*/
        /* We use dissagregated cuts therefore we generate a column for each product*/
        M->coeff[M->nz] = 1.;
        M->rwnmbs[M->nz] = instance->t + j + 1; /*Fortran index*/
        (M->nz)++;
    }

    /* Finishes the structure */
    M->clpnts[1] = M->nz + 1; /* Fortran index */
    
    /* Call the PDCGM function that add the new columns to the current RMP */
    PDCGM_add_columns(PDCGM_env, M, NULL);
    
    /* Clean memory */
    PDCGM_free_SMatrix_CW(M);
    PDCGM_FREE(M);
    
    return 0;
}


/* ORACLE FUNCTION: Solves the pricing subproblem using the dual solution */
/* The oracle() procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short CLSPST_oracle(
    double *primal_violation,   /* total violation of the generated constraints, if any */
    double *dual_violation,     /* total violation of the generated columns, if any */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data */
{

    int i, j, k, t, l, nc, ret, max_m, max_n, max_nz;   /* auxiliary variables */
    
    double opt, cost,
           *u, *x, *s, *y,   /* solution vectors */
           *solution;    /* save the solution of a given subproblem */
  
    /* Coefficient matrix to store the generated columns */ 
    PDCGM_SMatrix_CW *M; 
    
    /* Get the pointer to the instance data */
    T_instance_data *instance = (T_instance_data*) instance_data;
    
    /* Get the current dual solution */
    u = PDCGM_get_dual_solution(PDCGM_env);
    
    /* Allocate the auxiliary matrix in memory */
    M = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    if(M == NULL) return -1;
    max_m = instance->t + instance->n + 1;   /* number of rows in the RMP (number of periods + convexity constraints (number of items) + obj function) */
    max_n = instance->n;  /* one column from each subproblem  */
    max_nz = max_n * max_m;
    PDCGM_set_SMatrix_CW(M, max_m, max_n, max_nz);
    
    /* Set initial values */
    M->m = max_m;
    M->n = M->nz = 0;
    solution = instance->buffer_double;
    *dual_violation = 0.0;

    /* Set aliases for readability purpose */
    x = solution;
    s = &(solution[instance->t]);
    y = &(solution[2*instance->t]);
      
    /* We add each subproblem solution as a column */
    k=0;
    for(j=0; j < instance->n; j++)
    {     
        /* Solve the subproblems (Wagner-Whitin Algorithm) */
        opt = 0.0;
        wagnerwhitin(j, u, solution, &opt, instance->n, instance->t, instance->f, instance->d, instance->a, instance->b, instance->h);
        *dual_violation += (opt - u[j+instance->t]);
        
        /* This if-statement allows us to include or not a given column. 
           Add the column if the corresponding subproblem provides a negative cost */
        if (opt - u[j+instance->t] < -TOL_ZERO)
        {
            /* Add the column */
            M->clpnts[k] = M->nz + 1; /* Fortran index */
            k++;
            cost = 0.0;
                    
            /* Set the cost corresponding to the generated column ---> objective function */
            for(t=0; t < instance->t; t++)
            {
                cost += instance->h[j] * s[t] + instance->f[j] * y[t];
            }

            /* Sum up all the terms in order to calculate the LOWER BOUND */  
            /* If the reduced cost is non-positive, add the column with its corresponding coefficients for the 
               capacity constraint (a_j*x_{jt} + b_j*y_{jt}). If the reduced cost is positive, add an infeasible 
               column with capacity coefficient equals to total capacity (C_t) */     
            for (t=0; t < instance->t; t++)
            {
                if( x[t] > TOL_ZERO )
                {
                    M->coeff[M->nz] = (instance->a[j] * x[t] + instance->b[j] * y[t]);
                    M->rwnmbs[M->nz] = t + 1 ;
                    (M->nz)++;
                }
            }
      
            /* Set the coefficient corresponding to the dual variable v_j (convexity constraints in the relaxed master problem)*/
            /* We use dissagregated cuts therefore we generate a column for each product*/
            M->coeff[M->nz] = 1.;
            M->rwnmbs[M->nz] = instance->t + j + 1; /*Fortran index*/
            (M->nz)++;
      
            /* Set the cost corresponding to the generated column ---> last row*/
            M->coeff[M->nz] = cost;
            M->rwnmbs[M->nz] = instance->n + instance->t + 1; /*Fortran index*/
            (M->nz)++;
        }          
    }
    M->clpnts[k] = M->nz+1; /* -------> Fortran index */  
    M->n = k;   /* number of columns in the RMP (number of items) */
    
    /* Add the generated columns to the coefficient matrix  */
    if(M->n > 0)
    {
        /* Call the PDCGM function that add the new columns to the current RMP */
        nc = PDCGM_add_columns(PDCGM_env, M, NULL);

        /* If the answer is negative, report the ERROR */
        if( nc < 0 ) return -1;     
    }
    
    /* Clean memory */  
    PDCGM_free_SMatrix_CW(M);
    PDCGM_FREE(M);
    
    /* Return 1: COLUMN GENERATION */
    return 1;
}


/* C MAIN FUNCTION */
int main(int arg_c, char *arg_v[])
{
    long i, j, k, ret,  /* auxiliaries */
         dim,  /* number of linking constraints in the RMP */
         nb_cvxty_constr,  /* number of convexity constraints in the RMP */
         max_ncols_oracle; /* maximum number of generated columns in this call */

    FILE *out_file;  /* output file pointer */
    char file_name[100];  /* inpute file name */
    
    double *b, *lo_box, *up_box;  /* RHS and bounds of dual variables */
    short  *rw_type;  /* type of each constraint in the RMP */
    int    max_outer = 1000;  /* maximum number of outer iterations */
    
    PDCGM  *CLSPST; /* PDCGM environment for the CLSPST */
    
    T_instance_data *instance; /* Stores the CSP instance data */
  
    /* Check if the file was specified */
    if(arg_c < 2) 
    {
        printf("\nFile name not specified.\n");        
        return 0;
    }  
    /* Copy the parameters */
    strcpy(file_name, arg_v[1]);

    /* Memory allocation of the instance data structure */
    instance = (T_instance_data *) malloc(sizeof(T_instance_data));
    
    /* Read data of original problem */    
    if(read_clspst(file_name, instance) < 1) 
    {      
        printf("\nInstance file not found.\n");
        return 0;
    }
  
    /* Set the number of linking constraints in the master problem */
    dim = instance->t;
    
    /* Set the number of convexity constraints in the master problem */
    nb_cvxty_constr = instance->n; 
    
    /* Set the maximum number of columns in a call to the oracle */
    max_ncols_oracle = instance->n;
  
    /* Allocate the auxiliary array */
    instance->buffer_double = (double *) malloc(sizeof(double) * 3 * instance->t);
      
    /* Allocate memory for b, rw_type and boxes */
    b = PDCGM_ALLOC(double, dim + nb_cvxty_constr + 1);
    rw_type = PDCGM_ALLOC(short, dim + nb_cvxty_constr + 1);
    lo_box = PDCGM_ALLOC(double, dim);
    up_box = PDCGM_ALLOC(double, dim);
    for (i=0; i < dim; i++) 
    {
        b[i]  = instance->C;
        rw_type[i] = LESS_EQUAL;
        lo_box[i] = -1.e+3;
        up_box[i] = 0.0;
    }
    for (i=dim; i<dim+nb_cvxty_constr; i++) 
    {
        b[i]  = 1.0;
        rw_type[i] = EQUAL;
    }
    b[dim+nb_cvxty_constr] = 0.0;
    rw_type[dim+nb_cvxty_constr] = OBJECTIVE;
  
    #ifdef SAVE_TO_FILE
    /* Save to an output file */
    {
        out_file = fopen("output-clspst-pdcgm.txt", "a");
        fprintf(out_file, "\n%s ", file_name);
        fclose(out_file);
    }
    #endif    

    /* POPULATE THE INTERNAL DATA STRUCTURE OF PDCGM */ 
    CLSPST = PDCGM_set_data( 
            dim,  /* number of kept constraints in the RMP */
            nb_cvxty_constr,  /* number of convexity constraints in the RMP */
            (dim + nb_cvxty_constr + 1),  /* maximum number of nonzeros in a column */
            max_ncols_oracle,   /* maximum number of columns generated by the oracle */
            max_outer,   /* maximum number of outer iterations */
            NULL,  /* initial guess of the dual solution (may be NULL) */
            b,  /* RHS of each constraint in the RMP */
            rw_type,  /* type of each constraint (row) in the RMP */
            lo_box,   /* lower bound vector of the DUAL variables in the RMP */
            up_box,   /* upper bound vector of the DUAL variables in the RMP */
            instance);   /* instance data */
                
    /* Check if the PDCGM environment was created successfully */
    if(CLSPST == NULL)   
    {
        printf("-- ERROR: PDCGM_set_data() returned NULL --\n");
        fflush(stdout);
        return 0;
    }
            
    /* Set the optimality tolerance for the column generation algorithm */
    PDCGM_set_delta(CLSPST, CLSPST_DELTA);   
     
    /* Set the degree of optimality (parameter D > 1) */
    PDCGM_set_degree_of_optimality(CLSPST, 2.0);
    
    /* Set epsilon_max */
    PDCGM_set_max_opt_tol(CLSPST, 0.5);
    
    /* Set the verbose mode (how much information is printed) */
    PDCGM_set_verbosity(CLSPST, 1);
    
    /* Enable column elimination on insertion */
    PDCGM_set_column_elimination(CLSPST, 1);
    
    /* Start solving the problem with tighter bounds for the dual variables */
    //PDCGM_set_dual_box(CLSPST, -100.0, 100.0);
    
    /* Preloaded columns */
    set_initial_columns(CLSPST, instance);

    /* CALL THE COLUMN GENERATION PROCEDURE */
    ret = PDCGM_solve_MP(CLSPST, CLSPST_oracle);

    #ifdef SAVE_TO_FILE
    /* Save to an output file */
    {
        out_file = fopen("output-clspst-pdcgm.txt", "a");
        //fprintf(out_file, "\t%d \t%d ", instance->n, instance->t);
        fprintf(out_file, "\t%lf \t%lf \t%1.6E \t%d \t%lf \t%lf \t%lf ", CLSPST->lowerBnd, CLSPST->upperBnd, 
            CLSPST->rel_gap, CLSPST->outer, CLSPST->cputime_RMP, CLSPST->cputime_oracle, CLSPST->cputime_CG);
        fclose(out_file);
    }
    #endif    
    
    free(instance->buffer_double);
    free(instance->h);
    free(instance->d);
    free(instance->a);
    free(instance->f);
    free(instance->b);
    PDCGM_free(CLSPST);
    free(instance);
    free(b);
    free(rw_type);
    free(lo_box);
    free(up_box);
    
    return 0; 
}
