/* --------------------------------------------------------------------------
 * File: wagnerwhitin.h
 * DYNAMIC PROGRAMMING CODE - WAGNER WHITIN ALGORITHM TO BE USED IN THE 
 * CAPACITATED [MULTI ITEM] LOT SIZING PROBLEM WITH SETUP TIMES 
 * (WITHOUT BACKLOGGING OR LINKED LOT-SIZES) (clspst.c)
 * --------------------------------------------------------------------------
 * Authors: Pablo Gonzalez Brevis [P.Gonzalez-Brevis@sms.ed.ac.uk];
 *          Pedro Munari [pedro.munari@gmail.com]
 * Last update: 10/Jan/2011
 * --------------------------------------------------------------------------
 */

/*ASSUMPTIONS & DESCRIPTION OF DYNAMIC LOT-SIZING MODEL (SINGLE ITEM LOT SIZING PROBLEM - UNCAPACITATED):
 * 1. Demand d{j}{t} is known at the beginning of period 1 for every product
 * 2. Demand for period t must be met on time from inventory or from period t production. The cost c(x) of producing x units during any period is given by c(0) = 0, 
and for x > 0, c(x) = K + cx, where K is fixed cost for setting up production during a period, and c is the variable cost per unit of product j.
 * 3. At the end of period t, the inventory level s{j}{t} is observed, and a holding cost h{j}{t} is incurred. We let s{j}{0} denote the inventory level before period 1 production occurs (initial inventory).
 * 4. The goal is to determine a production level x{j}{t} of each product j in period t that minimizes the total cost of meeting (on time) the demands for periods 1,2,...,T for products 1,2,...,N.

	Reference: 
	-Wagner, H.M. and Whitin, T.M. "Dynamic version of the economic lot size model". Management Science,Vol.50, No. 12. Ten Most 	
	Influential Titles of Management Science's First Fifty Years (Dec.2004). pp.1770-1774. INFORMS
	-Winston, W.L. "Operations Research: Algorithms and Applications", 3rd Edition
*/

/* Bringing in the standard C library files */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>

#define TOL_ZERO 1E-8
#define INFTY_BD 1E+10

/* It solves single item lot sizing problems without capacity constraint */
int wagnerwhitin(
	int j,			  /* index of the subproblem */
	double *dual,	  /* dual solution from the RMP */
	double *solution, /* solution vector (x_{j1}, ..., x_{jT}, s_{j1}, ..., s_{jT}, y_{j1}, ..., y{jT}), with j FIXED */
	double *obj,	  /* cost associated to the production plan for item j*/
	int N,
	int T,
	double *f,  	/* setup cost f(j,t) (product j in in period t) */
	double *d,		/* demand d(j,t) (product j in period t) */
	double *a,		/* production coefficient a(j) (product j)*/
	double *b,		/* setup time -> loss of production capacity b(j,t) (prodcut j in period t)*/
	double *h);		/* holding cost*/

