/*
 *  COLUMN GENERATION PROCEDURE FOR THE 
 *  CUTTING STOCK PROBLEM (CSP)
 *  [Aggregated, k-best columns]
 *
 *  Author: Pedro Munari (pedro.munari@gmail.com)
 *  University of Sao Paulo, Brazil
 *  Last modified: 23/Jul/2012
 *
 *
 *  ON TOP OF HOPDM/PDCGM
 *  Written by:    Jacek Gondzio
 *                 Department of Mathematics and Statistics,
 *                 The University of Edinburgh,
 *                 Mayfield Road, King's Buildings.
 *                 EH9 3JZ Edinburgh, Scotland.
 *                 on leave from: Systems Research Institute,
 *                 Polish Academy of Sciences,
 *                 Newelska 6, 01-447 Warsaw, Poland.
 */

/* Bring in the standard C library files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Bring in the PDCGM/HOPDM library files */
#include "../../pdcgm/pdcgm.h"

/* Macros */
#define TOL_ZERO 1E-8

/* Flag: Save to a file when active */
#define SAVE_TO_FILE

/* Optimality tolerance */
#define CSP_DELTA 1.E-6

/* Maximum number of columns that can be generated at a time */
int MAX_NUMBER_COLS_IN_A_CALL = 100;

/* Data structure for the instance data */
typedef struct{
    double *w;        /* width of each item j = 1, ..., m */
    int *d;       /* demand of each item j = 1, ..., m */
    double W;     /* stock pieces length */
    long   m;     /* number of items */
    long   n;     /* number os stock pieces */
}T_instance_data;

/* Bring in the knapsack solver */
#include "include/knapsackBBk.h"

/* Read a CSP instance from the file specified as parameter */
int read_csp(
    char   *file_name, /* name of the file that describes the instance */
    T_instance_data *instance) /* pointer to the instance data */
{

    /* Statement of variables */
    FILE *myfile;
    long k, aux;
    char *ret, firstline[100];
    
    /* Open the file */
    myfile = fopen(file_name, "r");
    if(myfile == NULL) return 0;

    /* Ignores the first line */
    ret = fgets(firstline, 100, myfile);
        
    /* Read the values W m */
    aux = fscanf(myfile, "%ld", &(instance->m));
    aux = fscanf(myfile, "%lf", &(instance->W));
    
    instance->n = 10000;
    
    /* Memory allocation */
    instance->w = PDCGM_ALLOC(double, (instance->m));
    if(instance->w == NULL) return -1;
    instance->d = PDCGM_ALLOC(int, (instance->m));
    if(instance->d == NULL) return -1;
    
    /* Read the length of the items */
    for(k=0; k<instance->m; k++) 
    {
        aux = fscanf(myfile, "%lf", &(instance->w[k]));
        aux = fscanf(myfile, "%d", &(instance->d[k]));
    }
    
    /* Close the file */
    fclose(myfile);
    
    /* Return successfully */
    return 1;
}

/* Set initial columns to the RMP using homogeneous cuts */
/* Homogeneous cuts: a_{ii} = W/w_i; a_{ij} = 0 */
int set_initial_columns(
    PDCGM  *PDCGM_env,         /* PDCGM environment */
    T_instance_data *instance) /* Pointer to the instance data */
{
    int i, j, max_m, max_n, max_nz;
    PDCGM_SMatrix_CW *M;  /* Matrix with the generated column(s) */

    /* Set initial dimensions */
    M = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    if(M == NULL) return -1;
    max_m = instance->m + 2; /* number of rows in the RMP (m + convexity + obj function) */
    max_n = instance->m; /* number of columns in the RMP (initial plus slack) */
    max_nz = max_n * max_m;
    PDCGM_set_SMatrix_CW(M, max_m, max_n, max_nz);
        
    /* Set the initial columns */
    M->m = max_m; 
    M->n = max_n; 
    M->nz= 0;
    for(i=0; i<instance->m; i++)
    {
        /* Set a new column */
        M->clpnts[i] = M->nz + 1; /* -------> Fortran index */

        /* Set the entries of the column */
        M->coeff[ M->nz ] = floor(instance->W / (instance->w[i] * 1.0));
        M->rwnmbs[ M->nz ] = i + 1; /* -------> Fortran index */
        (M->nz)++;

        /* Set the coefficient in the convexity constraints */
        M->coeff[ M->nz ] = 1.;
        M->rwnmbs[ M->nz ] = instance->m + 1; /* -------> Fortran index */
        (M->nz)++;

        /* Set the cost corresponding to the generated column */
        M->coeff[ M->nz ] = 1.;
        M->rwnmbs[ M->nz ] = instance->m + 2; /* -------> Fortran index */
        (M->nz)++;
    }
        
    /* End of columns! */
    M->clpnts[ instance->m ] = M->nz + 1; /* -------> Fortran index */
    
    /* Call the PDCGM function that add the new columns to the current RMP */
    PDCGM_add_columns(PDCGM_env, M, NULL);
    
    /* Clean memory */
    PDCGM_free_SMatrix_CW(M);
    PDCGM_FREE(M);
    
    return 0;
}

/* SOLVE THE SUBPROBLEM (KNAPSACK) */
int solve_subproblem(
    double *u,         /* dual solution from the RMP */
    int ***X,          /* return the k-best solutions of the knapsack problem */
    double **f_X,      /* return the objective value of the solutions in X */
    int *k_max,        /* number of solutions required from the knapsack solver */
    T_instance_data *instance) /* pointer to the instance data */
{

    int n_itens, /* number of items in the knapsack */
        Lcap,    /* knapsack maximum capacity */   
        i, j;    /* auxiliar variables */
    
    knapsackBBk_Tdata *knap_data; /* knapsack problem data */    

    /* Allocate memory */
    n_itens = instance->m;  
    knap_data = (knapsackBBk_Tdata *) malloc(sizeof(knapsackBBk_Tdata) * n_itens);                     
    if(knap_data == NULL) return -1;
   
    /* Populate the knapsack problem */
    Lcap = instance->W;
    for(i=0; i < n_itens; i++)
    {
        if(u[i] > TOL_ZERO) knap_data[i].utilidade = u[i];   /* utility of an item (cost) */
        else knap_data[i].utilidade = 0.0;
        knap_data[i].l = instance->w[i]; /* width of an item (coefficient in the constraint) */ 
        knap_data[i].demanda = instance->d[i];     /* demand of an item ### Can be replaced by: instance->d[i] */
    }
   
    branch_bound_k(knap_data, n_itens, Lcap, k_max, X, f_X);

    /*
        printf("\nKNAP:\n");
        for(i=0; i<instance->m; i++) printf("%lf ", u[i]);
        printf("\n");
        for(i=0; i<instance->m; i++) printf("%d ", instance->w[i]);
        printf("\n");
        for(i=0; i<instance->m; i++) printf("%d ", instance->d[i]);
        printf("\n");

        for(i=0;(i<*k_max) && ((*f_X)[i] > 1.0);i++)
        {
            printf("[%d] %lf\t", i, (*f_X)[i]);
            //for(j=0;j<n_itens;j++) printf("X[%d][%d] = %d\n", i, j, (*X)[i][j]);
        }      
        printf("\n");

        scanf("%d", &i);
    */

    /* Clean the memory */
    free(knap_data);
    
    return 1;
}
/* END solve_subproblem */

/* ORACLE FUNCTION: Solves the pricing subproblem using the dual solution */
/* The oracle() procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short CSP_oracle(
    double *primal_violation,   /* total violation of the generated constraints, if any */
    double *dual_violation,     /* total violation of the generated columns, if any */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data */
{

    int i, j, k, nc, ret, max_m, max_n, max_nz,
        **ksolutions,  /* k-best solutions of the knapsack problem */
        *solution;     /* auxiliar variable */
    
    double  *u,        /* dual solution vector */
            *f_ksolutions;  /* objective value of the solutions in 'ksolutions' */

    /* Coefficient matrix to store the generated columns */ 
    PDCGM_SMatrix_CW *M; 
    
    /* Get the pointer to the instance data */
    T_instance_data *instance = (T_instance_data*) instance_data;
    
    /* Get the current dual solution */
    u = PDCGM_get_dual_solution(PDCGM_env);
    
    /* Allocate the auxiliary matrix in memory */
    M = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    if(M == NULL) return -1;
    max_m = instance->m + 2; /* number of rows in the RMP (m + convexity + obj function) */
    max_n = MAX_NUMBER_COLS_IN_A_CALL;
    max_nz = max_n * max_m;
    PDCGM_set_SMatrix_CW(M, max_m, max_n, max_nz);
    
    /* Set initial values */
    M->m = max_m;
    M->n = M->nz = 0;
    nc = MAX_NUMBER_COLS_IN_A_CALL;
    *dual_violation = 0.0;
    ret = 0;
    
    /* Solve the subproblem (knapsack) */
    solve_subproblem(u, &ksolutions, &f_ksolutions, &nc, instance);
    *dual_violation = f_ksolutions[0];
    
    /* Check the reduced cost */
    k=0;
    if(*dual_violation - 1.0 > 1.0e-6)
    {
        /* Insert the columns generated by the knapsack solver */
        for(j=0; (j < nc) && (f_ksolutions[j] - 1.0 > 1.0e-6); j++)
        {
            /* Get the next solution to generated the column */
            solution = ksolutions[j];
 
            /* Put the column in the sparse representation */
            M->clpnts[j] = k + 1; /* -------> Fortran index */
            for(i=0; i < instance->m; i++)
            {
                /* Check if it is a nonzero entry */
                if(solution[i] > TOL_ZERO) 
                {
                    /* Set a new entry in the column */
                    M->coeff[k] = (double) solution[i];
                    M->rwnmbs[k] = i + 1; /* -------> Fortran index */
                    k++;
                }
            }

            /* Set the entry in the convexity constraint */
            M->coeff[k] = 1.0;
            M->rwnmbs[k] = instance->m + 1; /* -------> Fortran index */
            k++;
            
            /* Set the cost corresponding to the generated column */
            M->coeff[k] = 1.0;
            M->rwnmbs[k] = instance->m + 2; /* -------> Fortran index */
            k++;   
        }
        /* Reset the number of generated columns */
        nc = j;
    }
    else 
    {
        /* Set a column with only one entry (in the convexity constraint) */
        M->coeff[k] = 1.0;
        M->rwnmbs[k] = instance->m + 1; /* -------> Fortran index */
        k++;
        
        /* Set the number of generated columns */
        nc = 1;
    }
    /* End of columns! */
    M->clpnts[nc] = k + 1; /* -------> Fortran index */

    /* Set the number of columns and nonzeros */
    M->n = nc; 
    M->nz= k;  

    /* Add the generated columns to the coefficient matrix  */
    if(M->n > 0)
    {
        /* Call the PDCGM function that add the new columns to the current RMP */
        nc = PDCGM_add_columns(PDCGM_env, M, NULL);

        /* If the answer is negative, report the ERROR */
        if( nc < 0 ) return -1;     
    }
            
    /* Update the objective value of the oracle */
    if(*dual_violation - 1.0 > 1.0e-6) *dual_violation = instance->n * (1 - *dual_violation); /* == N * (1 - x*u) */
    else *dual_violation = 0.0;

    /* Clean memory */  
    PDCGM_free_SMatrix_CW(M);
    PDCGM_FREE(M);
    if(ksolutions != NULL) 
    {
        for(i = 0; i < MAX_NUMBER_COLS_IN_A_CALL; i++) free(ksolutions[i]);
        free(ksolutions);
    }
    if(f_ksolutions != NULL) free(f_ksolutions);    
        
    /* Return 1: Column generation */
    return 1;
}

/* C MAIN FUNCTION */
int main(int arg_c, char *arg_v[])
{
    long i, j, k, ret,   /* auxiliaries */
         dim,   /* number of linking constraints in the RMP */
         nb_cvxty_constr;    /* number of convexity constraints in the RMP */

    FILE *out_file;  /* output file pointer */
    char file_name[100];  /* inpute file name */
    
    double *b, *lo_box, *up_box;  /* RHS and bounds of dual variables */
    short  *rw_type;  /* type of each constraint in the RMP */
    int    max_outer = 5000;  /* maximum number of outer iterations */
    
    PDCGM  *CSP;  /* PDCGM environment for the CSP */
    
    T_instance_data *instance; /* Stores the CSP instance data */

    /* Check if the file was specified */
    if (arg_c > 1) 
    {
        /* Copy the parameters */
        strcpy(file_name, arg_v[1]);
        if (arg_c > 2 && atoi(arg_v[2]) > 0) MAX_NUMBER_COLS_IN_A_CALL = atoi(arg_v[2]);
    }
    else
    {
        printf("\nFile name not specified.\n");        
        return 0;
    }
    
    /* Memory allocation of the instance data structure */
    instance = (T_instance_data *) malloc(sizeof(T_instance_data));
    
    /* Read the problem */
    if(read_csp(file_name, instance) < 1)
    {
        printf("\nThe file '%s' could not be read.\n", file_name);
        return 0;
    }
 
    /* Set the dimensions of the instance */
    dim = instance->m;
    nb_cvxty_constr = 1; 

    /* Allocate memory for b, rw_type and boxes */
    b = PDCGM_ALLOC(double, dim + nb_cvxty_constr + 1);
    rw_type = PDCGM_ALLOC(short, dim + nb_cvxty_constr + 1);
    lo_box = PDCGM_ALLOC(double, dim);
    up_box = PDCGM_ALLOC(double, dim);
    /* Set initial values */
    for (k=0; k<dim; k++) 
    {
        b[k] = instance->d[k];
        rw_type[k] = GREATER_EQUAL;
        lo_box[k] = 0.0;
        up_box[k] = 1.e+3;
    }
    for (k=dim; k<dim+nb_cvxty_constr; k++) 
    {
        b[k] = instance->n;
        rw_type[k] = LESS_EQUAL;
    }
    b[dim+nb_cvxty_constr] = 0.0;
    rw_type[dim+nb_cvxty_constr] = OBJECTIVE;

    /* POPULATE THE INTERNAL DATA STRUCTURE OF PDCGM */ 
    CSP = PDCGM_set_data( 
            dim,  /* number of kept constraints in the RMP */
            nb_cvxty_constr,  /* number of convexity constraints in the RMP */
            (dim + nb_cvxty_constr + 1),  /* maximum number of nonzeros in a column */
            MAX_NUMBER_COLS_IN_A_CALL,    /* maximum number of columns generated by the oracle */
            max_outer,     /* maximum number of outer iterations */
            NULL, /* initial guess of the dual solution (may be NULL) */
            b,    /* RHS of each constraint in the RMP */
            rw_type,  /* type of each constraint (row) in the RMP */
            lo_box,   /* lower bound vector of the DUAL variables in the RMP */
            up_box,   /* upper bound vector of the DUAL variables in the RMP */
            instance);   /* instance data */
                
    /* Check if the PDCGM environment was created successfully */
    if(CSP == NULL)   
    {
        printf("-- ERROR: PDCGM_set_data() returned NULL --\n");
        fflush(stdout);
        return 0;
    }
            
    /* Set the optimality tolerance for the column generation algorithm */
    PDCGM_set_delta(CSP, CSP_DELTA);   
     
    /* Set the degree of optimality (parameter D > 1) */
    PDCGM_set_degree_of_optimality(CSP, 5.0);
    
    /* Set epsilon_max */
    PDCGM_set_max_opt_tol(CSP, 0.5);
    
    /* Set the verbose mode (how much information is printed) */
    PDCGM_set_verbosity(CSP, 1);
    
    /* Enable column elimination on insertion */
    PDCGM_set_column_elimination(CSP, 1);
    
    /* Start solving the problem with tighter bounds for the dual variables */
    //PDCGM_set_dual_box(CSP, -100.0, 100.0);
    
    /* Preloaded columns */
    set_initial_columns(CSP, instance);

    /* CALL THE COLUMN GENERATION PROCEDURE */
    ret = PDCGM_solve_MP(CSP, CSP_oracle);
    
    #ifdef SAVE_TO_FILE
        /* Save to an output file */
        out_file = fopen("output-csp-pdcgm.txt", "a");
        fprintf(out_file, "\n%s ", file_name);
        //fprintf(out_file, "\t%ld \t%lf ", instance->m, instance->W);
        fprintf(out_file, "\t%lf \t%lf \t%1.6E \t%d \t%lf \t%lf \t%lf ", CSP->lowerBnd, CSP->upperBnd, 
                CSP->rel_gap, CSP->outer, CSP->cputime_RMP, CSP->cputime_oracle, CSP->cputime_CG);
        close(out_file); 
    #endif
    
    /* Clean the memory before returning */
    PDCGM_free(CSP);
    free(CSP);
    free(b);
    free(rw_type);
    free(lo_box);
    free(up_box);
    free(instance);

    return 0; 
}

