/*
 *  HOPDM  (Higher Order Primal-Dual Method)
 *  Copyright:     Jacek Gondzio, 1990, 2010.
 *
 *  PDCGM  (Primal-Dual Column Generation Method)
 *  Copyright:     Jacek Gondzio, 1996, 2010.
 *
 *  Using the PDCGM to solve a simple convex problem.
 *
 *  Written by:  - Jacek Gondzio
 *                 Department of Mathematics and Statistics,
 *                 The University of Edinburgh,
 *                 Mayfield Road, King's Buildings.
 *                 EH9 3JZ Edinburgh, Scotland.
 *                 on leave from: Systems Research Institute,
 *                 Polish Academy of Sciences,
 *                 Newelska 6, 01-447 Warsaw, Poland.
 *
 *               - Pedro Munari
 *                 University of Sao Paulo, Brazil
 *                 pedro.munari@gmail.com
 *
 *  Last modified: July, 2012.
 */

/* 
	File: qpexample3.
	It is a complete example of how to use PDCGM as a cutting plane algorithm
	in order to minimize a convex function. This implementation is similar to 
	qpexample2, but the columns are sent as sparse matrix to the solver. We 
	recommend reading the description given in file qpexample1.c before going 
	through the following code. 
*/

/* Bring in standard C library headers */
#include <stdlib.h>
#include <stdio.h>

/* Bring in the PDCGM/HOPDM library headers */
#include "../../pdcgm/pdcgm.h"

/* ORACLE FUNCTION: Evaluate f(y) = min{ y*y + 3*y + 1 } for a given point y */
/* The oracle() procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short oracle(
    double *primal_violation, 	/* total violation of the generated constraints (not used in this example) */
    double *dual_violation, 	/* total violation of the generated columns (the value f(y)) */
  	PDCGM  *PDCGM_env,			/* PDCGM environment */
    void   *instance_data)   	/* instance data (not used in this example) */
{

    /* Auxiliary variables for computing the cost coefficient */
    double f_value, f_derivative, cost;

    /* Dual solution vector y */
    double *y;
    
    /* Auxiliary matrix to store the generated columns (sparse representation) */	
	PDCGM_SMatrix_CW M; 	

    /* Print a message to the user */
    printf("[PDCGM] Hello from the Oracle function\n"); 

    /* Get a pointer to the dual solution of the current master problem */
    y = PDCGM_get_dual_solution(PDCGM_env);
    
    /* Compute f(y) */
    f_value = y[0] * y[0] + 3.0 * y[0] + 1.0;
    *dual_violation = -f_value - y[1];
    
    /* Compute the cost of the column: c = f(y) - s(y) * y (see comments above) */ 
    f_derivative = 2.0 * y[0] + 3.0; 
    cost = f_value - f_derivative * y[0];
    printf("--> y = (%.6e, %.6e), f(y)= %.6e, f'(y)= %.6e, cost = %.6e\n", y[0], y[1], f_value, f_derivative, cost);
    
    /* Set the new column: [s(y), 1, -cost]^t */
    PDCGM_set_SMatrix_CW(&M, 3, 1, 3);
    M.m = 3;
    M.n = 1;
    M.nz = 3;
    M.coeff[0] = f_derivative;
    M.coeff[1] = 1.0;
    M.coeff[2] = -cost;
    M.rwnmbs[0] = 1;
    M.rwnmbs[1] = 2;
    M.rwnmbs[2] = 3;
    M.clpnts[0] = 1;
    M.clpnts[1] = 4; 

    /* Add the new columns to the current master problem */
    PDCGM_add_columns(PDCGM_env, &M, NULL);

    /* Clean memory used by the auxiliary matrix */		
    PDCGM_free_SMatrix_CW(&M);

    /* Print the matrix in a file or to the standard output */
    /*
        FILE *f= fopen("M.txt", "w");
        if (f != NULL) print_Matrix(M, 1, f);
        else printf("failed in opening FILE f.\n"); 
        fclose(f);
    */

    /* Return 1: a column was generated */ 
    return 1;
}

/* C MAIN FUNCTION */
int main(void)
{
    int    k, ret; /* auxiliary variables */
    
    long   dim = 1, /* number of linking constraints in the master problem */
           nb_cvxty_constr = 1, /* number of convexity constraints in the master problem */
           max_n_cols_oracle = 1, /* maximum number of columns that are generated in each call to the oracle */
           max_outer = 1000; /* maximum number of outer iterations */
    
    double delta = 1.e-6, /* optimality tolerance */
           *b, *y0, *lo_box, *up_box, /* RHS, initial guess, and bounds */
           *y_opt; /* optimal solution */
    short  *row_type; /* type of each constraint in the master problem */
    
    PDCGM  *PDCGM_env; /* data structure of PDCGM */
	PDCGM_SMatrix_CW M; /* Auxiliary matrix to store the initial columns (sparse representation) */		

    /* b, initial guess y0, and boxes */ 
    b = PDCGM_ALLOC(double, dim + nb_cvxty_constr + 1);
    y0 = PDCGM_ALLOC(double, dim + nb_cvxty_constr + 1);
    row_type = PDCGM_ALLOC(short, dim + nb_cvxty_constr + 1);
    lo_box = PDCGM_ALLOC(double, dim);
    up_box = PDCGM_ALLOC(double, dim);
    
    /* Set the data of the linking constraint and the corresponding dual variable */
    for(k=0; k < dim; k++) 
    {
        b[k] = 0.0;
        y0[k] = 1.0; /* y0 = 1.0 is the initial guess */
        row_type[k] = EQUAL;
        lo_box[k] = -10.0;
        up_box[k] =  10.0;
    }
    /* Set the data of the convexity constraint and the corresponding dual variable */
    for (k=dim; k < dim + nb_cvxty_constr; k++) 
    {
        b[k] = 1.0;
        y0[k] = 0.0;
        row_type[k] = EQUAL;
    }
    /* Set the objective function costs */
    b[dim + nb_cvxty_constr] = 0.0;
    row_type[dim + nb_cvxty_constr] = OBJECTIVE;

    /* POPULATE THE INTERNAL DATA STRUCTURE OF PDCGM */	
	PDCGM_env = PDCGM_set_data( 
        dim,  /* number of kept constraints in the RMP */
        nb_cvxty_constr,  /* number of convexity constraints in the RMP */
        (dim + nb_cvxty_constr + 1),  /* maximum number of nonzeros in a column */
        max_n_cols_oracle,    /* maximum number of columns generated by the oracle */
        max_outer,     /* maximum number of outer iterations */
        y0, /* initial guess of the dual solution (may be NULL) */
        b,  /* RHS of each constraint in the RMP */
        row_type, /* type of each constraint (row) in the RMP */
        lo_box,   /* lower bound vector of the DUAL variables in the RMP */
        up_box,   /* upper bound vector of the DUAL variables in the RMP */
        NULL);   /* instance data */
	 			
	/* Check if the PDCGM environment was created successfully */
	if(PDCGM_env == NULL)   
	{
	    printf("-- ERROR: PDCGM_set_data() returned NULL --\n");
	    fflush(stdout);
	    return 0;
	}
			
	/* Set the optimality tolerance for the column generation algorithm */
	PDCGM_set_delta(PDCGM_env, delta);   
	 
	/* Set the degree of optimality (parameter D > 1) */
    PDCGM_set_degree_of_optimality(PDCGM_env, 5.0);
    
    /* Set the verbose mode (how much information is printed) */
    PDCGM_set_verbosity(PDCGM_env, 1);
    
    /* Enable column elimination on insertion */
    PDCGM_set_column_elimination(PDCGM_env, 1);

    /* Suggest initial columns: a cut using y = -1 */  
    PDCGM_set_SMatrix_CW(&M, 3, 1, 3);
    M.m = 3;
    M.n = 1;
    M.nz = 3;
    M.coeff[0] = 3.0;
    M.coeff[1] = 1.0;
    M.coeff[2] = -1.0;
    M.rwnmbs[0] = 1;
    M.rwnmbs[1] = 2;
    M.rwnmbs[2] = 3;
    M.clpnts[0] = 1;
    M.clpnts[1] = 4; 
    
    /* Add the new columns to the master problem */
    PDCGM_add_columns(PDCGM_env, &M, NULL);

	/* CALL THE COLUMN GENERATION PROCEDURE */
	ret = PDCGM_solve_MP(PDCGM_env, oracle);
	
    /* Printf the optimal solution */
    printf("\nPDCGM_solve_MP() ends with code = %d\n\n", ret);
    y_opt = PDCGM_get_dual_solution(PDCGM_env);
    for (k=0; k < dim; k++) printf("y_opt[%d] = %.6f\n", k, y_opt[k]);
    printf("Optimal value: %.6f\n\n", -PDCGM_env->upperBnd);

    /* Clean the memory before returning */
	PDCGM_free_SMatrix_CW(&M);
	PDCGM_free(PDCGM_env);
	free(PDCGM_env);
	free(b);
	free(row_type);
	free(lo_box);
	free(up_box);
	
	return 0;
}
