/*
 *  PDCGM  (Primal-Dual Column Generation Method)
 *  Copyright:     Jacek Gondzio, 1996, 2013.
 *
 *  COLUMN GENERATION FOR MULTICOMMODITY NETWORK FLOW PROBLEMS
 *  USING TWO DIFFERRENT FORMULATIONS: AGGREGATED AND DISAGGREGATED
 * 
 *  Authors: Pedro Munari [munari@dep.ufscar.br], 
 *           Pablo González-Brevis, 
 *           Jacek Gondzio
 *  
 *  MAIN FILE
 *  Last Modified: July, 2013.
 * 
 *  On top of HOPDM  (Higher Order Primal-Dual Method)
 *  Copyright:     Jacek Gondzio, 1990, 2010.
 *
 *  Using the Dijkstra algorithm implemented by:
 *      Jacek Gondzio and Robert Sarkissian
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <assert.h>
#include <math.h>
#include <omp.h>  

#define MCNF_TOL_ZERO 1.E-6
static int USE_ACTIVE_SET = 1;

/* Bring in the PDCGM/HOPDM headers */
#include "../../pdcgm/pdcgm.h"

/* Bring in the auxiliary functions */
#include "include/oracle.c"

/* ORACLE FUNCTION FOR THE LINEAR MCNF WITH *AGGREGATED* MASTER PROBLEM */
/* The oracle() procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short MCNF_linear_agg_oracle(
    double *primal_violation,   /* total violation of the generated constraints (not used in this application) */
    double *dual_violation,     /* total violation of the generated columns (the value f(y)) */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data (not used in this example) */
{
    
    int i, j, nr, nc, answer_type,
        convexity_row = Network->m, /* index of the convexity constraint in the coefficient matrix */
        objective_row = Network->m + 1; /* index of the objective function in the coefficient matrix */
        
    double cost, *d,
           *u, /* a pointer to the dual solution of the current master problem */
           *column; /* auxiliary array used to store the generated column */
    
    printf ("Hello from MCNF_linear_agg_oracle() -- Linear multicommodity network flow with aggregated MP.\n");
    fflush(stdout);
    
    /* Set the answer type: column generation */
    answer_type = 1; 
    
    /* Reset the number of columns added in the last interation */
    PDCGM_env->ncols_last = 0;
    
    /* Reset the pricing subproblem solver */
    Reset_Dijkstra (Network);
    
    /* Set the max number of entries in the generated column */
    nr = Network->m + 2; /* number of rows = number of arcs + convexity row +  objective row */
    
    /* Reset the dual_violation (the value of the oracle) */
    *dual_violation = 0.0;

    /* Get a pointer to the dual solution of the current master problem */
    u = PDCGM_get_dual_solution(PDCGM_env);
    
    if(USE_ACTIVE_SET)
    {
        /* Update the active set */
        if(PDCGM_env->outer > 0)
        {
            PDCGM_SMatrix_CW M;
            double *total_flow;
            double *lambda;
            int lambda_n;
            
            PDCGM_get_pointer_to_master_columns (PDCGM_env, &M);
            PDCGM_get_pointer_to_master_solution(PDCGM_env, &lambda, &lambda_n);
            total_flow = PDCGM_ALLOC(double, Network->m);
            
            for(i=0; i<M.n; i++)
            {
                for (j = M.clpnts[i] - 1; j < M.clpnts[i+1] - 1; j++)
                {
                    if(M.rwnmbs[j] - 1 < Network->m)
                        total_flow[ M.rwnmbs[j] - 1 ] -= M.coeff[j] * lambda[i];
                }
            }
            
            j=0;
            for(i=0; i<Network->m && j < 0.1 * Network->m; i++) 
            {
                if( (active_set[i] == 0) && (total_flow[i] > Network->K[i]) )
                {
                    PDCGM_env->rwstat[i] = GREATER_EQUAL;
                    active_set[i] = 1;
                    active_set_n++;
                    
                    u[i] = 0.0;
                    answer_type = 0;
                    j++;
                }
                else if( (active_set[i] == 1) && (u[i] < 0.0001) && (total_flow[i] < 0.9 * Network->K[i]) )
                {
                    PDCGM_env->rwstat[i] = OBJECTIVE;
                    active_set[i] = 0;
                    active_set_n--;
                    
                    u[i] = 0.0;
                    j--;
                }
            }
            free(total_flow);
        }    
    }
    
    /* Allocate memory for a column vector */
    column = PDCGM_ALLOC(double, nr);
    
    /* Set the network costs using the dual solution and the linear coefficient costs */ 
    for (j = 0; j < Network->m; j++) 
    { 
        if(USE_ACTIVE_SET)
        {
            if(active_set[j] == 1) net_cost[j] = u[j] + Network->cost[j];
            else net_cost[j] = Network->cost[j];
        }
        else net_cost[j] = u[j] + Network->cost[j];
    }
    
    /* Solve the pricing subproblems of the first type (one shortest path for each commodity) */
    #pragma omp parallel for private(i)
    for (i = 0; i < nb_demande; i++)
    {        
        /* Call the Dijkstra algorithm to solve the pricing subproblem */
        if(dijkstra_from_node[(demande + i)->s] == 0) Dijkstra_solve(Network, (demande + i)->s, (demande + i)->t, net_cost);
    }
    
    /* Solve the pricing subproblem (shortest path) */
    for (i = 0; i < nb_demande; i++)
    {
        /* Generate the column from the optimal path */
        add_solution_from_Dijkstra(Network, column, (demande + i)->s, (demande + i)->t, (demande + i)->val);
        
        /* Set the vector of solution costs */
        d = d_tableau + (demande + i)->s * Network->n;
        
        /* Set the relative cost (the value of the oracle)  */
        *dual_violation += d[(demande + i)->t] * (demande + i)->val;
    }
    
    /* Update the relative cost: add the dual associated to the convexity constraint */
    *dual_violation -= 1.0 * u[convexity_row];
    nc = 0;
    
    /* Check if the relative cost is negative */
    if( PDCGM_env->outer < 1 || *dual_violation < -MCNF_TOL_ZERO)
    {
        /* Set the entry in the convexity row */
        column[convexity_row] = 1.0;
        
        /* Set the objective function cost */
        cost = 0.0;
        for (j = 0; j < nr-2; j++)
        {
            if(column[j] < -MCNF_TOL_ZERO) cost += column[j] * Network->cost[j];
        }
        column[objective_row] = -cost; 
        
        /* Add the new column to the current master problem */
        nc = PDCGM_add_dense_columns(PDCGM_env, column, NULL, nr, 1);
        
        if(PDCGM_env->rwstat[convexity_row] == OBJECTIVE) 
        {
            PDCGM_env->rwstat[convexity_row] = GREATER_EQUAL;
        }
    }
    
    if(nc == 0) *dual_violation = 0.0;
    
    /* Free up the memory allocated */
    free(column);

    return answer_type;
}


/* ORACLE FUNCTION FOR THE LINEAR MCNF WITH *DISAGGREGATED* MASTER PROBLEM */
/* The oracle() procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short MCNF_linear_disagg_oracle(
    double *primal_violation,   /* total violation of the generated constraints (not used in this application) */
    double *dual_violation,     /* total violation of the generated columns (the value f(y)) */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data (not used in this example) */
{
    
    int i, j, j_max, nr, nc, nnz, ret;
        
    double aux_double, cost, answer_type, *d, 
           *u; /* a pointer to the dual solution of the current master problem */
        
    PDCGM_SMatrix_CW M, M2; /* Auxiliary matrix to store the generated columns (sparse representation) */
    PriorityQueue_type *priority_queue; /* priority queue used to sort elements */
    
    printf ("Hello from MCNF_linear_disagg_oracle() -- Linear multicommodity network flow with disaggregated MP.\n");
    
    /* Set the answer type: column generation */
    if(PDCGM_env->outer > 0) answer_type = 1; 
    else answer_type = 0; 
    
    /* Reset the number of columns added in the last interation */
    PDCGM_env->ncols_last = 0;
    
    /* Reset the pricing subproblem solver */
    Reset_Dijkstra (Network);
    
    /* Set the number of dense entries in each generated column */
    nr = Network->m + nb_demande + 1; /* number of rows = number of arcs + convexity rows + objective row */
    
    /* Set the max number of columns to be generated */
    nc = nb_demande;
    
    /* Set the max number of nonzeros in a column */
    nnz = Network->m + 2;
    
    /* Reset the dual_violation (the value of the oracle) */
    *dual_violation = 0.0;
    
    /* Get a pointer to the dual solution of the current master problem */
    u = PDCGM_get_dual_solution(PDCGM_env);

    if(USE_ACTIVE_SET)
    {
        /* Update the active set */
        if(PDCGM_env->outer > 0)
        {
            PDCGM_SMatrix_CW M;
            double *total_flow;
            double *lambda;
            int lambda_n;
            
            PDCGM_get_pointer_to_master_columns (PDCGM_env, &M);
            PDCGM_get_pointer_to_master_solution(PDCGM_env, &lambda, &lambda_n);
            total_flow = PDCGM_ALLOC(double, Network->m);
            
            for(i=0; i<M.n; i++)
            {
                for (j = M.clpnts[i] - 1; j < M.clpnts[i+1] - 1; j++)
                {
                    if(M.rwnmbs[j] - 1 < Network->m)
                        total_flow[ M.rwnmbs[j] - 1 ] -= M.coeff[j] * lambda[i];
                }
            }
            
            j_max = 0.1 * Network->m;            
            j = 0;
            for(i=0; i<Network->m && j < j_max; i++) 
            {
                if( (active_set[i] == 0) && (total_flow[i] > Network->K[i]) )
                {
                    PDCGM_env->rwstat[i] = EQUAL;
                    active_set[i] = 1;
                    active_set_n++;
                    
                    u[i] = 0.001;
                    answer_type = 0;
                    j++;
                }
                else if( (active_set[i] == 1) && (total_flow[i] < 0.9 * Network->K[i]) )
                {
                    PDCGM_env->rwstat[i] = OBJECTIVE;
                    active_set[i] = 0;
                    active_set_n--;
                    
                    u[i] = 0.0;
                    j--;
                }
            }
            free(total_flow);
        }    
    }
    
    j = 0;
    if(nb_demande > 60000) j_max = nb_demande / 15.0 + 1;
    else if(nb_demande > 20000) j_max = nb_demande / 10.0 + 1;
    else if(nb_demande > 10000) j_max = nb_demande / 5.0 + 1;
    else j_max = nb_demande;
    
    for(i = Network->m; i < Network->m + nb_demande && j < j_max; i++) 
    {
        if(PDCGM_env->rwstat[i] == OBJECTIVE) 
        {
            PDCGM_env->rwstat[i] = GREATER_EQUAL;
            j++;
            u[i] = 100.0;
        }
    }
    if(j > 0) answer_type = 0;
    
    /* Set the network costs using the dual solution and the linear coefficient costs */ 
    for (j = 0; j < Network->m; j++) 
    { 
        if(USE_ACTIVE_SET)
        {
            if(active_set[j] == 1) net_cost[j] = u[j] + Network->cost[j];
            else net_cost[j] = Network->cost[j];
        }
        else net_cost[j] = u[j] + Network->cost[j];
    }

    /* Set the sparse matrix that will store the generated columns */
    PDCGM_set_SMatrix_CW(&M, nr, nc, nc*nnz);
    M.m = nr; /* number of rows */
    M.n = 0;  /* no columns yet */
    M.nz = 0; /* no coefficient yet */ 
    M.clpnts[0] = 1; /* index of the first column */
    
    
    /* Set the sparse matrix that will store the generated columns */
    PDCGM_set_SMatrix_CW(&M2, nr, nc, nc*nnz);
    M2.m = nr; /* number of rows */
    M2.n = 0;  /* no columns yet */
    M2.nz = 0; /* no coefficient yet */ 
    M2.clpnts[0] = 1; /* index of the first column */
    
    /* Allocate a memory for a queue that will be used to sort the values */
    priority_queue = (PriorityQueue_type *) malloc( sizeof(PriorityQueue_type) );
    ret = PriorityQueue_allocate(priority_queue, nb_demande);
    if(ret < 0) return -1;
    
    /* Solve the pricing subproblems of the first type (one shortest path for each commodity) */
    #pragma omp parallel for private(i)
    for (i = 0; i < nb_demande; i++)
    {        
        /* Call the Dijkstra algorithm to solve the pricing subproblem */
        if(dijkstra_from_node[(demande + i)->s] == 0) Dijkstra_solve(Network, (demande + i)->s, (demande + i)->t, net_cost);
    }
    
    /* Solve the pricing subproblems of the first type (one shortest path for each commodity) */
    for (i = 0; i < nb_demande; i++)
    {        
        /* Set the vector of solution costs */
        d = d_tableau + (demande + i)->s * Network->n;

        if( (PDCGM_env->rwstat[Network->m+i] == GREATER_EQUAL) || d[(demande + i)->t] * (demande + i)->val - u[Network->m + i] < -MCNF_TOL_ZERO)
        {
            /* Generate the column from the optimal path */
            get_solution_from_Dijkstra(Network, &M, (demande + i)->s, (demande + i)->t, (demande + i)->val);
        
            /* Set the entry in the convexity row */
            M.coeff[M.nz] = 1.0;
            M.rwnmbs[M.nz] = Network->m + i + 1;
            M.nz++;
            
            /* Set the objective function cost */
            cost = 0.0;
            for (j = M.clpnts[M.n] - 1; j < M.nz - 1; j++)
            {
                cost += M.coeff[j] * Network->cost[M.rwnmbs[j] - 1];
            }
            M.coeff[M.nz] = -cost;
            M.rwnmbs[M.nz] = M.m;
            M.nz++;

            PriorityQueue_insert(priority_queue, d[(demande + i)->t] * (demande + i)->val - u[Network->m + i], M.n);

            /* Update the number of columns and set the first index of the next column */
            M.n++;
        }
        M.clpnts[M.n] = M.nz + 1;
        
        /* Set the relative cost (the value of the oracle)  */
        if(d[(demande + i)->t] * (demande + i)->val - u[Network->m + i] < 0) 
            *dual_violation += d[(demande + i)->t] * (demande + i)->val - u[Network->m + i];
    
        if(PDCGM_env->rwstat[Network->m+i] == GREATER_EQUAL) PDCGM_env->rwstat[Network->m+i] = EQUAL;
    }
    
    while( ( (PDCGM_env->outer < 1) || (M2.n < nb_demande) ) && (priority_queue->n > 0) && PriorityQueue_getNext(priority_queue, &aux_double, &i) )
    {
        for (j = M.clpnts[i] - 1; j < M.clpnts[i+1] - 1; j++)
        {
            M2.coeff[M2.nz] = M.coeff[j];
            M2.rwnmbs[M2.nz] = M.rwnmbs[j];
            M2.nz++;
        }
        M2.n++;
        M2.clpnts[M2.n] = M2.nz + 1;
    }

    PriorityQueue_free(priority_queue);
    free(priority_queue);
        
    /* Add the new columns to the current master problem */
    nc = PDCGM_add_columns(PDCGM_env, &M2, NULL);

    /* Clean memory used by the auxiliary matrix */     
    PDCGM_free_SMatrix_CW(&M);
    PDCGM_free_SMatrix_CW(&M2);
    
    if(nc == 0) *dual_violation = 0.0;
    
    return answer_type;
}

/* ORACLE FUNCTION FOR THE LINEAR MCNF WITH *DISAGGREGATED* MASTER PROBLEM */
/* The oracle() procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short MCNF_linear_grouped_oracle(
    double *primal_violation,   /* total violation of the generated constraints (not used in this application) */
    double *dual_violation,     /* total violation of the generated columns (the value f(y)) */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data (not used in this example) */
{
    
    int i, j, j_max, k, h, r, nr, nc;
        
    double aux_double, answer_type, *d,
           *u, /* a pointer to the dual solution of the current master problem */
           *column;
        
    printf ("Hello from MCNF_linear_grouped_oracle() -- Linear multicommodity network flow with (semi-)aggregated MP.\n");
    
    /* Set the answer type: column generation */
    if(PDCGM_env->outer > 0) answer_type = 1; 
    else answer_type = 0; 

    /* Reset the number of columns added in the last interation */
    PDCGM_env->ncols_last = 0;
        
    /* Reset the pricing subproblem solver */
    Reset_Dijkstra (Network);
    
    /* Set the number of dense entries in each generated column */
    nr = Network->m + source_nodes_n + 1; /* number of rows = number of arcs + convexity rows + objective row */
    
    /* Set the max number of columns to be generated */
    nc = Network->n;
    
    /* Reset the dual_violation (the value of the oracle) */
    *dual_violation = 0.0;
    
    /* Get a pointer to the dual solution of the current master problem */
    u = PDCGM_get_dual_solution(PDCGM_env);

    if(USE_ACTIVE_SET)
    {
        /* Update the active set */
        if(PDCGM_env->outer > 0)
        {
            PDCGM_SMatrix_CW M;
            double *total_flow;
            double *lambda;
            int lambda_n;
            
            PDCGM_get_pointer_to_master_columns (PDCGM_env, &M);
            PDCGM_get_pointer_to_master_solution(PDCGM_env, &lambda, &lambda_n);
            total_flow = PDCGM_ALLOC(double, Network->m);
            
            for(i=0; i<M.n; i++)
            {
                for (j = M.clpnts[i] - 1; j < M.clpnts[i+1] - 1; j++)
                {
                    if(M.rwnmbs[j] - 1 < Network->m)
                        total_flow[ M.rwnmbs[j] - 1 ] -= M.coeff[j] * lambda[i];
                }
            }
            
            j = 0;
            for(i=0; i<Network->m && j < 0.1 * Network->m; i++) 
            {
                if( (active_set[i] == 0) && (total_flow[i] > Network->K[i]) )
                {
                    printf("+");
                    
                    PDCGM_env->rwstat[i] = EQUAL;
                    active_set[i] = 1;
                    active_set_n++;
                    
                    u[i] = 0.001;
                    answer_type = 0;
                    j++;
                }
                else if( (active_set[i] == 1) && (total_flow[i] < 0.85 * Network->K[i]) )
                {
                    printf("-");
                    
                    PDCGM_env->rwstat[i] = OBJECTIVE;
                    active_set[i] = 0;
                    active_set_n--;
                    
                    u[i] = 0.0;
                    j--;
                }
            }
            
            free(total_flow);
        }
    }
    
    j = 0;
    j_max = source_nodes_n/2.0 + 1;
    for(i = Network->m; i < Network->m + source_nodes_n && j < j_max; i++) 
    {
        if(PDCGM_env->rwstat[i] == OBJECTIVE) 
        {
            PDCGM_env->rwstat[i] = GREATER_EQUAL;
            j++;
            u[i] = 100.0;
        }
    }
    if(j > 0) answer_type = 0;
    
    /* Set the network costs using the dual solution and the linear coefficient costs */ 
    for (j = 0; j < Network->m; j++) 
    { 
        if(USE_ACTIVE_SET)
        {
            if(active_set[j] == 1) net_cost[j] = u[j] + Network->cost[j];
            else net_cost[j] = Network->cost[j];
        }
        else net_cost[j] = u[j] + Network->cost[j];
    }

    /* Allocate memory for a column vector */
    column = PDCGM_ALLOC(double, nr);

    /* Solve the pricing subproblems of the first type (one shortest path for each commodity) */
    #pragma omp parallel for private(i)
    for (i = 0; i < nb_demande; i++)
    {        
        /* Call the Dijkstra algorithm to solve the pricing subproblem */
        if(dijkstra_from_node[(demande + i)->s] == 0) Dijkstra_solve(Network, (demande + i)->s, (demande + i)->t, net_cost);
    }
    
    /* Solve the pricing subproblems (one shortest path for each source node) */
    r = 0;
    nc = 0;
    for (h = 0; h < Network->n; h++)
    {        
        if(source_nodes_len[h] > 0) 
        {
            /* Initialize the column */    
            for (j = 0; j < nr; j++) column[j] = 0.0;
            
            aux_double = 0.0;
            for (k = 0; k < source_nodes_len[h]; k++)
            {
                /* Get the k-th commodity in the group */
                i = source_nodes_commodities[source_nodes[h] + k];
                
                /* Set the vector of solution costs */
                d = d_tableau + (demande + i)->s * Network->n;
                
                /* Generate the column from the optimal path */
                add_solution_from_Dijkstra(Network, column, (demande + i)->s, (demande + i)->t, (demande + i)->val);
                
                /* Set the relative cost (the value of the oracle)  */
                aux_double += d[(demande + i)->t] * (demande + i)->val - u[Network->m + r];
            }

            /* Check if the column has a negative relative cost  */
            if(PDCGM_env->outer < 1 || aux_double < -MCNF_TOL_ZERO)
            {
                /* Set the entry in the convexity row */
                column[Network->m + r] = 1.0;
                
                /* Set the objective function cost */
                column[nr-1] = 0.0;
                for (j = 0; j < Network->m; j++) column[nr-1] -= column[j] * Network->cost[j];
                
                /* Add the new columns to the current master problem */
                nc += PDCGM_add_dense_columns(PDCGM_env, column, NULL, nr, 1);
                
                /* Update the relative cost (the value of the oracle)  */
                *dual_violation += aux_double; 
            }
            
            /* Go to the next group */
            r++;
        }
    }
    
    if(nc == 0) *dual_violation = 0.0;
    
    free(column);
    
    return answer_type;
}

/* Solve the multicommodity network flow problem */
int MCNF_linear_solve(Graph * G, Demande * dem, int nb_dem, int aggregated)
{
    int dim = G->m, /* number of linking constraints in the problem: number of arcs */
        nb_sub_pb,   /* number of subproblems (= number of convexity constraints) */
        max_n_cols = 10000,
        max_outer = 5000, 
        k;
    
    PDCGM *PDCGM_env; /* PDCGM environment */
    
    FILE *out_file; /* output file pointer  */
    
    /* Set the number of subproblems according to the formulation */
    if(aggregated == 1) nb_sub_pb = 1; /* only 1 aggregated subproblem */
    else if(aggregated == 2) nb_sub_pb = source_nodes_n; /* the subproblems are aggregated by same source node */
    else nb_sub_pb = nb_dem; /* number of commodities */
    
    /* Allocate memory for the MP formulation */
    double *b  = PDCGM_ALLOC(double, dim + nb_sub_pb + 1); /* RHS of the master problem */
    double *y0 = PDCGM_ALLOC(double, dim + nb_sub_pb + 1); /* initial guess for the dual solution */
    double *lo_box = PDCGM_ALLOC(double, dim); /* lower bounds of y */
    double *up_box = PDCGM_ALLOC(double, dim); /* upper bounds of y */
    short  *row_type = PDCGM_ALLOC(short, dim + nb_sub_pb + 1); /* type of each constraint in the master problem */

    /* Set the linking constraints and the associated duals */
    for(k = 0; k < dim; k++) 
    {
        b[k]  = -Network->K[k];
        y0[k] = 0.0;
        if(USE_ACTIVE_SET) row_type[k] = OBJECTIVE;
        else row_type[k] = EQUAL;
        lo_box[k]  = 0.0;
        if(aggregated > 0) up_box[k]  = 100.0;
        else up_box[k]  = 1.0;
    }
    
    /* Set the convexity constraints and the associated duals */
    for (k = dim; k < dim + nb_sub_pb; k++) 
    {
        b[k]  = 1.0;
        y0[k] = 0.0;
        row_type[k] = OBJECTIVE; // REMARK: WE RESET AS EQUALITY INSIDE THE ORACLE
    }
    
    /* Set the objective function (last row in the coefficient matrix) */
    b[dim+nb_sub_pb] = 0.0;
    y0[dim+nb_sub_pb] = 0.0;
    row_type[dim+nb_sub_pb] = OBJECTIVE;

    /* POPULATE THE INTERNAL DATA STRUCTURE OF PDCGM */ 
    PDCGM_env = PDCGM_set_data( 
        dim,  /* number of kept constraints in the RMP */
        nb_sub_pb,  /* number of convexity constraints in the RMP */
        (dim + nb_sub_pb + 1),  /* maximum number of nonzeros in a column */
        nb_sub_pb,     /* maximum number of columns generated by the oracle */
        max_outer,     /* maximum number of outer iterations */
        y0, /* initial guess of the dual solution (may be NULL) */
        b,  /* RHS of each constraint in the RMP */
        row_type, /* type of each constraint (row) in the RMP */
        lo_box,   /* lower bound vector of the DUAL variables in the RMP */
        up_box,   /* upper bound vector of the DUAL variables in the RMP */
        NULL);   /* instance data */
                
    /* Check if the PDCGM environment was created successfully */
    if(PDCGM_env == NULL)   
    {
        printf("-- ERROR: PDCGM_set_data() returned NULL --\n");
        fflush(stdout);
        return 0;
    }
            
    /* Set the optimality tolerance for the column generation algorithm */
    PDCGM_set_delta(PDCGM_env, 1.E-5);   
    
    /* Set the verbose mode (how much information is printed) */
    PDCGM_set_verbosity(PDCGM_env, 1);
    
    /* Set the maximum number of columns as a large value */
    if(aggregated == 1) 
    {
        max_n_cols = 1000;
    }
    else if(aggregated == 2) 
    {
        if(max_n_cols < 10 * G->m) max_n_cols = 10 * G->m;
    }
    else
    {
        if(max_n_cols < 2 * nb_sub_pb) max_n_cols = 2 * nb_sub_pb;
    }
    
    PDCGM_set_max_number_columns(PDCGM_env, max_n_cols);
    PDCGM_env->start_from_reduce_matrix = 0;
    
    /* Tell PDCGM that the first iterations may be difficult */
    PDCGM_env->use_difficult_iters = 1; 
    
    /* Set maximum CPU time */
    PDCGM_set_max_cputime(PDCGM_env, 50000.0);
    
    /* Enable column elimination on *insertion* */
    PDCGM_set_column_elimination(PDCGM_env, 1);
    
    /* Start solving the master problem with tighter bounds */
    if(aggregated == 1) PDCGM_set_dual_box(PDCGM_env, 0.0, 10.0);
    
    /* Reset the level of centrality required from the interior point method */
    if(aggregated == 1) 
    {
        PDCGM_set_centrality_value(PDCGM_env, 1);
    }
    else if(aggregated == 2) 
    {
        PDCGM_set_centrality_value(PDCGM_env, 1);
    }
    else 
    {
        PDCGM_set_centrality_value(PDCGM_env, 0);
        PDCGM_env->WS_strategy = 0;    
    }
    
    /* Set an (artificial) initial column to avoid infeasibility */
    //double *initial_column = PDCGM_ALLOC(double, dim + nb_sub_pb + 1);
    //for(k=dim; k < dim + nb_sub_pb; k++) initial_column[k] = 1.0;
    //initial_column[dim + nb_sub_pb] = 1.E+2;    
    //PDCGM_add_dense_columns(PDCGM_env, initial_column, NULL, dim + nb_sub_pb + 1, 1);

    /* START THE COLUMN GENERATION PROCEDURE */
    
    /* Solve the LINEAR multicommodity network flow problem */
    if(aggregated == 1)
    {
        /* Set the degree of optimality (parameter D > 1) */
        PDCGM_set_degree_of_optimality(PDCGM_env, 1.1);

        /* Call PDCGM */
        PDCGM_solve_MP(PDCGM_env, MCNF_linear_agg_oracle);
    }
    else if(aggregated == 2)
    {
        /* Set the degree of optimality (parameter D > 1) */
        PDCGM_set_degree_of_optimality(PDCGM_env, 5.0);

        /* Call PDCGM */
        PDCGM_solve_MP(PDCGM_env, MCNF_linear_grouped_oracle);
    }
    else
    {
        /* Set the degree of optimality (parameter D > 1) */
        PDCGM_set_degree_of_optimality(PDCGM_env, 10.0);

        /* Call PDCGM */
        PDCGM_solve_MP(PDCGM_env, MCNF_linear_disagg_oracle);
    }
    
    /* Print the solution */
    double *lambda;
    int lambda_n;
    PDCGM_get_pointer_to_master_solution(PDCGM_env, &lambda, &lambda_n);
    
    printf("Active sets: %4.2lf \n", (1.0 * active_set_n) / G->m);
    printf("Optimal value: %1.5E \n\n", max_cap*max_cost*PDCGM_env->upperBnd);
    
    /* Printf the results in an output file */
    out_file = fopen("output-mcnf-pdcgm.txt", "a");
    fprintf(out_file, "\t%1.6E \t%1.6E \t%1.6E \t%lf \t%d \t%lf \t%lf \t%lf ",  
        max_cap*max_cost*PDCGM_env->lowerBnd, max_cap*max_cost*PDCGM_env->upperBnd, 
        PDCGM_env->rel_gap, (100.0 * active_set_n) / G->m, PDCGM_env->outer, PDCGM_env->cputime_RMP, PDCGM_env->cputime_oracle, PDCGM_env->cputime_CG);
    fflush(out_file);
    fclose(out_file);
    
    /* Clean the memory before returning */
    PDCGM_free(PDCGM_env);
    free(PDCGM_env);
    free(b);
    free(y0);
    free(row_type);
    free(lo_box);
    free(up_box);
    
    return 0;
}

/* C MAIN FUNCTION */
int main (int argc, char *argv[])
{
    /* Auxiliary variables */
    int nb_dem, aggregated, oriented, begin_zero, with_cost;
        
    Graph G; /* graph representing the network */
    Demande *dem; /* demand of each commodity */
    
    char DefltFlnme[1000]; /* instance file name */
    FILE *out_file; /* output file pointer  */

    /* Parse the command line receive as an input */
    parse_input (argc, argv, &aggregated, &oriented, DefltFlnme, &begin_zero, &with_cost);

    /* Summarize the input data */  
    printf ("\nProblem filename: %s\n", DefltFlnme);

    out_file = fopen("output-mcnf-pdcgm.txt", "a");
    fprintf(out_file, "\n%s ", DefltFlnme);
    fflush(out_file); 
    fclose(out_file);
        
    /* Read the instance data */
    read_instance(DefltFlnme, begin_zero, aggregated);

    /* Set initial parameters and initialize the oracle data structure */
    set_network_Dijkstra(&G, &dem, &nb_dem, oriented);
    init_Dijkstra(&G, dem, nb_dem);
    
    /* Solve the problem by column generation */
    MCNF_linear_solve(&G, dem, nb_dem, aggregated);

    return 0;
}
