// ------------------------------------------------------------------
//  PDCGM  (Primal-Dual Column Generation Method)
//  See Gondzio, González-Brevis and Munari, EJOR, 2012.
//      Gondzio and Sarkissian, 1996.
//
//  On top of: HOPDM (Higher Order Primal-Dual Method)
//  Copyright: Jacek Gondzio, 1990, 2010.
//
//  COLUMN GENERATION FOR THE MULTIPLE KERNEL LEARNING PROBLEM
//  
//  Authors: Pedro Munari [pedro.munari@gmail.com], 
//           Pablo González-Brevis, 
//           Jacek Gondzio
//            
//  Last Modified: September, 2012.
//
//  Using Shogun toolbox http://www.shogun-toolbox.org
//  to efficiently solve the pricing subproblem 
// ------------------------------------------------------------------


// Bring in basic C and C++ headers
#include <iostream>
#include <fstream>
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// Bring in PDCGM/HOPDM headers 
// They are implemented in C so we have to use 'extern' 
extern "C" {
#include "../../pdcgm/pdcgm.h"
}

// Bring in Shogun headers 
#include <shogun/labels/BinaryLabels.h>
#include <shogun/features/DenseFeatures.h>
#include <shogun/kernel/GaussianKernel.h>
#include <shogun/kernel/PolyKernel.h>
#include <shogun/kernel/CombinedKernel.h>
#include <shogun/kernel/normalizer/AvgDiagKernelNormalizer.h>
#include <shogun/classifier/svm/LibSVM.h>
#include <shogun/base/init.h>
#include <shogun/lib/common.h>
#include <shogun/io/SGIO.h>

// Define namespaces
using namespace std;
using namespace shogun;

#define MKL_TOL_ZERO 1.E-6
//#define USE_NORMALIZER

// Data structure for the instance data
typedef struct
{
    int32_t n_vectors;   // number of input vectors x_i
    int32_t n_features;  // number of features in the input vectors
    int32_t n_kernels;   // number of kernels that are used
    int32_t n_training;  // number of training input vector
    
    int32_t* labels;     // labels of input vectors
    SGMatrix< float64_t > training_vectors; // input vectors used to train the svm
    SGMatrix< float64_t > test_vectors;     // input vectors used to test the svm
    CKernel** kernels;   // kernels currently used in the svm
    CLibSVM *svm;        // svm object used to call shogun
} 
T_instance_data;

// ----------------------------------------------------------------------------
// Function: MKL_print_message
// ----------------------------------------------------------------------------
// Printing function used in shogun
// ----------------------------------------------------------------------------
void MKL_print_message(FILE* target, const char* str)
{
    fprintf(target, "%s", str);
}

// ----------------------------------------------------------------------------
// Function: MKL_compute_kernels
// ----------------------------------------------------------------------------
// Compute all the kernel matrices of type k(x_i, x_j). The x_i must be the
// training vectors and in case 'n_test > 0' then x_j are the test vectors. 
// Otherwise, x_j are also the training vectors. 
// ----------------------------------------------------------------------------
int MKL_compute_kernels(
    T_instance_data *instance,
    int32_t n_test)
{
    int32_t aux_int;
    float64_t aux_double;

    // Assign the training features to a features object 
    CDenseFeatures<float64_t> *training_features = new CDenseFeatures<float64_t>(instance->training_vectors);
    SG_REF(training_features);
    
    // Create an array of single feature vectors
    SGMatrix< float64_t > single_training_mat = SGMatrix< float64_t >(1, instance->n_training);
    CDenseFeatures<float64_t> **single_training_features = new CDenseFeatures<float64_t>*[instance->n_features];
    
    for(int32_t i = 0; i < instance->n_features; i++) 
    {
        // Copy the feature from all input vectors
        for(int32_t j = 0; j < instance->n_training; j++) single_training_mat(0,j) = instance->training_vectors(i,j);
        single_training_features[i] = new CDenseFeatures<float64_t>(single_training_mat);
        SG_REF(single_training_features[i]);
    }
    
    // Check if test vectors should be used
    CDenseFeatures<float64_t> *test_features;
    CDenseFeatures<float64_t> **single_test_features;
    if(n_test > 0)
    {
        // Assign the test features to a features object 
        test_features = new CDenseFeatures<float64_t>(instance->test_vectors);
        SG_REF(test_features);
    
        // Create an array of single test vectors
        SGMatrix< float64_t > single_test_mat = SGMatrix< float64_t >(1, n_test);
        single_test_features = new CDenseFeatures<float64_t>*[instance->n_features];
    
        for(int32_t i = 0; i < instance->n_features; i++) 
        {
            // Copy the feature from all input vectors
            for(int32_t j = 0; j < n_test; j++) single_test_mat(0,j) = instance->test_vectors(i,j);
            single_test_features[i] = new CDenseFeatures<float64_t>(single_test_mat);
            SG_REF(single_test_features[i]);
        }
    }
    else
    {
        // Use only training vectors
        test_features = training_features;
        single_test_features = single_training_features;
    }
    
    // Set the array of kernels
    instance->n_kernels = 13 * (instance->n_features + 1);
    instance->kernels = new CKernel*[instance->n_kernels];
    int32_t i = 0;
    
    #ifdef USE_NORMALIZER
    // Set the array of normalizers (for gaussian only)
    CAvgDiagKernelNormalizer **normalizer = new CAvgDiagKernelNormalizer*[10 * (instance->n_features + 1)];
    #endif
    
    // Create Gaussian kernels with cache 10MB and widths 0.125, 0.25, 0.5, 1.0, ...., 64 
    aux_double = 0.125;
    for(int32_t k = 0; k < 10; k++)
    {
        // Set the kernel using all features
        instance->kernels[i] = new CGaussianKernel(100, aux_double);
        #ifdef USE_NORMALIZER
        normalizer[i] = new CAvgDiagKernelNormalizer;
        instance->kernels[i]->set_normalizer(normalizer[i]);
        #endif
        instance->kernels[i]->init(training_features, test_features);
        i++;
        
        // Set the kernel for each (single) feature
        for(int32_t j = 0; j < instance->n_features; j++) 
        {
            instance->kernels[i] = new CGaussianKernel(100, aux_double);
            #ifdef USE_NORMALIZER
            normalizer[i] = new CAvgDiagKernelNormalizer;
            instance->kernels[i]->set_normalizer(normalizer[i]);
            #endif
            instance->kernels[i]->init(single_training_features[j], single_test_features[j]);
            i++;
        }
    
        // Increase the width
        aux_double *= 2.0;
    }
    
    // Create Polynomial kernels with cache 10MB and degree 1 to 3
    aux_int = 1;
    for(int32_t k = 0; k < 3; k++)
    {
        instance->kernels[i] = new CPolyKernel(100, aux_int);
        // Remark: Poynomial kernels are already normalized
        instance->kernels[i]->init(training_features,test_features);
        i++;
        
        // Set the kernel for each (single) feature
        for(int32_t j = 0; j < instance->n_features; j++) 
        {
            instance->kernels[i] = new CPolyKernel(100, aux_int);
            // Remark: Poynomial kernels are already normalized
            instance->kernels[i]->init(single_training_features[j], single_test_features[j]);
            i++;
        }
        
        // Increase the degree
        aux_int++;
    }

    return 0;
}

// ----------------------------------------------------------------------------
// Function: MKL_read_instance
// ----------------------------------------------------------------------------
// Read an instance file and fill in the data structures. Also, allocate the
// kernels in memory and initialize them. 
// ----------------------------------------------------------------------------
int MKL_read_instance(
    char   *file_name,         // path of the instance file 
    T_instance_data *instance) // data structure to store the input
{
    int32_t aux_int, n_test;
    float64_t aux_double;

    instance->n_vectors  = 0; // number of input vectors x_i
    instance->n_features = 0; // number of features in the input vectors
    instance->n_kernels  = 0; // number of kernels that are used
    instance->n_training = 0; // number of training input vector

    // Read the dimensions of the instance
    ifstream input_file(file_name);
    if (input_file.is_open())
    {
        // Read the number of input vectors 
        input_file >> aux_int;
        instance->n_vectors = aux_int;
        
        // Read the number of features 
        input_file >> aux_int;
        instance->n_features = aux_int;
        
        // Read how many input vectors should be used for training 
        input_file >> aux_int;
        instance->n_training = aux_int;
    }    

    if(instance->n_vectors < 1 || instance->n_features < 1)
    {
        cout << "### The input file is not valid." << endl;
        return -1;
    }
    
    // Allocate memory for the matrices that store training and test input vectors
    n_test = instance->n_vectors - instance->n_training;
    instance->training_vectors = SGMatrix< float64_t >(instance->n_features, instance->n_training);
    if(n_test > 0) instance->test_vectors = SGMatrix< float64_t >(instance->n_features, n_test);
    
    // Allocate memory for the array of labels
    instance->labels = new int32_t[instance->n_vectors];
    
    // Read information about each training vector
    int32_t t = 0;
    for(int32_t j = 0; j < instance->n_training; j++)
    {
        // Set all the features of the current input vector
        for(int32_t i = 0; i < instance->n_features; i++) 
        {
            input_file >> aux_double;
            instance->training_vectors(i,j) = aux_double;
        }
        
        // Set the label of the current input vector 
        input_file >> aux_double;
        instance->labels[t++] = aux_double;
    }
    
    // Read information about each training vector
    for(int32_t j = 0; j < n_test; j++)
    {
        // Set all the features of the current input vector
        for(int32_t i = 0; i < instance->n_features; i++) 
        {
            input_file >> aux_double;
            instance->test_vectors(i,j) = aux_double;
        }
        
        // Set the label of the current input vector 
        input_file >> aux_double;
        instance->labels[t++] = aux_double;
    }
    input_file.close();
    
    // Print the information
    //instance->training_vectors.display_matrix();
    //instance->test_vectors.display_matrix();
    //for(int32_t i = 0; i < instance->n_vectors; i++) cout << instance->labels[i] << endl;
    
    // Compute the kernels for the training vectors
    MKL_compute_kernels(instance, 0);
    
    // Create the svm object
    instance->svm = new CLibSVM;
    
    return 0;
}

// ----------------------------------------------------------------------------
// Function: MKL_solve_SVM 
// ----------------------------------------------------------------------------
// Given the kernel weigths in 'betas', this function creates an SVM with 
// a combined kernel and calls shogun to solve the SVM. The optimal solution
// of the SVM is used to generate a column which is returned together with_cost
// the relative cost of the column 
// ----------------------------------------------------------------------------
int MKL_solve_SVM(
    T_instance_data *instance,   // ( ->) Pointer to instance data 
    double C,                    // ( ->) Regularization parameter of the SVM
    SGVector< float64_t > betas, // ( ->) Kernel weights
    double *Sk_alpha,            // (<- ) Array with the Sk(alpha) values for each k
    double *sum_betak_Sk_alpha)  // (<- ) Sum betak * Sk(alpha) over k
{
    
    // Set the combined kernel
    CCombinedKernel* kernel = new CCombinedKernel();
    SG_REF(kernel);
    
    for(int32_t i = 0; i < instance->n_kernels; i++) kernel->append_kernel(instance->kernels[i]);
    kernel->set_subkernel_weights(betas); 
 
    // Create a labels object
    CBinaryLabels *bin_labels = new CBinaryLabels(instance->n_training);
    SG_REF(bin_labels);
    
    // Read the labels and assign to the corresponding object
    for (int32_t i=0; i<instance->n_training; i++) 
        bin_labels->set_int_label(i, instance->labels[i]);

    // Create svm with C = 100
    //instance->svm = new CLibSVM(C, kernel, bin_labels);
    instance->svm->set_C(100, 100);
    instance->svm->set_kernel(kernel);
    instance->svm->set_labels(bin_labels);
    
    // Train the machine
    (instance->svm)->train();

    // Get the support vectors in the optimal solution
    int32_t n_support_vectors = (instance->svm)->get_num_support_vectors();
    SGVector< int32_t > support_vectors = (instance->svm)->get_support_vectors();
    
    // Get the corresponding alphas 
    SGVector< float64_t > alphas = (instance->svm)->get_alphas();
    // REMARK: In shogun, alpha_i is already multiplied by y_i (the label)
    
    // Compute the sum of alphas 
    float64_t sum_alpha=0.0;
    for (int32_t i=0; i<n_support_vectors; i++) sum_alpha += fabs(alphas[i]);

    // Compute the sum of alphas 
    float64_t sum_alpha_y=0.0;
    for (int32_t i=0; i<n_support_vectors; i++) sum_alpha_y += alphas[i];

    // Initialize the returning parameters
    *sum_betak_Sk_alpha = 0.0;
    for (int32_t i = 0; i < instance->n_kernels; i++) Sk_alpha[i] = -sum_alpha;

    //cout << "sum_alpha = " << sum_alpha << endl;
    //cout << "sum_alpha_y = " << sum_alpha_y << endl;
    
    // Get the first kernel in the set of kernels
    CKernel* kn = ((CCombinedKernel*)kernel)->get_first_kernel();

    // Compute the Sk values for the obtained alphas
    int32_t k = 0;
    while (kn)
    {
        // Get the support vectors and compute Sk
        for (int32_t i = 0; i < n_support_vectors; i++)
        {
            for (int32_t j = 0; j < n_support_vectors; j++)
            {
                Sk_alpha[k] += 0.5 * alphas[i] * alphas[j] * kn->kernel(support_vectors[i], support_vectors[j]);
            }
        }
        
        // Positive value are irrelevante
        if(Sk_alpha[k] > MKL_TOL_ZERO) Sk_alpha[k] = 0.0;
        
        // Update the sum of Sk values (multiplied by the kernel weigths)
        *sum_betak_Sk_alpha += betas[k] * Sk_alpha[k];
        
        // Get the next kernel in the set
        SG_UNREF(kn);
        kn = ((CCombinedKernel*) kernel)->get_next_kernel();
        k++;
    }

    // Print the computed results
    //for (int32_t i = 0; i < instance->n_kernels; i++)
    //  cout << "Sk_alpha[" << i << "] = " << Sk_alpha[i] << endl;
    //cout << "---> sum_betak_Sk_alpha = " << *sum_betak_Sk_alpha << endl;

    // Free up memory
    //SG_UNREF(kernel);
        
    return 0;
}


// ----------------------------------------------------------------------------
// Function: MKL_oracle 
// ----------------------------------------------------------------------------
// Oracle function: call an SVM with combined kernels in order to generate a 
// column and add it to the RMP. The oracle() procedure must return which of 
// the following elements was generated: 0 -> nothing was generated, 
// 1 -> one or more columns, 2 -> one or more valid inequalities
// ----------------------------------------------------------------------------
static short MKL_oracle(
    double *primal_violation,   // total violation of the generated constraints (not used in this application)
    double *dual_violation,     // total violation of the generated columns (the value f(y))
    PDCGM  *PDCGM_env,          // PDCGM environment
    void   *instance_data)      // instance data (not used in this example)
{
    // Cast the pointer to instance data
    T_instance_data *instance = (T_instance_data *) instance_data;
        
    // Statement of auxiliary variables
    int nr, nc, answer_type, 
        convexity_row = instance->n_kernels; // index of the convexity constraint in the coefficient matrix 
        
    double *u,      // a pointer to the dual solution of the current master problem
           *column; // auxiliary array used to store the generated column
    
    cout << "Hello from MKL_oracle()" << endl;
    
    // Set the answer type: column generation
    answer_type = 1; 

    // Set the max number of entries in the generated column
    nr = instance->n_kernels + 2; // number of rows = number of kernels + convexity row + objective row 
    
    // Reset the dual_violation (the value of the oracle)
    *dual_violation = 0.0;

    // Allocate memory for a column vector
    column = new double[nr]; //PDCGM_ALLOC(double, nr);

    // Get a pointer to the dual solution of the current master problem 
    u = PDCGM_get_dual_solution(PDCGM_env);
    
    // Define the weight of each kernel using the dual solution
    SGVector< float64_t > betas = SGVector< float64_t >(instance->n_kernels);
    for(int32_t i = 0; i < instance->n_kernels; i++) betas[i] = u[i];
    
    // Solve the pricing subproblem (SVM with combined kernels)
    MKL_solve_SVM(instance, 10.0, betas, column, dual_violation);
    
    // Change the sign of entries in the linking constraints
    for(int32_t i = 0; i < instance->n_kernels; i++) column[i] = -column[i];
    
    // Update the relative cost: add the dual associated to the convexity constraint
    *dual_violation -= u[nr-2];
    nc = 0;
    
    // Check if the relative cost is negative
    if( PDCGM_env->outer < 1 || *dual_violation < -MKL_TOL_ZERO)
    {
        // Set the entry in the convexity row
        column[convexity_row] = 1.0;
        
        // There is no objective cost
        column[convexity_row+1] = 0.0;
        //for(int32_t i = 0; i < nr; i++) cout << column[i] << " ";
        
        // Add the new column to the current master problem
        nc = PDCGM_add_dense_columns(PDCGM_env, column, NULL, nr, 1);
    }
    
    cout << endl;
    
    if(nc == 0)
    {
        *dual_violation = 0.0;
    }
    
    // Free up the allocated memory
    delete [] column;

    return answer_type;
}


// ----------------------------------------------------------------------------
// Function: MKL_solve 
// ----------------------------------------------------------------------------
// Sets the master problem and call PDCGM to solve it
// ----------------------------------------------------------------------------
int MKL_solve(
    T_instance_data *instance,
    SGVector< float64_t > betas)
{
    int dim = instance->n_kernels,  // number of linking constraints in the problem: number of kernels 
        nb_sub_pb = 1,      // number of subproblems (= number of convexity constraints): one subproblem 
        max_outer = 1000,   // maximum number of outer iterations
        k, ret;             // auxiliaries
    
    PDCGM *PDCGM_env; // PDCGM environment

    // Allocate memory for the MP formulation 
    double *column = new double[dim + nb_sub_pb + 1];
    double *b  = new double[dim + nb_sub_pb + 1]; // RHS of the master problem
    double *u0 = new double[dim + nb_sub_pb + 1]; // initial guess for the dual solution u 
    double *lo_box = new double[dim]; // lower bounds of u 
    double *up_box = new double[dim]; // upper bounds of u 
    short  *row_type = new short[dim + nb_sub_pb + 1]; // type of each constraint in the master problem

    printf ("Hello from solve_MKL()\n");
    fflush(stdout);

    // Set the linking constraints and the associated duals 
    for(k = 0; k < dim; k++) 
    {
        b[k]  = 0.0;
        u0[k] = 0.0;
        row_type[k] = GREATER_EQUAL;
        lo_box[k]  = 0.0;
        up_box[k]  = 10.0;
    }
    u0[0] = 1.0;
    
    // Set the convexity constraint and the associated dual 
    for (k = dim; k < dim + nb_sub_pb; k++) 
    {
        b[k]  = 1.0;
        u0[k] = 0.0;
    row_type[k] = EQUAL;
//  lo_box[k]  = -1E6;
//  up_box[k]  = 0.;
    }
    
    // Set the objective function (last row in the coefficient matrix)
    b[dim+nb_sub_pb] = 0.0;
    u0[dim+nb_sub_pb] = 0.0;
    row_type[dim+nb_sub_pb] = OBJECTIVE;

    // POPULATE THE INTERNAL DATA STRUCTURE OF PDCGM 
    PDCGM_env = PDCGM_set_data( 
        dim,       // number of kept constraints in the RMP 
        nb_sub_pb, // number of convexity constraints in the RMP 
        (dim + nb_sub_pb + 1),  // maximum number of nonzeros in a column 
        nb_sub_pb, // maximum number of columns generated by the oracle
        max_outer, // maximum number of outer iterations
        u0,   // initial guess of the dual solution (may be NULL) 
        b,    // RHS of each constraint in the RMP
        row_type,  // type of each constraint (row) in the RMP
        lo_box,    // lower bound vector of the DUAL variables in the RMP
        up_box,    // upper bound vector of the DUAL variables in the RMP
        instance);     // instance data
                
    // Check if the PDCGM environment was created successfully
    if(PDCGM_env == NULL)   
    {
        printf("-- ERROR: PDCGM_set_data() returned NULL --\n");
        fflush(stdout);
        return 0;
    }
            
    // Set the optimality tolerance for the column generation algorithm
    PDCGM_set_delta(PDCGM_env, 1E-2);
    
    // Set the verbose mode (how much information is printed)
    PDCGM_set_verbosity(PDCGM_env, 0);
    
    // Set maximum CPU time 
    //PDCGM_set_max_cputime(PDCGM_env, 3000.0);
    
    // Enable column elimination on *insertion* 
    PDCGM_set_column_elimination(PDCGM_env, 1);
    
    // Set the degree of optimality (parameter D > 1)
    PDCGM_set_degree_of_optimality(PDCGM_env, 10.);

    // Set epsilon_max
    PDCGM_set_max_opt_tol(PDCGM_env, 0.5);

    // SET AN INITIAL COLUMN (AUXILIARY VARIABLE RHO)
    for(k = 0; k < dim; k++) column[k] = -1.0; // components in the linking constraints
    column[dim] = 0.0;   // convexity
    column[dim+1] = -1.0; // objective cost

    // Call the PDCGM function that add auxiliary columns to the RMP
    //PDCGM_add_auxiliary_columns_dense(PDCGM_env, column, dim + nb_sub_pb + 1, 1);
    PDCGM_add_dense_columns(PDCGM_env, column, NULL, dim + nb_sub_pb + 1, 1);

    cout << "Calling PDCGM_solve_MP()" << endl;

    // Call PDCGM to solve the MP
    ret = PDCGM_solve_MP(PDCGM_env, MKL_oracle);
    
    // Set the returning parameters with the optimal betas
    if(ret == PDCGM_COMPLETE) 
    {
        for(k = 0; k < dim; k++) betas[k] = PDCGM_env->u[k];
    }
    
    ofstream output_file("output-mkl-pdcgm.txt", ios::out | ios::app);
    output_file << -PDCGM_env->upperBnd << " " << -PDCGM_env->lowerBnd << " " << PDCGM_env->rel_gap << " " 
        << PDCGM_env->outer << " " << PDCGM_env->cputime_RMP
        << " " << PDCGM_env->cputime_oracle << " " << PDCGM_env->cputime_CG << " ";
    output_file.close();
    
    // Clean memory before returning
    PDCGM_free(PDCGM_env);
    delete PDCGM_env;
    delete column;
    delete b;
    delete u0;
    delete row_type;
    delete lo_box;
    delete up_box;
    
    // Return successfully
    return 0;
}


// ----------------------------------------------------------------------------
// Function: MKL_classify 
// ----------------------------------------------------------------------------
int MKL_classify(
    T_instance_data *instance,   // ( ->) Pointer to instance data 
    double C,                    // ( ->) Regularization parameter of the SVM
    SGVector< float64_t > betas) // ( ->) Kernel weights
{
    int32_t n_test;

    // Classify on training vectors
    {
        // Create a labels object
        CBinaryLabels *bin_labels = new CBinaryLabels(instance->n_training);
        SG_REF(bin_labels);
    
        // Read the labels and assign to the corresponding object
        for (int32_t i = 0; i < instance->n_training; i++) 
            bin_labels->set_int_label(i, instance->labels[i]);
    
        // Compute classification error
        CBinaryLabels* classification = CBinaryLabels::obtain_from_generic( (instance->svm)->apply() );

        int32_t terr=0;
        for(int32_t i = 0; i < instance->n_training; i++)
        {
            if( bin_labels->get_int_label(i) == classification->get_int_label(i) ) terr++;
        }
        SG_SPRINT("> Accuracy on training set: %d/%d (%lf) \n", terr, instance->n_training, 1.0*terr/instance->n_training);
    }

    // Classify on test vectors
    
    n_test = instance->n_vectors - instance->n_training;
    if(n_test < 1) return -1;

    // Create a labels object
    CBinaryLabels *bin_labels = new CBinaryLabels(n_test);
    SG_REF(bin_labels);
    
    // Read the labels and assign to the corresponding object
    for (int32_t i = 0; i < n_test; i++) 
        bin_labels->set_int_label(i, instance->labels[instance->n_training + i]);

    // Compute the kernels for the test vectors
    MKL_compute_kernels(instance, n_test);
        
    // Set the combined kernel
    CCombinedKernel* kernel = new CCombinedKernel();
    SG_REF(kernel);

    for(int32_t i = 0; i < instance->n_kernels; i++) kernel->append_kernel(instance->kernels[i]);
    kernel->set_subkernel_weights(betas); 
 
    // Change the kernels in the current svm
    instance->svm->set_kernel(kernel);
    
    // Compute classification error
    CBinaryLabels* classification = CBinaryLabels::obtain_from_generic( (instance->svm)->apply_binary() );

    int32_t terr=0;
    for(int32_t i = 0; i < n_test; i++)
    {
        if( bin_labels->get_int_label(i) == classification->get_int_label(i) ) terr++;
    }
    SG_SPRINT("> Accuracy on test set: %d/%d (%lf) \n\n", terr, n_test, 1.0*terr/n_test);

    ofstream output_file("output-mkl-pdcgm.txt", ios::out | ios::app);
    output_file << 1.0*terr/n_test << " ";
    output_file.close();
        
    // Free up memory
    //SG_UNREF(kernel);
    //SG_UNREF(instance->svm);

    return 0;
}


// ----------------------------------------------------------------------------
// C MAIN FUNCTION
// ----------------------------------------------------------------------------
int main (int argc, char *argv[])
{
    // Auxiliary variables
    //int i, j;
    int ret;
    char file_name[1024];
    T_instance_data instance;

    // Check if the file was specified
    if (argc > 1) strcpy(file_name, argv[1]);
    else
    {
        cout << endl << "File name not specified." << endl;        
        return 0;
    }

    // Initialize shogun environment    
    init_shogun(&MKL_print_message);

    // Read the instance file
    ret = MKL_read_instance(file_name, &instance);

    ofstream output_file("output-mkl-pdcgm.txt", ios::out | ios::app);
    output_file << endl << file_name << " ";
    //output_file << instance.n_vectors << " " << instance.n_training << " " << instance.n_features << " " << instance.n_kernels << " ";
    output_file.close();

    // Solve the problem
    if(ret == 0)
    {
        SGVector< float64_t > betas = SGVector< float64_t >(instance.n_kernels);
        for(int32_t i = 0; i < instance.n_kernels; i++) betas[i] = 0.0;
        //betas[0] = 1.0;

        // Call the MKL solver
        MKL_solve(&instance, betas);
        
        // Print the optimal betas
 
       cout << endl;
        int32_t k=0;
        for(int32_t i = 0; i < instance.n_kernels; i++) 
        {
        //    cout << "beta*[" << i << "] = " << betas[i] << endl;
            if(betas[i] > 1E-2/instance.n_kernels){
     k++;
    }else{
    betas[i]=0.;    
    }
        
    }
        cout << endl << "> Number of kernels used: " << k << endl; 
        ofstream output_file("output-mkl-pdcgm.txt", ios::out | ios::app);
        output_file << k << " "; 
        output_file.close();
    
        // Check the accuracy of the svm using the text vectors
        MKL_classify(&instance, 100.0, betas); 
    }
        
    // Close shogun environment
    exit_shogun();
    
    return 0;
}

