/******************************************************************************/
/* SMPS READER FOR TWO-STAGE STOCHASTIC LINEAR PROGRAMMING PROBLEMS           */
/* Author: Pedro Munari [pedro.munari@gmail.com]                              */
/* Last modified: JAN/2013                                                    */
/******************************************************************************/

/******************************************************************************/
/* ATTENTION: This reader supports two-stage stochastic problems only. The    */
/*            .STO file must be described by INDEP and/or BLOCK sections      */ 
/******************************************************************************/

/******************************************************************************/
/* SMPS DATA STRUCTURE                                                        */
/******************************************************************************/

#define MAX_LINE_SIZE 64
#define PROBLEM_NAME_SIZE 47

#define TYPE_FIELD_SIZE 2
#define NAME_FIELD_SIZE 8
#define VALUE_FIELD_SIZE 12

#define MAX_LINE_SIZE_EXT 1000
#define PROBLEM_NAME_SIZE_EXT 1000

#define TYPE_FIELD_SIZE_EXT 2
#define NAME_FIELD_SIZE_EXT 100
#define VALUE_FIELD_SIZE_EXT 100

#define FIELD_1_START 2
#define FIELD_2_START 5
#define FIELD_3_START 15
#define FIELD_4_START 25
#define FIELD_5_START 40
#define FIELD_6_START 50

#define TOL_ZERO 1E-10

typedef char T_MPS_name[101];

/* Type to classify the domain of a variable */
typedef enum{REA, INT, BIN} T_domain;
/* REA -> Real */
/* INT -> Integer */
/* BIN -> Binary */

/* Type to classify a variable */
typedef enum{FX, BD, PL, FR, MI, LO, UP} T_variable_type;
/* FX -> Fixed (type 0) */
/* BD -> Bounded (type 1) */
/* PL -> Plus or Nonnegative (type 2) */
/* FR -> Free (type 3) */
/* MI -> Minus or Nonpositive (x_j <= 0) */
/* LO -> Lower bound ( l_j <= x_j <= +\infty ) */
/* UP -> Upper bound (   0 <= x_j <= u_j     ) */

/* Type to classify a constraint */
typedef enum{G, L, E, N} T_constraint_type;
/* G -> ax >= b */
/* L -> ax <= b */
/* E -> ax == b */
/* N -> non-constraining row */

/* Type to define variables of options (selection) [range: 0 -> 255] */
typedef unsigned char T_option;

/* Data structure for the vector of variables in the MPS data */
typedef struct
{
    T_MPS_name name;			/* name of the variable specified in the MPS file */
    T_variable_type type;       /* variable type (FiXed, BounDed, PLus, ...) */
    T_domain domain;            /* the domain whose the variable belongs (REAl, INTeger, BINary) */
    double lower_bound;         /* the lower bound of the variable */
    double upper_bound;         /* the upper bound of the variable */
}T_MPS_variable;

/* Data structure for the vector of variables in the MPS data */
typedef struct
{
    T_MPS_name name;	  		/* name of the variable specified in the MPS file */
    T_constraint_type type;     /* variable type (FiXed, BounDed, PLus, ...) */
}T_MPS_constraint;

/* Type of a node in the list of rows or columns */
typedef struct _node
{
    int position;
    T_MPS_name name;
    int type;
    struct _node* next;
}T_MPS_node;

/* Type that defines a list with a head node and a counter */
typedef struct list
{
    T_MPS_node *head;
    int counter;
}T_MPS_list;

/* Data structure for the vector of periods in the TIME file */
typedef struct
{
    int *first_variable_period;   /* index of the first column in period i */
    int *first_constraint_period; /* index of the first row in period i */
    
}T_SMPS_periods;

/* Data structure for entries (variable, constraint, value) in the BLOCk structure */
typedef struct
{
    int i_variable;   /* index of the variable in the array MPS_variables */
    int i_constraint; /* index of the constraint in the array MPS_constraint */
    double value;     /* value of the entry */
}
T_SMPS_BLOCK_values;

/* Data structure for the array of all blocks in the STOCHASTIC file */
typedef struct
{
    double probability;  /* probability associated to the current block */    

    int n_entries; /* number of entries in the block */
    int first_entry; /* position of the first entry of the block in the array of entries */

}T_SMPS_BLOCK_description;

/* Data structure for the vector of blocks in the STOCHASTIC file */
typedef struct
{
    T_MPS_name name;  /* name of the block */
    int period;   /* number of the period which contains the block */
    
    int first_description; /* position of the first description of the block */
    int n_description; /* number of descriptions of the block */
    
}T_SMPS_BLOCK;

/* Data structure for entries (variable, constraint, value) in the INDEP structure */
typedef struct
{
    double value; /* value of the entry */
    double probability; /* probability of the value */
}
T_SMPS_INDEP_values;

/* Data structure for the vector of INDEPENDENT distribution in the STOCHASTIC file */
typedef struct
{
    int i_variable;   /* index of the variable in the array MPS_variables */
    int i_constraint; /* index of the constraint in the array MPS_constraint */
    
    int first_entry; /* position of the first description of the block */
    int n_entries;   /* number of descriptions of the block */
    
}T_SMPS_INDEP;

/* Data structure for sparse matrices (Rowwise and Columnwise together) */
typedef struct
{
    int    nr;          /* number of rows in the matrix */
    int    nc;          /* number of colums in the matrix */
    int    nc_max;      /* maximum number of colums in the matrix */
    int    nnz;         /* number of nonzero entries in the matrix */
    int    max_size;    /* number of positions allocated to store the non-zero entries */
    
    /* Number of nonzero entries */
    int    *col_len;     /* number of nonzero entries in a column */
    int    *row_len;     /* number of nonzero entries in a row */

    /* Colum-wise structure */
    int    *C_col_ind;    /* vector of column index for each nonzero entry */
    int    *C_row_ind;    /* vector of row index for each nonzero entry */
    double *C_value;      /* nonzero entries */
    
    /* Row-wise structure */
    int    *R_row_ind;    /* vector of row index for each nonzero entry */ 
    int    *R_col_ind;    /* vector of column index for each nonzero entry */
    double *R_value;      /* nonzero entries */
    
}SMatrix_type;

/* Data structure for sparse matrices of stochastic problems */
typedef struct
{
    int nr;      /* number of rows in the matrix */
    int nr_max;  /* maximum number of rows in the matrix */
    int nc;      /* number of colums in the matrix */
    int nc_max;  /* maximum number of colums in the matrix */
    int nnz;     /* number of nonzero entries in the matrix */
    int nnz_max; /* number of positions allocated to store the non-zero entries */
    int nS;      /* number of scenarios */  
    
    /* Number of nonzero entries */
    int    *col_len;     /* number of nonzero entries in each column */
    int    *row_len;     /* number of nonzero entries in each row */

    /* Colum-wise structure */
    int    *C_col_ind;    /* vector of column index for each nonzero entry */
    int    *C_row_ind;    /* vector of row index for each nonzero entry */
    double *C_value;  /* nonzero entries of the EVP */
    
    /* Row-wise structure */
    int    *R_row_ind;    /* vector of row index for each nonzero entry */ 
    int    *R_col_ind;    /* vector of column index for each nonzero entry */
    double *R_value;  /* nonzero entries of the EVP */

    /* Stochastic values */
    int n_STO;
    int n_STO_max;
    
    int    *STO_ind; /* row/column index (rowwise/columnwise repr) of the stochastic coefficient */
     
    int    *STO_C_pos;    /* position of the corresponding original value in C_value_EVP */
    double **STO_C_value; /* stochastic value associated to a coefficient in C_value_EVP */
                          /* nS arrays: one for each scenario */
                          
    int    *STO_R_pos;    /* position of the corresponding original value in R_value_EVP */
    double **STO_R_value; /* stochastic value associated to a coefficient in R_value_EVP */
                          /* nS arrays: one for each scenario */
    
}STO_SMatrix_type;

/* Data structure for sparse matrices of stochastic problems */
typedef struct
{
    int nr;      /* number of entries in the vector */
    int nnz;     /* number of nonzero entries in the vector */
    int nnz_max; /* number of positions allocated to store the non-zero entries */
    int nS;      /* number of scenarios */  
    
    int    *index;      /* vector of index for each nonzero entry */
    double *value;  /* nonzero entries of the EVP */
    
    /* Stochastic values */
    int n_STO;
    int n_STO_max;
    
    int    *STO_pos;    /* position of the corresponding original value in value_EVP */
    double **STO_value; /* stochastic value associated to a coefficient in value_EVP */
                        /* nS arrays: one for each scenario */
    
}STO_SVector_type;

/* Data structure that encapsulates all the information regarding the instance */
typedef struct
{
    /* CORE file */
    SMatrix_type *A; /* coefficient matrix in the sparse data structure */
	double *c;       /* vector of costs */
	double *RHS;     /* vector of RHS values */
	double *RANGES;  /* vector of ranges on constraints */
    T_MPS_constraint *MPS_constraints; /* Information about the constraints in the CORE file */
    T_MPS_variable   *MPS_variables;   /* Information about the variables in the CORE file */
    
    /* TIME file */
    T_SMPS_periods* SMPS_periods; /* Information about the periods in the TIME file */
    
    /* STOCHASTIC file */
    T_SMPS_BLOCK *SMPS_block; /* Array of blocks in a STOCHASTIC file */
    T_SMPS_BLOCK_description *SMPS_block_description; /* Description of the blocks in the BLOCK structure */
    T_SMPS_BLOCK_values      *SMPS_block_values;      /* Entries of the blocks in the BLOCK structure */

    T_SMPS_INDEP *SMPS_indep; /* Array of independent distribution in a STOCHASTIC file */
    T_SMPS_INDEP_values *SMPS_indep_values; /* Values associated to each entry in the INDEP structure */
    
    /* Dimensions */
    int nr; /* number of rows in the CORE problem */
    int nc; /* number of columns in the CORE problem */
    int nI; /* number of independent random variables in the STOCHASTIC file */
    int nB; /* number of blocks in the STOCHASTIC file */
    int nS; /* number of scenarios in the problem */
    
}T_SMPS_DATA;

/******************************************************************************/
/* READ AN INSTANCE FROM THREE FILES: .COR, .TIM AND .STO                     */
/******************************************************************************/
int readSMPS(
    char *mps_file_name,        /* ( ->) CORE data file in the MPS format */
    char *TIME_file_name,       /* ( ->) TIME data file */
    char *STO_file_name,        /* ( ->) STOCHASTIC data file */
    T_SMPS_DATA* SMPS_data);     /* (<- ) pointer to the instance data */

/* Initializes an empty sparse matrix in Rowwise format */
int SMatrix_initMatrix_RW(SMatrix_type *S, int nr, int nc, int max_nnz);

/* Initializes an empty sparse matrix in Columnwise format */
int SMatrix_initMatrix_CW(SMatrix_type *S, int nr, int nc, int max_nnz);

