/******************************************************************************/
/* SMPS READER FOR TWO-STAGE STOCHASTIC LINEAR PROGRAMMING PROBLEMS           */
/* Author: Pedro Munari [pedro.munari@gmail.com]                              */
/* Last modified: JAN/2013                                                    */
/******************************************************************************/
/* LIBRARY TO STORE AND HANDLE SPARSE MATRICES                                */
/******************************************************************************/

/* Initializes an empty sparse matrix in Rowwise format */
int SMatrix_initMatrix_RW(SMatrix_type *S, int nr, int nc, int max_nnz);

/* Initializes an empty sparse matrix in Columnwise format */
int SMatrix_initMatrix_CW(SMatrix_type *S, int nr, int nc, int max_nnz);

/* Reallocates the sparse matrix with more available columns */
int SMatrix_reallocateColumns_CW(SMatrix_type *S);

/* Reallocates the sparse matrix with more available nonzero entries */
int SMatrix_reallocateArrays_CW(SMatrix_type *S);

/* Reallocates the vectors to drop blank spaces */
int SMatrix_fitArrays_CW(SMatrix_type *S);

/* Reallocates the vectors to drop empty columns */
int SMatrix_fitColumns_CW(SMatrix_type *S, int nc);

/* Create a Rowwise representation from the columnwise one */
int SMatrix_createRWfromCW(SMatrix_type *S);

/* Computes the inner product of the vector v and a column of the sparse matrix S */
void SMatrix_innerProductColumn(
     double *result,   /* (<- ) result of the inner product to be computed */
     int index,        /* ( ->) index of the column of S */
     SMatrix_type *S,  /* ( ->) sparse matrix in the column compression format */
     double *v);        /* ( ->) vector to compute the inner product */

/* Scatter a column from the sparse matrix to the vector v */
void SMatrix_scatterColumn(
     double *v,         /* ( ->) vector to receive the column scattered */
     int index,         /* ( ->) index of the column of S to be scattered */
     SMatrix_type *S); /* ( ->) sparse matrix in the column compression format */

/* Scatter the negative of a column from the sparse matrix to the vector v */
void SMatrix_scatterColumnNeg_flags(
     double *v,        /* (<->) vector to receive the column scattered */
     T_option *v_flag, /* ( ->) flag for nonzeros in the v array */
     int *v_indices,   /* (<->) indices of nonzeros in the v array */
     int *n_v,         /* (<->) number of nonzeros in the v array */
     int index,        /* ( ->) index of the column of S to be scattered */
     SMatrix_type *S);/* ( ->) sparse matrix in the column compression format */

/* Scatter the negative of a column from the sparse matrix to the vector v */
void SMatrix_sumColumnTimesScalar(
     double scalar,    /* ( ->) a number that multiplies the column */
     double *v,        /* (<->) vector to receive the column scattered */
     T_option *v_flag, /* ( ->) flag for nonzeros in the v array */
     int *v_indices,   /* (<->) indices of nonzeros in the v array */
     int *n_v,         /* (<->) number of nonzeros in the v array */
     int index,        /* ( ->) index of the column of S to be scattered */
     SMatrix_type *S);  /* ( ->) sparse matrix in the column compression format */

/* Computes the inner product of the vector v and a row from the sparse matrix S */
void SMatrix_innerProductRow(
     double *result,   /* (<- ) result of the inner product to be computed */
     int index,        /* ( ->) index of the row from S */
     SMatrix_type *S,  /* ( ->) sparse matrix in the column compression format */
     double *v);        /* ( ->) vector to compute the inner product */

void SMatrix_scatterRowNeg_flags(
     double *v,        /* (<->) vector to receive the row scattered */
     T_option *v_flag, /* ( ->) flag for nonzeros in the v array */
     int *v_indices,   /* (<->) indices of nonzeros in the v array */
     int *n_v,         /* (<->) number of nonzeros in the v array */
     int index,        /* ( ->) index of the row of S to be scattered */
     SMatrix_type *S);  /* ( ->) sparse matrix in the column compression format */

/* Free the memory used by a sparse matrix */
void SMatrix_freeMatrix(SMatrix_type *S);

/* Initializes an empty stochastic sparse matrix in Columnwise format */
int STO_SMatrix_initMatrix_CW(STO_SMatrix_type *S, int nr, int nc, int nnz, int nS);

/* Initializes an empty stochastic sparse matrix in Columnwise format */
int STO_SMatrix_initMatrix_RW(STO_SMatrix_type *S, int nr, int nc, int nnz, int nS);

/* Initialize the arrays of scenarios coefficients */
int STO_SMatrix_initScenarios_CW(STO_SMatrix_type *S, int nS);

/* Initialize the arrays of scenarios coefficients */
int STO_SMatrix_initScenarios_RW(STO_SMatrix_type *S, int nS);

/* Set a new stochastic coefficient in the vector */
int STO_SMatrix_addStochasticCoeff_RW(
    STO_SMatrix_type *S, /* stochastic sparse matrix */
    int s,         /* number of the scenario which will receive the coefficient */
    int row_ind,   /* row index of the coefficient in the matrix */
    int col_ind,   /* column index of the coefficient in the matrix */
    double value); /* coefficient value */

/* Free the memory used by a sparse matrix */
void STO_SMatrix_freeMatrix(STO_SMatrix_type *S);

/* Initializes an empty stochastic sparse vector */
int STO_SVector_initVector(STO_SVector_type *v, int nr, int nnz, int nS);

/* Initialize the arrays of scenarios coefficients */
int STO_SVector_initScenarios(STO_SVector_type *v, int nS);

/* Set a new stochastic coefficient in the vector */
int STO_SVector_addStochasticCoeff(
    STO_SVector_type *v, /* stochastic sparse vector */
    int s,        /* number of the scenario which will receive the coefficient */
    int index,    /* index of the coefficient in the vector */
    double value); /* coefficient value */

/* Free the memory used by the sparse vector */
void STO_SVector_freeVector(STO_SVector_type *v);

/* Create a Rowwise representation from the Columnwise */
int STO_SMatrix_createRWfromCW(STO_SMatrix_type *S, int nS);

/* Create a Columnwise representation from the Rowwise */
int STO_SMatrix_createCWfromRW(STO_SMatrix_type *S, int nS);
