/*
 *  COLUMN GENERATION PROCEDURE FOR SOLVING
 *  TWO-STAGE STOCHASTIC PROGRAMMING PROBLEMS (TSSP)
 *  [Aggregated, k-best columns]
 *
 *  Authors: Pablo Gonzalez-Brevis (p.gonzalez-brevis@sms.ed.ac.uk)
 *           Pedro Munari (munari@dep.ufscar.br)
 *
 *  Last modified: Jul/2013
 *
 *  ON TOP OF HOPDM/PDCGM
 *  Written by:    Jacek Gondzio
 *                 Department of Mathematics and Statistics,
 *                 The University of Edinburgh,
 *                 Mayfield Road, King's Buildings.
 *                 EH9 3JZ Edinburgh, Scotland.
 *                 on leave from: Systems Research Institute,
 *                 Polish Academy of Sciences,
 *                 Newelska 6, 01-447 Warsaw, Poland.
 */
 
/* Include libraries and auxiliar files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define SAVE_TO_FILE

/* Optimality tolerance */
#define TSSP_DELTA 1.E-5

/* Degree of optimality D (must be > 1) */
#define TSSP_DEGOPT 5.0 // 1.1

/* Bounds on dual variables */
#define BOUND_DUAL 1.E+5

/* Scaling factor for the convexity constraints */
#define CNVXTY_VALUE 100.0

/* Scaling factor for the convexity constraints */
double OBJ_SCALING = 1.0;

/* Scaling factor for the RHS */
double RHS_SCALING = 1.0;

/* Bring in the SMPS reader files */
#include "smps_reader.h"
#include "sparse_matrix.h"

/* Bring in the PDCGM/HOPDM library files */
#include "../../../pdcgm/pdcgm.h"

/* Bring in the CPLEX library files */
#include <ilcplex/cplex.h>

/* Data structure for the instance data */
typedef struct
{
    T_SMPS_DATA  *SMPS_data; /* instance data in the SMPS format */
    
    int nr_sub,  /* number of rows in the coefficient matrix of the subproblem */
        nc_sub,  /* number of columns in the coefficient matrix of the subproblem */
        nnz_sub; /* number of nonzeroes in the coefficient matrix of the subproblem */
    
    /* Stochastic data */
    double *scenario_probability; /* The probability of each scenario */
    
    STO_SVector_type *h; /* RHS of each second stage row, for each scenario 
                            These are the coefficient costs of the Benders subproblems */
                         
    STO_SVector_type *q; /* Costs of each second stage variable, for each scenario 
                            These are the RHS values of the Benders subproblems */
    
    STO_SMatrix_type *T; /* Second stage matrices corresponding to first stage variables,
                            for each scenario (ROWWISE and COLUMNWISE REPRESENTATION) */

    STO_SMatrix_type *Wt; /* Second stage matrices corresponding to second stage variables,
                             for each scenario (ROWWISE and COLUMNWISE REPRESENTATION) 
                             These are the coefficient matrices of the Benders subproblems */

    /* Auxiliary dense vectors */
    char *row_type; /* type of each constraint of the subproblem */
    
    double *cost,  /* objective costs of the subproblem */  
           *RHS,   /* right hand side of the contraints the subproblem */   
           *low_b,
           *upp_b; /* lower and upper bounds of the variables of the subproblem */
           
    /* CPLEX variables */
    CPXENVptr    *env_sub; /* CPLEX environment for the subproblem of each scenario */ 
    CPXLPptr     *lp_sub;  /* CPLEX environment for the subproblem of each scenario */
}  
T_instance_data;

/*---------------------------------------------------------------------------*/
/* Increment the less significative counter (and apply carry-over if needed) */
/*---------------------------------------------------------------------------*/
int TSSP_increment_array_counter(
    int *array_counter_ind, 
    int *array_counter_len, 
    int array_counter_n);
            
/*--------------------------------------------------------*/
/* SOLVE THE DETERMINISTIC EQUIVALENT PROBLEM USING HOPDM */
/*--------------------------------------------------------*/
int TSSP_solve_DEP(T_SMPS_DATA* SMPS_data);

/*--------------------------------------------------------*/
/* SOLVE THE DECOMPOSED PROBLEM USING PDCGM               */
/*--------------------------------------------------------*/
int TSSP_solve_decomposed(int aggregated, T_SMPS_DATA* SMPS_data);

/*------------------------------------------------------------*/
/* SOLVE THE CORE FORMULATION (WITHOUT SCENARIOS) USING HOPDM */
/*------------------------------------------------------------*/
int TSSP_solve_EVP(double *optimal_value, double *x,T_instance_data* instance);

