/*
 * AUXILIARY FUNCTIONS TO SET AND SOLVE A PROBLEM ON TOP OF CPLEX 
 * THE PROBLEM CORRESPONDS TO THE BENDERS SUBPROBLEM OF A COLUMN GENERATION
 * SCHEME TO SOLVE TWO-STAGE STOCHASTIC PROBLEMS
 *
 * Author: Pedro Munari (munari@dep.ufscar.br)
 *         UFSCar - Federal University of Sao Carlos, Brazil
 *
 * Last modified: Feb/2013
 *
 */

/* SETUP OF THE CPLEX VARIABLES */
int TSSP_CPLEX_setup_vars(
    CPXENVptr *p_env,   /* pointer to the CPLEX environment */
    CPXLPptr  *p_lp,    /* pointer to the problem in the internal structure */
    int obj_sense)      /* 1 -> maximization, -1 -> minimization */
{

    /* local CPLEX vars */
    int status = 0;
    CPXENVptr env = NULL;
    CPXLPptr  lp  = NULL;

    /* Initialize the CPLEX environment */
    env = CPXopenCPLEX (&status);
    /* If an error occurs, the status value indicates the reason for
       failure.  A call to CPXgeterrorstring will produce the text of
       the error message.  Note that CPXopenCPLEX produces no output,
       so the only way to see the cause of the error is to use
       CPXgeterrorstring.  For other CPLEX routines, the errors will
       be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON.  */
    if ( env == NULL ) 
    {
        char  errmsg[1024];
        fprintf (stderr, "Could not open CPLEX environment.\n");
        CPXgeterrorstring (env, status, errmsg);
        fprintf (stderr, "%s", errmsg);
        return -1;
    }

    /* Turn on output to the screen */
    status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_OFF);
    if ( status ) 
    {
        fprintf (stderr, "Failure to turn on screen indicator, error %d.\n", status);
        return -1;
    }

    /* Turn on data checking */
    status = CPXsetintparam (env, CPX_PARAM_DATACHECK, CPX_ON);
    if ( status ) 
    {
        fprintf (stderr, "Failure to turn on data checking, error %d.\n", status);
        return -1;
    }

    /* Create the problem */
    lp = CPXcreateprob (env, &status, "PROBLEM");

    /* A returned pointer of NULL may mean that not enough memory
       was available or there was some other problem.  In the case of 
       failure, an error message will have been written to the error 
       channel from inside CPLEX.  Here, the setting of
       the parameter CPX_PARAM_SCRIND causes the error message to
       appear on stdout.  */
    if(lp == NULL)
    {
        fprintf (stderr, "Failed to create LP.\n");
        return -1;
    }

    /* Set optimization direction according to the parameter obj_sense */
    if(obj_sense == 1) CPXchgobjsen (env, lp, CPX_MAX);
    else if (obj_sense == -1) CPXchgobjsen (env, lp, CPX_MIN);
    else
    {
        fprintf (stderr, "Failed to create LP (objective sense).\n");
        return -1;
    }
    
    /* Set the parameters to return */
    *p_env = env;
    *p_lp  = lp;
    
    /* Return successfully */
    return 1;
}

/* FREE UP THE CPLEX VARS ALLOCATED BY BENDERS_CP_setup_vars() */
int TSSP_CPLEX_free_vars(
    CPXENVptr *p_env,   /* pointer to the CPLEX environment */
    CPXLPptr  *p_lp)    /* pointer to the problem in the internal structure */
{

    int status = 0;
    
    /* Free up the problem as allocated by CPXcreateprob, if necessary */
    if ( (p_lp != NULL ) && (*p_lp != NULL ) )
    {
       status = CPXfreeprob (*p_env, p_lp);
       if( status ) fprintf (stderr, "CPXfreeprob failed, error code %d.\n", status);
    }

    /* Free up the CPLEX environment, if necessary */
    if ( ( p_env != NULL ) && ( *p_env != NULL ) )
    {
        status = CPXcloseCPLEX (p_env);

        /* Note that CPXcloseCPLEX produces no output,
           so the only way to see the cause of the error is to use
           CPXgeterrorstring.  For other CPLEX routines, the errors will
           be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON. */
        if ( status ) 
        {
           char  errmsg[1024];
           fprintf (stderr, "Could not close CPLEX environment.\n");
           CPXgeterrorstring (*p_env, status, errmsg);
           fprintf (stderr, "%s", errmsg);
        }
    }
    
    return 1;
}

/* SET THE *DUAL* FORMULATION OF THE BENDERS SUBPROBLEM */
int TSSP_set_subproblem(
    T_instance_data *instance)  /* instance data */
{ 
    if( (instance == NULL) || (instance->SMPS_data == NULL) ) return 0;
    if( (instance->h == NULL) || (instance->T == NULL) ) return 0;
    
    T_SMPS_DATA *SMPS_data = instance->SMPS_data; /* Alias */
    
    int status = 0, /* auxiliaries */
        nI = SMPS_data->nI, /* number of stochastic independent variables */
        nB = SMPS_data->nB, /* number of stochastic blocks */
        nS = SMPS_data->nS; /* number of scenarios */
    
    int h, i, j, k, r, s, t, cbeg=0, cend=0,
        n_STO_q, n_STO_h, n_STO_Wt, n_STO_T,
        nr = SMPS_data->nr,
        nc = SMPS_data->nc,
        nr_p1  = (SMPS_data->SMPS_periods)->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
        nc_p1  = (SMPS_data->SMPS_periods)->first_variable_period[1],   /* number of variables in the first stage */
        nr_p2  = nr - nr_p1, /* number of constraints in the second stage */
        nc_p2  = nc - nc_p1, /* number of variables in the second stage */
        nr_sub  = nc_p2,     /* number of constraints in the SUBPROBLEM (dual formulation) */
        nc_sub  = nr_p2,     /* number of variables in the SUBPROBLEM (dual formulation) */
        nnz_sub = (SMPS_data->A)->nnz + nc_sub; /* a GUESS for the number of nonzeroes in the coefficient matrix of the SUBPROBLEM */
    
    /* Auxiliary arrays that are used to generate scenarios */
    int array_counter_n=0, *array_counter_ind=NULL, *array_counter_len=NULL;
    
    instance->nr_sub = nr_sub;
    instance->nc_sub = nc_sub;
    instance->nnz_sub = nnz_sub;
    
    /* Memory allocation */
    instance->low_b = (double *) malloc( sizeof(double) * nc_sub );
    if(instance->low_b == NULL) return -1;
    
    instance->upp_b = (double *) malloc( sizeof(double) * nc_sub );
    if(instance->upp_b == NULL) return -1;
    
    instance->cost = (double *) malloc( sizeof(double) * nc_sub );
    if(instance->cost == NULL) return -1;
    
    instance->RHS = (double *) malloc( sizeof(double) * nr_sub );
    if(instance->RHS == NULL) return -1;
    
    instance->row_type = (char *)  malloc( sizeof(char) * nr_sub );
    if(instance->row_type == NULL) return -1;
    
    if( STO_SVector_initVector(instance->h, nc_sub, nc_sub, nS) != 1 ) return -1;
    (instance->h)->nr = 0;
    (instance->h)->nnz = 0;
    (instance->h)->n_STO = 0;
    
    if( STO_SVector_initVector(instance->q, nr_sub, nr_sub, nS) != 1 ) return -1;
    (instance->q)->nr = 0;
    (instance->q)->nnz = 0;
    (instance->q)->n_STO = 0;
    
    if( STO_SMatrix_initMatrix_CW(instance->T, nr_p2, nc_p1, nnz_sub, nS) != 1 ) return -1;
    (instance->T)->nr = 0;
    (instance->T)->nc = 0;
    (instance->T)->nnz = 0;
    (instance->T)->n_STO = 0;
    
    if( STO_SMatrix_initMatrix_RW(instance->Wt, nr_sub, nc_sub, nnz_sub, nS) != 1 ) return -1;
    (instance->Wt)->nr = 0;
    (instance->Wt)->nc = 0;
    (instance->Wt)->nnz = 0;
    (instance->Wt)->n_STO = 0;
    
    /* Set the type of each variable in the subproblem (dual formulation) */
    for(i = 0; i < nc_sub; i++) 
    {
        if(SMPS_data->MPS_constraints[nr_p1 + i].type == E)
        {
            /* The dual variable must be FREE */
            instance->low_b[i] = -CPX_INFBOUND;
            instance->upp_b[i] = CPX_INFBOUND;
        }
        else if(SMPS_data->MPS_constraints[nr_p1 + i].type == G)
        {
            /* The dual variable must be NON-NEGATIVE */
            instance->low_b[i] = -CPX_INFBOUND;
            instance->upp_b[i] = 0.0;
        }
        else if(SMPS_data->MPS_constraints[nr_p1 + i].type == L)
        {
            /* The dual variable must be NON-POSITIVE */
            instance->low_b[i] = 0.0;
            instance->upp_b[i] = CPX_INFBOUND;
        }
        else
        {
            /* The dual variable must be FIXED */
            instance->low_b[i] = 0.0;
            instance->upp_b[i] = 0.0;
        }
    }
    
     /* Set the the costs of the subproblem (dual formulation) 
        -- the same as the RHS of the second stage constraints */
    for(i = 0; i < nc_sub; i++)
    {
        if( fabs(SMPS_data->RHS[nr_p1 + i]) > TOL_ZERO )
        {
            (instance->h)->index[ (instance->h)->nnz ] = i;
            (instance->h)->value[ (instance->h)->nnz ] = SMPS_data->RHS[nr_p1 + i];
            ((instance->h)->nnz)++;
        }
    }
    (instance->h)->nr = nc_sub;
    
    /* SET THE AUXILIARY MATRIX T (COLUMNWISE SPARSE REPRESENTATION) */
    /* Coefficient matrix corresponding to the first stage variables on second stage constraints */
    k = 0;
    cbeg = 0;
    cend = 0;
    
    /* Set T using the coefficients of the first stage columns which are in second stage rows */
    for(i = 0; i < nc_p1; i++) 
    {
        /* Set the initial position of the new column */
        (instance->T)->C_col_ind[i] = k;
        
        /* Get the initial position of the CORE column */
        cbeg = (SMPS_data->A)->C_col_ind[i];
        cend = cbeg + (SMPS_data->A)->col_len[i];
        
        /* Copy the second stage coefficients only */
        for(j = cbeg; j < cend; j++)
        {
            /* Check if it is an entry in a second stage row */
            if((SMPS_data->A)->C_row_ind[j] >= nr_p1)
            {
                (instance->T)->C_row_ind[k] = (SMPS_data->A)->C_row_ind[j] - nr_p1;
                (instance->T)->C_value[k]   = (SMPS_data->A)->C_value[j];
                
                /* Increase the number of entries in the corresponding column and row */
                (instance->T)->col_len[i]++;
                (instance->T)->row_len[ (instance->T)->C_row_ind[k] ]++;
                
                /* Increase the number of nonzeros */
                k++;
            }
        }        
    }
    /* Set the final dimensions of the matrix */
    (instance->T)->nr = nr_p2;
    (instance->T)->nc = nc_p1;
    (instance->T)->nnz = k;
    
    /* Create the rowwise representation of T in order to improve the 
       performance on future computations */
    STO_SMatrix_createRWfromCW(instance->T, nS);
    
    /* Set the type of the subproblem constraints */
    for(i = 0; i < nr_sub; i++) 
    {
        if(SMPS_data->MPS_variables[nc_p1 + i].type == PL)
        {
            /* 'GREATER-THAN-OR-EQUAL-TO' CONSTRAINT */
            instance->row_type[i] = 'G';
        }
        else if(SMPS_data->MPS_variables[nc_p1 + i].type == MI)
        {
            /* 'LESS-THAN-OR-EQUAL-TO' CONSTRAINT */
            instance->row_type[i] = 'L';
        }
        else if(SMPS_data->MPS_variables[nc_p1 + i].type == FR)
        {
            /* EQUALITY CONSTRAINT */
            instance->row_type[i] = 'E';
        }
        else 
        {
            /* THE CURRENT VERSION DOES NOT SUPPORT BOUNDED VARIABLES */
            printf("\nSorry, the current version does not support bounded variables.\n");
            return -2;
        }
    }
    
    /* Set the RHS of the second stage constraints */
    for(i = 0; i < nr_sub; i++)
    {
        if( fabs(SMPS_data->c[nc_p1 + i]) > TOL_ZERO )
        {
            (instance->q)->index[ (instance->q)->nnz ] = i;
            (instance->q)->value[ (instance->q)->nnz ] = -SMPS_data->c[nc_p1 + i];
            ((instance->q)->nnz)++;
        }
    }
    (instance->q)->nr = nr_sub;
   
    /* SET THE COEFFICIENT MATRIX OF THE SUBPROBLEM USING A ROWWISE STOCHASTIC SPARSE REPRESENTATION */
    
    /* Set the rows of W^T using the corresponding secong stage columns */
    k=0;
    for(i = 0; i < nr_sub; i++) 
    {
        /* Set the initial position of the new row */
        (instance->Wt)->R_row_ind[i] = k;
        
        /* Get the initial position of the corresponding column */
        cbeg = (SMPS_data->A)->C_col_ind[nc_p1 + i];
        cend = cbeg + (SMPS_data->A)->col_len[nc_p1 + i];
        
        /* Copy the second stage coefficients only */
        for(j = cbeg; j < cend; j++) 
        {
            /* Check if t is an entry in a second stage row */
            if( ((SMPS_data->A)->C_row_ind[j] >= nr_p1) && ((SMPS_data->A)->C_row_ind[j] < nr) ) 
            {
                (instance->Wt)->R_col_ind[k]   = (SMPS_data->A)->C_row_ind[j] - nr_p1;
                (instance->Wt)->R_value[k] = (SMPS_data->A)->C_value[j];
                
                /* Increase the number of entries in the corresponding column and row */
                ((instance->Wt)->row_len[i])++;
                ((instance->Wt)->col_len[ (instance->Wt)->R_col_ind[k] ])++;
                
                /* Increase the number of nonzeros */
                k++;
            }
        }        
    }
    (instance->Wt)->nr = nr_sub;
    (instance->Wt)->nc = nc_sub;
    (instance->Wt)->nnz = k;
    instance->nnz_sub = k;
    
    /* Set the problem on the CPLEX environment */
    
    /* State and allocate auxiliary arrays that will help setting the stochastic information */
    if(nI + nB > 0)
    {
        /* Array of counters: one counter for each BLOCK/INDEP */
        array_counter_n = nI + nB;
        array_counter_ind = (int *) malloc( sizeof(int) * array_counter_n );
        
        /* Array to store the length of each BLOCK/INDEP */
        array_counter_len = (int *) malloc( sizeof(int) * array_counter_n );
        
        /* Set the length of each BLOCK/INDEP */    
        for(i = 0; i < nI; i++) array_counter_len[i] = (SMPS_data->SMPS_indep[i]).n_entries;
        for(i = 0; i < nB; i++) array_counter_len[nI + i] = (SMPS_data->SMPS_block[i]).n_description;
        
        /* Initialize the counters */
        for(i = 0; i < array_counter_n; i++) array_counter_ind[i] = 0;
        
    }
    
    /* Modify the second stage coefficients according to the scenarios */
    s = 0;
    do
    {
        /* Modify the data according to the SMPS data */
        instance->scenario_probability[s] = 1.0;
        n_STO_q  = 0;
        n_STO_h  = 0;
        n_STO_Wt = 0;
        n_STO_T  = 0;
        
        /* Check if the SMPS data about the scenarios is available */
        if(nI + nB > 0)
        {
            /* Set the scenario according to the current counter values */
            for(i = 0; i < array_counter_n; i++)
            {
                if(i < nI)
                {
                    /* Get the information about the i-th INDEP description */
                    
                    /* Get the corresponding row */
                    r = SMPS_data->SMPS_indep[i].i_constraint;
                    if(r == 0) r = -1;
                    else r -= nr_p1;
                    /* |--> In the SMPS data, the row 0 describes the objective function, so we have to discard 
                            this row when setting the matrix on CPLEX (it requires a separate array to describe 
                            the objective). In addition, we have to discard the first stage constraints */
                    
                    /* Get the array_counter_ind[i]-th entry from the i-th INDEP description */
                    k = SMPS_data->SMPS_indep[i].first_entry + array_counter_ind[i];
                    
                    if(SMPS_data->SMPS_indep[i].i_variable < 0) 
                    {
                        /* Changes in the RHS of the original problem 
                           => changes in the cost of the subproblem (dual formulation) */
                        STO_SVector_addStochasticCoeff(instance->h, s, r, SMPS_data->SMPS_indep_values[k].value / RHS_SCALING);
                        n_STO_h++;
                        
                    }
                    else
                    {
                        /* Check errors */
                        if(SMPS_data->SMPS_indep[i].i_variable >= nc_p1 + nc_p2) 
                        {
                            printf("*** ERROR! The column index is not valid.");
                            return -1;
                        }
                        
                        /* Get the corresponding column */
                        h = SMPS_data->SMPS_indep[i].i_variable;
                        
                        if(r < 0)
                        {
                            /* Check errors */
                            if(h < nc_p1) 
                            {
                                printf("*** ERROR! h < nc_p1 when setting RHS from INDEP.");
                                return -1;
                            }
                            
                            /* Changes in the objective function of the original problem 
                               ==> changes in the RHS of the subproblem (dual formulation) */
                             STO_SVector_addStochasticCoeff(instance->q, s, h-nc_p1, -SMPS_data->SMPS_indep_values[k].value / OBJ_SCALING);
                             n_STO_q++;
                                
                        }
                        else
                        {
                            /* Changes in the coefficient matrix */
                            
                            /* Check if we should modify T or W */
                            if(h < nc_p1)
                            {
                                /* Modify T */
                                STO_SMatrix_addStochasticCoeff_RW(instance->T, s, r, h, SMPS_data->SMPS_indep_values[k].value);
                                n_STO_T++;
                            }
                            else
                            {
                                /* Reset the column index in order to ignore the first stage columns */
                                h -= nc_p1;
                                
                                /* Modify W */
                                STO_SMatrix_addStochasticCoeff_RW(instance->Wt, s, h, r, SMPS_data->SMPS_indep_values[k].value);
                                n_STO_Wt++;
                            }
                        }
                    }
                    
                    /* Update the probability of the current scenario */
                    instance->scenario_probability[s] *= SMPS_data->SMPS_indep_values[k].probability;
                }
                else
                {
                    /* Get the array_counter_ind[i]-th entry from the (i-nI)-th BLOCK description */
                    j = SMPS_data->SMPS_block[i - nI].first_description + array_counter_ind[i];
                    
                    for(k = SMPS_data->SMPS_block_description[j].first_entry; 
                        k < SMPS_data->SMPS_block_description[j].first_entry + SMPS_data->SMPS_block_description[j].n_entries; k++)
                    {
                        /* Get the corresponding row */
                        r = SMPS_data->SMPS_block_values[k].i_constraint;
                        if(r == 0) r = -1;
                        else r -= nr_p1;
                        /* |--> In the SMPS data, the row 0 describes the objective function, so we have to discard 
                                this row when setting the matrix on CPLEX (it requires a separate array to describe 
                                the objective). In addition, we have to discard the first stage constraints */
                            
                        if(SMPS_data->SMPS_block_values[k].i_variable < 0) 
                        {
                            /* Changes in the RHS of the original problem 
                               ==> changes in the cost of the subproblem (dual formulation) */
                            STO_SVector_addStochasticCoeff(instance->h, s, r, SMPS_data->SMPS_block_values[k].value / RHS_SCALING);
                            n_STO_h++;
                        }
                        else
                        {
                            /* Check errors */
                            if(SMPS_data->SMPS_block_values[k].i_variable >= nc_p1 + nc_p2) 
                            {
                                printf("*** ERROR! The column index is not valid.");
                                fflush(stdout);
                                return -1;
                            }
                            
                            /* Get the corresponding column */
                            h = SMPS_data->SMPS_block_values[k].i_variable;
                            
                            if(r < 0)
                            {
                                /* Check errors */
                                if(h < nc_p1) 
                                {
                                    printf("*** ERROR! h < nc_p1 when setting RHS from BLOCK.");
                                    fflush(stdout);
                                    return -1;
                                }
                                
                                /* Changes in the objective function of the original problem 
                                   ==> changes in the RHS of the subproblem (dual formulation) */
                                STO_SVector_addStochasticCoeff(instance->q, s, h-nc_p1, -SMPS_data->SMPS_block_values[k].value / OBJ_SCALING);
                                n_STO_q++;
                            }
                            else
                            {
                                /* Changes in the coefficient matrix */
                                
                                /* Check if we should modify T or W */
                                if(h < nc_p1)
                                {
                                    /* Modify T */
                                    STO_SMatrix_addStochasticCoeff_RW(instance->T, s, r, h, SMPS_data->SMPS_block_values[k].value);
                                    n_STO_T++;
                                }
                                else
                                {
                                    /* Reset the column index in order to ignore the first stage columns */
                                    h -= nc_p1;
                                    
                                    /* Modify W */
                                    STO_SMatrix_addStochasticCoeff_RW(instance->Wt, s, h, r, SMPS_data->SMPS_block_values[k].value);
                                    n_STO_Wt++;
                                }
                            }
                        }
                    }
                    
                    /* Update the probability of the current scenario */
                    instance->scenario_probability[s] *= SMPS_data->SMPS_block_description[j].probability;
                }
            }
            
            /* Reset the number of stochastic coefficients in each vector and matrix */
            (instance->h)->n_STO  = n_STO_h;
            (instance->q)->n_STO  = n_STO_q;
            (instance->T)->n_STO  = n_STO_T;
            (instance->Wt)->n_STO = n_STO_Wt;
            
            /* Go to the information of the next scenario by incrementing array_counter  */
            TSSP_increment_array_counter(array_counter_ind, array_counter_len, array_counter_n);
        }
        
        /* Go to the next scenario */
        s++;
    }
    while(s < nS);

     /* Free the allocated memory */
    free(array_counter_ind);
    free(array_counter_len);

    for(i = 0; i < instance->nc_sub; i++) instance->cost[i] = 0.0;
    for(i = 0; i < instance->nr_sub; i++) instance->RHS[i] = 0.0;
    
    /* Set the RHS */
    for(t = 0; t < (instance->q)->nnz; t++)
        instance->RHS[ (instance->q)->index[t] ] = (instance->q)->value[t];
    
    /* Set the costs */
    for(t = 0; t < (instance->h)->nnz; t++)
        instance->cost[ (instance->h)->index[t] ] = (instance->h)->value[t];
    
    /* Set the CPLEX environment (minimization problem) */
    if( TSSP_CPLEX_setup_vars(&(instance->env_sub[0]), &(instance->lp_sub[0]), -1) == -1 )
    {
       /* The CPLEX vars were not allocated */
       return -1;
    }
    
    /* Set the variables on the CPLEX data structure */
    status = CPXnewcols (instance->env_sub[0], instance->lp_sub[0], instance->nc_sub, instance->cost, instance->low_b, instance->upp_b, NULL, NULL);
    if(status) return -1;
    
    /* Set the constraints on the CPLEX data structure */
    status = CPXaddrows(instance->env_sub[0], instance->lp_sub[0], 0, instance->nr_sub, instance->nnz_sub, 
                instance->RHS, instance->row_type, (instance->Wt)->R_row_ind, 
                (instance->Wt)->R_col_ind, (instance->Wt)->R_value, NULL, NULL);
    if(status) return -1;
    
    /* Return successfully */
    return 1;
}

/* Function: TSSP_solve_subproblem 
   Description: Solves the Benders subproblem by using CPLEX. In case 
   the subproblem is bounded, the parameter z is set as the optimal solution 
   (extreme point). Otherwise, z is set as an extreme ray of the subproblem.
   The function returns 1 if z corresponds to an extreme point, 2 if z corresponds 
   to an extreme ray, and <= 0 if an error occurs. */
int TSSP_solve_subproblem(
    int   s,          /* ( ->) index of the subproblem (scenario) */
    double *obj_value, /* (<- ) objective value of the subproblem (may be an extreme point or a ray) */
    double *z,         /* (<- ) solution of the subproblem (may be an extreme point or a ray) */
    double *x,         /* ( ->) current primal solution of the restricted master problem */
    T_instance_data *instance) /* ( ->) instance data */
{ 
    if( (instance == NULL) || (instance->SMPS_data == NULL) ) return 0;
    if( (instance->env_sub[0] == NULL) || (instance->lp_sub[0] == NULL) ) return 0;
    if( (s < 0) || (s >= instance->SMPS_data->nS) ) return 0;
    if( (x == NULL) || (z == NULL) ) return 0;
    
    int i, j, k, t, aux_int, rbeg, rend, /* auxiliaries */
         nr_p1  = instance->SMPS_data->SMPS_periods->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
         nr_p2  = instance->SMPS_data->nr - nr_p1, /* number of constraints in the second stage */
         n_costs;      /* maximum number of modified costs in the subproblem */
         
    int  status, solstat, /* auxiliaries */
         *int_vector=NULL; 
                          
    double aux_double, /* auxiliaries */
           *double_vector=NULL;

    /* Get the total number of columns in the subproblem */
    n_costs = CPXgetnumcols(instance->env_sub[0], instance->lp_sub[0]);
    
    /* Allocate memory */
    aux_int = n_costs;
    if(aux_int < (instance->h)->n_STO)  aux_int = (instance->h)->n_STO;
    if(aux_int < (instance->q)->n_STO)  aux_int = (instance->q)->n_STO;
    if(aux_int < (instance->Wt)->n_STO) aux_int = (instance->Wt)->n_STO;
    
    int_vector = (int *) malloc( sizeof(int) * aux_int );
    double_vector = (double *) malloc( sizeof(double) * aux_int );
    
    /* Intialize the vector */
    for(i = 0; i < n_costs; i++) instance->cost[i] = 0.0;
    
    /* Modify the RHS according to the scenario */
    if((instance->h)->n_STO > 0)
    {
        for(t = 0; t < (instance->h)->n_STO; t++)
            (instance->h)->value[ (instance->h)->STO_pos[t] ] = (instance->h)->STO_value[s][t];
    }
    /* Set the RHS of the subproblem (costs of the original problem) */
    for(t = 0; t < (instance->h)->nnz; t++)
        instance->cost[ (instance->h)->index[t] ] = (instance->h)->value[t];
        
    /* Modify the matrix T according to the scenario */
    if((instance->T)->n_STO > 0)
    {
        for(t = 0; t < (instance->T)->n_STO; t++)
            (instance->T)->R_value[ (instance->T)->STO_R_pos[t] ] = (instance->T)->STO_R_value[s][t];
    }
    
    /* Compute the new costs using the primal solution x.
       First, compute the product T x (componentwise). Then, check if the result is nonzero and 
       if so subtract from the corresponding component from h. In case the result is zero, we 
       discard it, i.e., we do not add it to the arrays of modified costs (int_vector, double_vector) */
    k = 0;
    for(i = 0; i < nr_p2; i++) 
    {
        aux_double = 0.0;
        
        /* Check if the row is not empty */
        if((instance->T)->row_len[i] > 0)
        {
            /* Get the initial position of the column */
            rbeg = (instance->T)->R_row_ind[i];
            rend = rbeg + (instance->T)->row_len[i];
            
            /* Compute the product of x by the i-th row of T */
            for(j = rbeg; j < rend; j++)
            {
                /* Multiply the coefficient by the corresponding component of x */
                if( fabs(x[ (instance->T)->R_col_ind[j] ]) > TOL_ZERO )
                    aux_double += (instance->T)->R_value[j] * x[ (instance->T)->R_col_ind[j] ];
            }
        }
        
        instance->cost[i] -= aux_double;
            
        int_vector[k] = i;
        double_vector[k] = instance->cost[i];
        k++;
    }
    
    /* Change the objective functions costs in the CPLEX environment */
    if( CPXchgobj(instance->env_sub[0], instance->lp_sub[0], k, int_vector, double_vector) ) 
    {
       printf("\n### Failed to change the costs of the subproblem.\n");
       fflush(stdout);
       return -2;
    }
    
    /* Modify the RHS according to the scenario */
    if((instance->q)->n_STO > 0)
    {
        for(t = 0; t < (instance->q)->n_STO; t++)
            int_vector[t] = (instance->q)->index[ (instance->q)->STO_pos[t] ];
           
        /* Change the RHS in the CPLEX environment */
        if( CPXchgrhs(instance->env_sub[0], instance->lp_sub[0], (instance->q)->n_STO, int_vector, (instance->q)->STO_value[s]) ) 
        {
           printf("\n### Failed to change the RHS of the subproblem.\n");
           fflush(stdout);
           return -2;
        }
    }
    
    /* Modify the coefficient matrix according to the scenario */
    if((instance->Wt)->n_STO > 0)
    {
        for(t = 0; t < (instance->Wt)->n_STO; t++)
            int_vector[t] = (instance->Wt)->R_col_ind[ (instance->Wt)->STO_R_pos[t] ];
        
        /* Change the RHS in the CPLEX environment */
        if( CPXchgcoeflist(instance->env_sub[0], instance->lp_sub[0], (instance->Wt)->n_STO, (instance->Wt)->STO_ind, int_vector, (instance->Wt)->STO_R_value[s]) ) 
        {
           printf("\n### Failed to change the RHS of the subproblem.\n");
           fflush(stdout);
           return -2;
        }
    }
    
    free(int_vector);
    free(double_vector);
        
    /* Write the subproblem in the LP format to a file */
    #ifdef COMMMMENTT
        char out_file[1000] = "";
        sprintf(out_file,"CPLEX1_%d.lp",s);
        CPXwriteprob(instance->env_sub[0], instance->lp_sub[0], out_file, NULL);
    #endif
    
    /* Choose the primal simplex method (to be able to get a RAY, if necessary) */   
    if( (status = CPXsetintparam (instance->env_sub[0], CPX_PARAM_LPMETHOD, 1)) )
    {
        printf("\n### Failed to choose the simplex algorithm.\n");
        fflush(stdout);
        return -2;
    }

    /* Turn OFF scaling (to be able to get a RAY, if necessary) */
    if( CPXsetintparam (instance->env_sub[0], CPX_PARAM_SCAIND, 1) ) 
    {
        printf("\n### Failed to turn off scaling.\n");
        fflush(stdout);
        return -2;
    }

    /* Turn OFF presolver (to be able to get a RAY, if necessary) */
    if( CPXsetintparam (instance->env_sub[0], CPX_PARAM_PREIND, 0) ) 
    {
        printf("\n### Failed to turn off the presolver.\n");
        fflush(stdout);
        return -2;
    }
    
    /* Solve the current subproblem */
    if( CPXlpopt(instance->env_sub[0], instance->lp_sub[0]) ) 
    {
       printf("\n### Failed to solve the subproblem by using CPLEX.\n");
       fflush(stdout);
       return -2;
    }
    
    /* Get solution from CPLEX */
    status = CPXsolution (instance->env_sub[0], instance->lp_sub[0], &solstat, obj_value, z, NULL, NULL, NULL);
    
    /* Check the type of the solution obtained */
    if(solstat == CPX_STAT_UNBOUNDED)
    {
        /* Unbounded problem: get a ray of the dual set */
        //printf("  --- UNbounded subproblem --> Generate ray ---  ");
        if( CPXgetray(instance->env_sub[0], instance->lp_sub[0], z) )
        {
           printf("\n### Error: Failed to get the RAY\n");
           return -1;
        }
        
        /* An extreme ray has been generated */
        return 2;
    }
    //printf("  --- Bounded subproblem --> Generate EXTREME POINT ---  ");
    
    /* An extreme point has been generated */
    return 1;
}

int TSSP_solve_subproblem_with_upperbound(
    int   s,          /* ( ->) index of the subproblem (scenario) */
    double *obj_value, /* (<- ) objective value of the subproblem (may be an extreme point or a ray) */
    double *z,         /* (<- ) solution of the subproblem (may be an extreme point or a ray) */
    double *x,         /* ( ->) current primal solution of the restricted master problem */
    T_instance_data *instance) /* ( ->) instance data */
{ 
    if( (instance == NULL) || (instance->SMPS_data == NULL) ) return 0;
    if( (instance->env_sub[0] == NULL) || (instance->lp_sub[0] == NULL) ) return 0;
    if( (s < 0) || (s >= instance->SMPS_data->nS) ) return 0;
    if( (x == NULL) || (z == NULL) ) return 0;
    
    int i, j, k, t, aux_int, rbeg, rend, /* auxiliaries */
         nr_p1  = instance->SMPS_data->SMPS_periods->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
         nr_p2  = instance->SMPS_data->nr - nr_p1, /* number of constraints in the second stage */
         n_costs;      /* maximum number of modified costs in the subproblem */
         
    int  status, solstat, /* auxiliaries */
         *int_vector=NULL; 
                          
    double aux_double, /* auxiliaries */
           *double_vector=NULL;

    /* Get the total number of columns in the subproblem */
    n_costs = CPXgetnumcols(instance->env_sub[0], instance->lp_sub[0]);
    int nrows = CPXgetnumrows(instance->env_sub[0], instance->lp_sub[0]);
    
    /* Allocate memory */
    aux_int = n_costs;
    if(aux_int < (instance->h)->n_STO)  aux_int = (instance->h)->n_STO;
    if(aux_int < (instance->q)->n_STO)  aux_int = (instance->q)->n_STO;
    if(aux_int < (instance->Wt)->n_STO) aux_int = (instance->Wt)->n_STO;
    
    int_vector = (int *) malloc( sizeof(int) * aux_int );
    double_vector = (double *) malloc( sizeof(double) * aux_int );
    
    /* Intialize the vector */
    for(i = 0; i < n_costs; i++) instance->cost[i] = 0.0;
    
    /* Modify the RHS according to the scenario */
    if((instance->h)->n_STO > 0)
    {
        for(t = 0; t < (instance->h)->n_STO; t++)
            (instance->h)->value[ (instance->h)->STO_pos[t] ] = (instance->h)->STO_value[s][t];
    }
    /* Set the RHS of the subproblem (costs of the original problem) */
    for(t = 0; t < (instance->h)->nnz; t++)
        instance->cost[ (instance->h)->index[t] ] = (instance->h)->value[t];
        
    /* Modify the matrix T according to the scenario */
    if((instance->T)->n_STO > 0)
    {
        for(t = 0; t < (instance->T)->n_STO; t++)
            (instance->T)->R_value[ (instance->T)->STO_R_pos[t] ] = (instance->T)->STO_R_value[s][t];
    }
    
    /* Compute the new costs using the primal solution x.
       First, compute the product T x (componentwise). Then, check if the result is nonzero and 
       if so subtract from the corresponding component from h. In case the result is zero, we 
       discard it, i.e., we do not add it to the arrays of modified costs (int_vector, double_vector) */
    k = 0;
    for(i = 0; i < nr_p2; i++) 
    {
        aux_double = 0.0;
        
        /* Check if the row is not empty */
        if((instance->T)->row_len[i] > 0)
        {
            /* Get the initial position of the column */
            rbeg = (instance->T)->R_row_ind[i];
            rend = rbeg + (instance->T)->row_len[i];
            
            /* Compute the product of x by the i-th row of T */
            for(j = rbeg; j < rend; j++)
            {
                /* Multiply the coefficient by the corresponding component of x */
                if( fabs(x[ (instance->T)->R_col_ind[j] ]) > TOL_ZERO )
                    aux_double += (instance->T)->R_value[j] * x[ (instance->T)->R_col_ind[j] ];
            }
        }
        
        instance->cost[i] -= aux_double;
            
        int_vector[k] = i;
        double_vector[k] = instance->cost[i];
        k++;
    }
    
    /* Change the objective functions costs in the CPLEX environment */
    if( CPXchgobj(instance->env_sub[0], instance->lp_sub[0], k, int_vector, double_vector) ) 
    {
       printf("\n### Failed to change the costs of the subproblem.\n");
       fflush(stdout);
       return -2;
    }
    
    
    {  
    
    double *nRHS = (double *) malloc(sizeof(double));
    nRHS[0] = -100000.0;
    
    char *rowt = (char *) malloc(sizeof(char));
    rowt[0] = 'G';
    
    int *rmat = (int *) malloc(sizeof(int));
    rmat[0] = 0;
    
    status = CPXaddrows(instance->env_sub[0], instance->lp_sub[0], 0, 1, k, 
                nRHS, rowt, rmat, 
                int_vector, double_vector, NULL, NULL);
    if(status) return -1;
    
    free(nRHS);
    free(rowt);
    free(rmat);
    
    }
    
    
    /* Modify the RHS according to the scenario */
    if((instance->q)->n_STO > 0)
    {
        for(t = 0; t < (instance->q)->n_STO; t++)
            int_vector[t] = (instance->q)->index[ (instance->q)->STO_pos[t] ];
           
        /* Change the RHS in the CPLEX environment */
        if( CPXchgrhs(instance->env_sub[0], instance->lp_sub[0], (instance->q)->n_STO, int_vector, (instance->q)->STO_value[s]) ) 
        {
           printf("\n### Failed to change the RHS of the subproblem.\n");
           fflush(stdout);
           return -2;
        }
    }
    
    /* Modify the coefficient matrix according to the scenario */
    if((instance->Wt)->n_STO > 0)
    {
        for(t = 0; t < (instance->Wt)->n_STO; t++)
            int_vector[t] = (instance->Wt)->R_col_ind[ (instance->Wt)->STO_R_pos[t] ];
        
        /* Change the RHS in the CPLEX environment */
        if( CPXchgcoeflist(instance->env_sub[0], instance->lp_sub[0], (instance->Wt)->n_STO, (instance->Wt)->STO_ind, int_vector, (instance->Wt)->STO_R_value[s]) ) 
        {
           printf("\n### Failed to change the RHS of the subproblem.\n");
           fflush(stdout);
           return -2;
        }
    }
    
    free(int_vector);
    free(double_vector);
        
    /* Write the subproblem in the LP format to a file */
    #ifdef COMMMMENTT
        char out_file[1000] = "";
        sprintf(out_file,"CPLEX1_%d.lp",s);
        CPXwriteprob(instance->env_sub[0], instance->lp_sub[0], out_file, NULL);
    #endif
    
    /* Choose the primal simplex method (to be able to get a RAY, if necessary) */   
    if( (status = CPXsetintparam (instance->env_sub[0], CPX_PARAM_LPMETHOD, 1)) )
    {
        printf("\n### Failed to choose the simplex algorithm.\n");
        fflush(stdout);
        return -2;
    }

    /* Turn OFF scaling (to be able to get a RAY, if necessary) */
    if( CPXsetintparam (instance->env_sub[0], CPX_PARAM_SCAIND, 1) ) 
    {
        printf("\n### Failed to turn off scaling.\n");
        fflush(stdout);
        return -2;
    }

    /* Turn OFF presolver (to be able to get a RAY, if necessary) */
    if( CPXsetintparam (instance->env_sub[0], CPX_PARAM_PREIND, 0) ) 
    {
        printf("\n### Failed to turn off the presolver.\n");
        fflush(stdout);
        return -2;
    }
    
    /* Solve the current subproblem */
    if( CPXlpopt(instance->env_sub[0], instance->lp_sub[0]) ) 
    {
       printf("\n### Failed to solve the subproblem by using CPLEX.\n");
       fflush(stdout);
       return -2;
    }
    
    /* Get solution from CPLEX */
    status = CPXsolution (instance->env_sub[0], instance->lp_sub[0], &solstat, obj_value, z, NULL, NULL, NULL);
    
    //printf("  --- Bounded subproblem --> Generate EXTREME POINT ---  ");
    status = CPXdelrows(instance->env_sub[0], instance->lp_sub[0], nrows, nrows);
    
    /* An extreme point has been generated */
    return 1;
}

/* Function: TSSP_solve_subproblem_with_box
   Description: Similar to TSSP_solve_subproblem() but it inserts
   a box to the feasible set, in order to have a bounded problem  */
int TSSP_solve_subproblem_with_box(
    int   s,          /* ( ->) index of the subproblem (scenario) */
    double *obj_value, /* (<- ) objective value of the subproblem (may be an extreme point or a ray) */
    double *z,         /* (<- ) solution of the subproblem (may be an extreme point or a ray) */
    double *x,         /* ( ->) current primal solution of the restricted master problem */
    T_instance_data *instance) /* ( ->) instance data */
{ 
    if( (instance == NULL) || (instance->SMPS_data == NULL) ) return 0;
    if( (instance->env_sub[0] == NULL) || (instance->lp_sub[0] == NULL) ) return 0;
    if( (s < 0) || (s >= instance->SMPS_data->nS) ) return 0;
    if( (x == NULL) || (z == NULL) ) return 0;
    
    int i, j, k, t, aux_int, rbeg, rend, /* auxiliaries */
         nr_p1  = instance->SMPS_data->SMPS_periods->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
         nr_p2  = instance->SMPS_data->nr - nr_p1, /* number of constraints in the second stage */
         n_costs;      /* maximum number of modified costs in the subproblem */
    
    char *bound_type;
    
    int  status, solstat, /* auxiliaries */
         *int_vector=NULL; /* array of indices of coefficients to be modified in the objective function */
                          
    double aux_double, /* auxiliaries */
           *double_vector=NULL; /* array of values of coefficients to be modified in the objective function */

    /* Get the total number of columns in the subproblem */
    n_costs = CPXgetnumcols(instance->env_sub[0], instance->lp_sub[0]);
    
    /* Allocate memory */
    aux_int = n_costs;
    if(aux_int < (instance->h)->n_STO)  aux_int = (instance->h)->n_STO;
    if(aux_int < (instance->q)->n_STO)  aux_int = (instance->q)->n_STO;
    if(aux_int < (instance->Wt)->n_STO) aux_int = (instance->Wt)->n_STO;
    
    int_vector = (int *) malloc( sizeof(int) * aux_int );
    double_vector = (double *) malloc( sizeof(double) * aux_int );
    bound_type = (char *) malloc( sizeof(char) * aux_int );
    
    /* Intialize the vector */
    for(i = 0; i < n_costs; i++) instance->cost[i] = 0.0;
    
    /* Modify the RHS according to the scenario */
    if((instance->h)->n_STO > 0)
    {
        for(t = 0; t < (instance->h)->n_STO; t++)
            (instance->h)->value[ (instance->h)->STO_pos[t] ] = (instance->h)->STO_value[s][t];
    }
    /* Set the RHS of the subproblem (costs of the original problem) */
    for(t = 0; t < (instance->h)->nnz; t++)
        instance->cost[ (instance->h)->index[t] ] = (instance->h)->value[t];
    
    /* Modify the matrix T according to the scenario */
    if((instance->T)->n_STO > 0)
    {
        for(t = 0; t < (instance->T)->n_STO; t++)
            (instance->T)->R_value[ (instance->T)->STO_R_pos[t] ] = (instance->T)->STO_R_value[s][t];
    }
    
    /* Compute the new costs using the primal solution x.
       First, compute the product T x (componentwise). Then, check if the result is nonzero and 
       if so subtract from the corresponding component from h. In case the result is zero, we 
       discard it, i.e., we do not add it to the arrays of modified costs (int_vector, double_vector) */
    k = 0;
    for(i = 0; i < nr_p2; i++) 
    {
        aux_double = 0.0;
        
        /* Check if the row is not empty */
        if((instance->T)->row_len[i] > 0)
        {
            /* Get the initial position of the column */
            rbeg = (instance->T)->R_row_ind[i];
            rend = rbeg + (instance->T)->row_len[i];
            
            /* Compute the product of x by the i-th row of T */
            for(j = rbeg; j < rend; j++)
            {
                /* Multiply the coefficient by the corresponding component of x */
                if( fabs(x[ (instance->T)->R_col_ind[j] ]) > TOL_ZERO )
                    aux_double += (instance->T)->R_value[j] * x[ (instance->T)->R_col_ind[j] ];
            }
        }
        instance->cost[i] -= aux_double;
            
        int_vector[k] = i;
        double_vector[k] = instance->cost[i];
        k++;
    }
    
    /* Change the objective functions costs in the CPLEX environment */
    if( CPXchgobj(instance->env_sub[0], instance->lp_sub[0], k, int_vector, double_vector) ) 
    {
       printf("\n### Failed to change the costs of the subproblem.\n");
       fflush(stdout);
       return -2;
    }
    
    /* Modify the RHS according to the scenario */
    if((instance->q)->n_STO > 0)
    {
        for(t = 0; t < (instance->q)->n_STO; t++)
            int_vector[t] = (instance->q)->index[ (instance->q)->STO_pos[t] ];
           
        /* Change the RHS in the CPLEX environment */
        if( CPXchgrhs(instance->env_sub[0], instance->lp_sub[0], (instance->q)->n_STO, int_vector, (instance->q)->STO_value[s]) ) 
        {
           printf("\n### Failed to change the RHS of the subproblem.\n");
           fflush(stdout);
           return -2;
        }
    }
    
    /* Modify the coefficient matrix according to the scenario */
    if((instance->Wt)->n_STO > 0)
    {
        for(t = 0; t < (instance->Wt)->n_STO; t++)
            int_vector[t] = (instance->Wt)->R_col_ind[ (instance->Wt)->STO_R_pos[t] ];
         
        /* Change the RHS in the CPLEX environment */
        if( CPXchgcoeflist(instance->env_sub[0], instance->lp_sub[0], (instance->Wt)->n_STO, (instance->Wt)->STO_ind, int_vector, (instance->Wt)->STO_R_value[s]) ) 
        {
           printf("\n### Failed to change the RHS of the subproblem.\n");
           fflush(stdout);
           return -2;
        }
    }
    
    /* Set new lower and upper bounds */
    for(i = 0; i < n_costs; i++) int_vector[i]  = i;
    for(i = 0; i < n_costs; i++) bound_type[i] = 'L';
    for(i = 0; i < n_costs; i++) 
        if(instance->low_b[i] < -10000.0) double_vector[i]  = -10000.0;
        else double_vector[i]  = instance->low_b[i];
    status = CPXchgbds (instance->env_sub[0], instance->lp_sub[0], n_costs, int_vector, bound_type, double_vector);
    for(i = 0; i < n_costs; i++) bound_type[i] = 'U';
    for(i = 0; i < n_costs; i++) 
        if(instance->upp_b[i] > 10000.0) double_vector[i]  = 10000.0;
        else double_vector[i]  = instance->upp_b[i];
    status = CPXchgbds (instance->env_sub[0], instance->lp_sub[0], n_costs, int_vector, bound_type, double_vector);
    
    /* Write the subproblem in the LP format to a file */
    #ifdef COMMMMENTT
        char out_file[1000] = "";
        sprintf(out_file,"CPLEX_%d.lp",s);
        CPXwriteprob(instance->env_sub[0], instance->lp_sub[0], out_file, NULL);
    #endif
    
    /* Choose the primal simplex method (to be able to get a RAY, if necessary) */   
    if( (status = CPXsetintparam (instance->env_sub[0], CPX_PARAM_LPMETHOD, 1)) )
    {
        printf("\n### Failed to choose the simplex algorithm.\n");
        fflush(stdout);
        return -2;
    }

    /* Turn OFF scaling (to be able to get a RAY, if necessary) */
    if( CPXsetintparam (instance->env_sub[0], CPX_PARAM_SCAIND, 1) ) 
    {
        printf("\n### Failed to turn off scaling.\n");
        fflush(stdout);
        return -2;
    }

    /* Turn OFF presolver (to be able to get a RAY, if necessary) */
    if( CPXsetintparam (instance->env_sub[0], CPX_PARAM_PREIND, 0) ) 
    {
        printf("\n### Failed to turn off the presolver.\n");
        fflush(stdout);
        return -2;
    }
    
    /* Solve the current subproblem */
    if( CPXlpopt(instance->env_sub[0], instance->lp_sub[0]) ) 
    {
       printf("\n### Failed to solve the subproblem by using CPLEX.\n");
       fflush(stdout);
       return -2;
    }
    
    /* Get solution from CPLEX */
    status = CPXsolution (instance->env_sub[0], instance->lp_sub[0], &solstat, obj_value, z, NULL, NULL, NULL);

    /* Restore the original lower and upper bounds */
    for(i = 0; i < n_costs; i++) int_vector[i]  = i;
    for(i = 0; i < n_costs; i++) bound_type[i] = 'L';
    status = CPXchgbds (instance->env_sub[0], instance->lp_sub[0], n_costs, int_vector, bound_type, instance->low_b);
    for(i = 0; i < n_costs; i++) bound_type[i] = 'U';
    status = CPXchgbds (instance->env_sub[0], instance->lp_sub[0], n_costs, int_vector, bound_type, instance->upp_b);
    
    free(int_vector);
    free(double_vector);
    free(bound_type);
    
    /* An extreme point has been generated */
    return 1;
}

