/*
 *  COLUMN GENERATION PROCEDURE FOR SOLVING 
 *  TWO-STAGE STOCHASTIC PROGRAMMING PROBLEMS (TSSP)
 *  [Aggregated, k-best columns]
 *
 *  Authors: Pablo Gonzalez-Brevis (p.gonzalez-brevis@sms.ed.ac.uk)
 *           Pedro Munari (munari@dep.ufscar.br)
 *
 *  Last modified: Jul/2013
 *
 *  ON TOP OF HOPDM/PDCGM
 *  Written by:    Jacek Gondzio
 *                 Department of Mathematics and Statistics,
 *                 The University of Edinburgh,
 *                 Mayfield Road, King's Buildings.
 *                 EH9 3JZ Edinburgh, Scotland.
 *                 on leave from: Systems Research Institute,
 *                 Polish Academy of Sciences,
 *                 Newelska 6, 01-447 Warsaw, Poland.
 */
 
/* Bring in the header files */
#include "include/tssp-pdcgm.h"
#include "include/tssp_subproblem.c"

/*---------------------------------------------------------------------------*/
/* Increment the less significative counter (and apply carry-over if needed) */
/*---------------------------------------------------------------------------*/
int TSSP_increment_array_counter(
    int *array_counter_ind, 
    int *array_counter_len, 
    int array_counter_n)
{
    char flag = 0;
    
    /* Start form the less significative counter (the last) */
    int i = array_counter_n - 1;
    
    do
    {
        array_counter_ind[i]++;
        flag = 0;
        
        if( (array_counter_ind[i] == array_counter_len[i]) && (i > 0) )
        {
            /* Carry-over: increment the precendent counter */
            array_counter_ind[i] = 0;
            flag = 1;
        }
        i--;
    }
    while( (flag) && (i >= 0) );

    return 1;
}


/*------------------------------------------------------------*/
/* SOLVE THE CORE FORMULATION (WITHOUT SCENARIOS) USING HOPDM */
/*------------------------------------------------------------*/
int TSSP_solve_EVP(
    double *optimal_value,
    double *x,
    T_instance_data *instance)
{
    int i, j, k, s, cbeg, cend;

    long   code = 0, presolve = 0, iter = 0, iolog = 0, iexit = 0, irobj, status,
           *clpnts = 0, nzQ = 0, *qcol_ptr = 0, lpqp = 0, NewtonEqnMethod = 0, restart = 0;
    
    INTS   *qrow_ind = 0, *rwnmbs = 0,
           *row_type = 0;  /* Type of constraints: EQUAL, GREATER_EQUAL, LESS_EQUAL, OBJECTIVE */
           
    double objcon = 0.0, *q_diag = 0, *q_coeff = 0, *low_b = 0, *upp_b = 0,
           errorb, erroru, errorc, asmall, alarge, opt_tol = 1.0e-6,
           *coeff, *RANGES = 0, *RHS = 0, *costs = 0;

    Pb *my_Pb;
    Pb_save *L_save = (Pb_save *) malloc( sizeof(Pb_save) );
    
    T_SMPS_DATA* SMPS_data = instance->SMPS_data;
    
    /* Set local variables */
    int nr = SMPS_data->nr, /* number of rows in the coefficient matrix (CORE description) */
        nc = SMPS_data->nc, /* number of columns in the coefficient matrix (CORE description) */
        nr_p1 = (SMPS_data->SMPS_periods)->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
        nc_p1 = (SMPS_data->SMPS_periods)->first_variable_period[1];   /* number of variables in the first stage */
    
    /* Set the row that describes the objective function */
    irobj = 1; /* Fortran index! */
    
    /* Set the arrays of variable bounds according to the HOPDM style */
    low_b = (double *) malloc( sizeof(double) * nc );
    upp_b = (double *) malloc( sizeof(double) * nc );
    for(i = 0; i < nc; i++) 
    {
        low_b[i] = SMPS_data->MPS_variables[i].lower_bound;
        if( (SMPS_data->MPS_variables[i].type == BD) || (SMPS_data->MPS_variables[i].type == FX) ) upp_b[i] = SMPS_data->MPS_variables[i].upper_bound;
        else upp_b[i] = 1.e30;
    }
    
    /* Set the array of row types according to the HOPDM style */

    row_type = (INTS *) malloc( sizeof(INTS) * nr );
    for(i = 0; i < nr; i++) 
    {
        if(SMPS_data->MPS_constraints[i].type == E) row_type[i] = 1;
        else if(SMPS_data->MPS_constraints[i].type == G) row_type[i] = 2;
        else if(SMPS_data->MPS_constraints[i].type == L) row_type[i] = 3;
        else row_type[i] = 4;
    }
    
    /* Set the array of RHS values */
    RHS = (double *) malloc( sizeof(double) * nr );
    for(i = 0; i < nr_p1; i++) RHS[i] = SMPS_data->RHS[i];

    /* Compute the corresponding EVP values */

    /* Vector h */
    if((instance->h)->n_STO > 0)
    {
        for(i = 0; i < (instance->h)->n_STO; i++) 
            (instance->h)->value[ (instance->h)->STO_pos[i] ] = instance->scenario_probability[0] * (instance->h)->STO_value[0][i];
            
        for(s = 1; s < SMPS_data->nS; s++) 
            for(i = 0; i < (instance->h)->n_STO; i++) 
                (instance->h)->value[ (instance->h)->STO_pos[i] ] += instance->scenario_probability[s] * (instance->h)->STO_value[s][i];
   }
   
   /* Vector q */
   if((instance->q)->n_STO > 0)
   { 
        for(i = 0; i < (instance->q)->n_STO; i++) 
            (instance->q)->value[ (instance->q)->STO_pos[i] ] = instance->scenario_probability[0] * (instance->q)->STO_value[0][i];
            
        for(s = 1; s < SMPS_data->nS; s++) 
            for(i = 0; i < (instance->q)->n_STO; i++) 
                (instance->q)->value[ (instance->q)->STO_pos[i] ] += instance->scenario_probability[s] * (instance->q)->STO_value[s][i];
    }
    
    /* matrix W */
    if((instance->Wt)->n_STO > 0)
    {
        
        for(i = 0; i < (instance->Wt)->n_STO; i++) 
            (instance->Wt)->R_value[ (instance->Wt)->STO_R_pos[i] ] = instance->scenario_probability[0] * (instance->Wt)->STO_R_value[0][i];
            
        for(s = 1; s < SMPS_data->nS; s++) 
            for(i = 0; i < (instance->Wt)->n_STO; i++) 
                (instance->Wt)->R_value[ (instance->Wt)->STO_R_pos[i] ] += instance->scenario_probability[s] * (instance->Wt)->STO_R_value[s][i];
    }
    
    /* matrix T */
    if((instance->T)->n_STO > 0)
    {
        for(i = 0; i < (instance->T)->n_STO; i++) 
            (instance->T)->R_value[ (instance->T)->STO_R_pos[i] ] = instance->scenario_probability[0] * (instance->T)->STO_R_value[0][i];
            
        for(s = 1; s < SMPS_data->nS; s++) 
            for(i = 0; i < (instance->T)->n_STO; i++) 
                (instance->T)->R_value[ (instance->T)->STO_R_pos[i] ] += instance->scenario_probability[s] * (instance->T)->STO_R_value[s][i];
        
        /* reset the columnwise representation of T */
        STO_SMatrix_createCWfromRW(instance->T, SMPS_data->nS);
    }
    
    /* Modify the second stage entries according to the EVP values */
    for(i = nr_p1; i < nr; i++) RHS[i] = 0.0;
    for(i = 0; i < (instance->h)->nnz; i++)
        RHS[ nr_p1 + (instance->h)->index[i] ] = (instance->h)->value[i];
    
    /* Set the array of RANGE values */
    RANGES = (double *) malloc( sizeof(double) * nr );
    if(SMPS_data->RANGES == NULL) for(i = 0; i < nr; i++) RANGES[i] = 1.e30;
    else for(i = 0; i < nr; i++) RANGES[i] = SMPS_data->RANGES[i];
    
    /* Set the array of RHS values */
    costs = (double *) malloc( sizeof(double) * nc );
    for(i = 0; i < nc_p1; i++) costs[i] = SMPS_data->c[i];
    
    /* Modify the second stage entries according to the EVP values */
    for(i = nc_p1; i < nc; i++) costs[i] = 0.0;
    for(i = 0; i < (instance->q)->nnz; i++)
        costs[ nc_p1 + (instance->q)->index[i] ] = -(instance->q)->value[i];
    
    /* Set the indices of the coeff matrix according to the HOPDM style */
    clpnts = (long*) malloc( sizeof(long) * (nc+1) );
    rwnmbs = (INTS*) malloc( sizeof(INTS) * ((SMPS_data->A)->nnz) );
    coeff = (double*) malloc( sizeof(double) * ((SMPS_data->A)->nnz) );
    
    /* Modify the objective function according to the EVP values */
    k=0;
    for(i = 0; i < nc_p1; i++) 
    {
        clpnts[i] = k + 1;
        
        cbeg = (SMPS_data->A)->C_col_ind[i];
        cend = cbeg + (SMPS_data->A)->col_len[i];
        
        for(j = cbeg; j < cend; j++) 
        {
            if( (SMPS_data->A)->C_row_ind[j] < nr_p1 )
            {
                rwnmbs[k] = (SMPS_data->A)->C_row_ind[j] + 1; 
                coeff[k] = (SMPS_data->A)->C_value[j];
                k++;
            }
        }
        
        cbeg = (instance->T)->C_col_ind[i];
        cend = cbeg + (instance->T)->col_len[i];
        
        for(j = cbeg; j < cend; j++) 
        {
            rwnmbs[k] = (instance->T)->C_row_ind[j] + nr_p1 + 1; 
            coeff[k] = (instance->T)->C_value[j];
            k++;
        }
    }
    for(i = nc_p1; i < nc; i++) 
    {
        clpnts[i] = k + 1;
        
        if(fabs(costs[i]) > TOL_ZERO)
        {
            rwnmbs[k] = 1; 
            coeff[k] = costs[i];
            k++;
        }
        
        cbeg = (instance->Wt)->R_row_ind[i-nc_p1];
        cend = cbeg + (instance->Wt)->row_len[i-nc_p1];
        
        for(j = cbeg; j < cend; j++) 
        {
            rwnmbs[k] = (instance->Wt)->R_col_ind[j] + nr_p1 + 1; 
            coeff[k] = (instance->Wt)->R_value[j];
            k++;
        }
    }
    clpnts[nc] = k + 1;

    /* Populate the data structure with the data of the current RMP */
    my_Pb = load_Pb (NULL, NULL, row_type, clpnts, rwnmbs, coeff,
                RHS, RANGES, low_b, upp_b, nr, nc, k, 
                nzQ, lpqp, objcon, q_diag, qcol_ptr, qrow_ind, q_coeff, 
                presolve, irobj, iolog, MINIMIZE, NewtonEqnMethod); 
    my_Pb->exit_code = 0;
    
    /* Call the preprocessor (it is not presolving) */
    preproc_Pb (my_Pb);
    
    printf("\nSolving the EVP ... \n");
    code = solve_Pb (my_Pb, restart, L_save, &opt_tol, &iexit,
           &errorb, &erroru, &errorc, &asmall, &alarge, &iter);

    if ( code != 0 ) 
    {
        printf("After solve_Pb: code= %ld\n", code);
        
        free_Pb(my_Pb);
        free(clpnts);
        free(rwnmbs);
        free(coeff);
        free(RHS);
        free(RANGES);
        free(row_type);
        free(low_b);
        free(upp_b);
        
        /* Return without success */
        return 0;
    } 
    
    double objective;
    double *primal = PDCGM_ALLOC (double, my_Pb->ncopy),
           *dual   = PDCGM_ALLOC (double, my_Pb->mcopy);

    print_Pb(my_Pb, "", "", 0, dual);

    get_opt_solution (my_Pb, &status, &objective, primal, dual, NULL, NULL);
    printf("Optimal value: %.8f (status= %ld)\n", objective, status);

    /* Copy the optimal solution (first stage variables only) */
    if(x != NULL) for(i = 0; i < nc_p1; i++) x[i] = primal[i]; 
    if(optimal_value != NULL) *optimal_value = objective;

    free(primal);
    free(dual);
    
    free_Pb(my_Pb);
    free(clpnts);
    free(rwnmbs);
    free(coeff);
    free(RHS);
    free(RANGES);
    free(row_type);
    free(low_b);
    free(upp_b);
    
    /* Return successfully */
    return 1;
}


/*--------------------------------------------------------*/
/* DISAGGREGATED ORACLE FUNCTION                          */
/*--------------------------------------------------------*/
/* The oracle procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short TSSP_disagg_oracle(
    double *primal_violation,   /* total violation of the generated constraints, if any */
    double *dual_violation,     /* total violation of the generated columns, if any */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data */
{
    /* Get the pointer to the instance data */
    T_instance_data *instance = (T_instance_data*) instance_data;
    if( (instance == NULL) || (instance->SMPS_data == NULL) ) return 0;
    
    short flag_ray=0;
    
    int i, j, k, s, ret, 
         nr_p1 = instance->SMPS_data->SMPS_periods->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
         nc_p1 = instance->SMPS_data->SMPS_periods->first_variable_period[1],   /* number of variables in the first stage */
         nr_p2  = instance->SMPS_data->nr - nr_p1, /* number of constraints in the second stage */
         nr_linking = nc_p1, /* number of LINKING constraints in the master problem (DUAL FORMULATION) */
         nr_convexity = instance->SMPS_data->nS, /* number of convexity constraints in the master problem (DUAL FORMULATION) */
         nr_RMP = nr_linking + nr_convexity + 1; /* total number constraints in the master problem (DUAL FORMULATION) */
         
    double aux_double, /* auxiliary */
           obj_value,  /* objective value returned by the subproblem */
           *x,         /* current dual solution (first stage variables) */
           *new_column;/* coefficients on the linking constraints of the generated column */
           
    double *z; /* solution of the subproblem (extreme point or extreme ray) */
    
    /* Coefficient matrix to store the generated columns */ 
    PDCGM_SMatrix_CW *M; 
    
    printf("\nHello from the ORACLE();\n");
    fflush(stdout);
    
    /* Get the current dual solution */
    x = PDCGM_get_dual_solution(PDCGM_env);
    
    /* Allocate memory for the solution of the subproblem */
    z = PDCGM_ALLOC(double, nr_p2);
    new_column = PDCGM_ALLOC(double, nr_linking);
    for(i = 0; i < nr_linking; i++) new_column[i] = 0.0;
    
    /* Allocate the auxiliary matrix in memory */
    M = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    if(M == NULL) return -1;
    PDCGM_set_SMatrix_CW(M, nr_RMP, nr_convexity, (nr_linking + 2) * nr_convexity);
    
    /* Set initial values */
    M->m = nr_RMP;
    M->n = 0;
    M->nz = 0;
    k = 0;
    ret = 0;
    *dual_violation = 0.0;
    
    for(s = 0; s < nr_convexity; s++)
    {
        /* Solve the Benders subproblem */
        obj_value = 0.0;
        if(PDCGM_env->outer == 0) ret = TSSP_solve_subproblem_with_box(s, &obj_value, z, x, instance);
        else ret = TSSP_solve_subproblem(s, &obj_value, z, x, instance);
        
        //printf("\nTSSP_solve_subproblem(%d): obj_value = %lf, ret = %d ", s, obj_value, ret);
        obj_value *= instance->scenario_probability[s];
        
        if(ret <= 0) 
        {
            printf("\n### ERROR ###\n");
            return -1;
        }
        
        if(ret == 2)
        { 
            /* The subproblem is unbounded: Extreme ray! */
            if(flag_ray == 0)
            {
                /* Discard any extreme point that may have been previously generated */
                M->n = 0;
                M->nz = 0;
                k = 0;
                *dual_violation = 0.0;
                flag_ray=1;
            }
            
            /* Check if the reduced cost is negative */
            if(obj_value < -TOL_ZERO)
            {
                /* Set the dual violation (reduced cost) */
                *dual_violation += obj_value;
                
                /* Start a new column */
                M->clpnts[M->n] = k+1;
                (M->n)++;
                
                /* Set the coefficients of the linking rows */
                if((instance->T)->n_STO > 0)
                {
                    for(i = 0; i < (instance->T)->n_STO; i++)
                        (instance->T)->R_value[ (instance->T)->STO_R_pos[i] ] = (instance->T)->STO_R_value[s][i];
                }
                for(i = 0; i < nr_p2; i++)
                {
                    for(j = (instance->T)->R_row_ind[i]; j < (instance->T)->R_row_ind[i] + (instance->T)->row_len[i]; j++)
                    {
                        /* Multiply each coefficient by the corresponding entry in z */
                        new_column[ (instance->T)->R_col_ind[j] ] += (instance->T)->R_value[j] * z[i];
                    }
                }               
                for(i = 0; i < nr_linking; i++)
                {
                    /* Check if the resulting coefficient is nonzero */
                    if(fabs(new_column[i]) > TOL_ZERO)
                    {
                        M->coeff[k] = instance->scenario_probability[s] * new_column[i];
                        M->rwnmbs[k] = i + 1; /* Fortran index! */
                        k++;
                    }
                    new_column[i] = 0.0;
                }
                
                /* Add the coefficient cost */
                aux_double = 0.0;
                if((instance->h)->n_STO > 0)
                {
                    for(i = 0; i < (instance->h)->n_STO; i++)
                        (instance->h)->value[ (instance->h)->STO_pos[i] ] = (instance->h)->STO_value[s][i];
                }
                for(i = 0; i < (instance->h)->nnz; i++)
                    aux_double += (instance->h)->value[i] * z[ (instance->h)->index[i] ];
                
                M->coeff[k] = instance->scenario_probability[s] * aux_double;
                M->rwnmbs[k] = nr_RMP; /* Fortran index! */
                k++;
                
            }
            
        }
        else if(flag_ray == 0)
        {
            
            /* If the subproblem solution is an extreme point, then sum 
               the value of dual component to the reduced cost */
            obj_value -= CNVXTY_VALUE * x[nr_linking + s];

            /* Check if the reduced cost is negative */
            if(obj_value < -TOL_ZERO || PDCGM_env->outer == 0)
            {
                /* Update the dual violation (reduced cost) */
                *dual_violation += obj_value;
                
                /* Start a new column */
                M->clpnts[M->n] = k+1;
                (M->n)++;
                
                /* Set the coefficients of the linking rows */
                if((instance->T)->n_STO > 0)
                {
                    for(i = 0; i < (instance->T)->n_STO; i++)
                        (instance->T)->R_value[ (instance->T)->STO_R_pos[i] ] = (instance->T)->STO_R_value[s][i];
                }
                for(i = 0; i < nr_p2; i++)
                {
                    for(j = (instance->T)->R_row_ind[i]; j < (instance->T)->R_row_ind[i] + (instance->T)->row_len[i]; j++)
                    {
                        /* Multiply each coefficient by the corresponding entry in z */
                        new_column[ (instance->T)->R_col_ind[j] ] += (instance->T)->R_value[j] * z[i];
                    }
                }               
                for(i = 0; i < nr_linking; i++)
                {
                    /* Check if the resulting coefficient is nonzero */
                    if(fabs(new_column[i]) > TOL_ZERO)
                    {
                        M->coeff[k] = instance->scenario_probability[s] * new_column[i];
                        M->rwnmbs[k] = i + 1; /* Fortran index! */
                        k++;
                    }
                    new_column[i] = 0.0;
                }
                
                /* Set the coefficient of the convexity row */
                M->coeff[k] = CNVXTY_VALUE;
                M->rwnmbs[k] = nr_linking + s + 1; /* Fortran index! */
                k++;
                
                /* Add the coefficient cost */
                aux_double = 0.0;
                 if((instance->h)->n_STO > 0)
                {
                    for(i = 0; i < (instance->h)->n_STO; i++)
                        (instance->h)->value[ (instance->h)->STO_pos[i] ] = (instance->h)->STO_value[s][i];
                }
                for(i = 0; i < (instance->h)->nnz; i++)
                    aux_double += (instance->h)->value[i] * z[ (instance->h)->index[i] ];
                
                M->coeff[k] = instance->scenario_probability[s] * aux_double;
                M->rwnmbs[k] = nr_RMP; /* Fortran index! */
                k++;
            }
        }
    }
    
    /* End of columns! */
    M->clpnts[M->n] = k+1; /* -------> Fortran index */
    M->nz= k;
    
    if(*dual_violation > -TOL_ZERO) *dual_violation = 0.0;
    
    /* Add the generated columns to the coefficient matrix  */
    if(M->n > 0)
    {
        /* Call the PDCGM function that add the new columns to the current RMP */
        ret = PDCGM_add_columns(PDCGM_env, M, NULL);

        /* If the returning value is negative, report the ERROR */
        if(ret < 0) return -1;      
    }
    
    /* Clean memory */  
    PDCGM_free_SMatrix_CW(M);
    PDCGM_FREE(M);
    PDCGM_FREE(z);
    PDCGM_FREE(new_column);
    
    return 1;
}


/*-----------------------------------------------------------------------------------*/
/* AGGREGATED ORACLE FUNCTION 1: EXTREME POINTS ARE AGGREGATED, BUT NOT EXTREME RAYS */
/*-----------------------------------------------------------------------------------*/
/* The oracle procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short TSSP_agg1_oracle(
    double *primal_violation,   /* total violation of the generated constraints, if any */
    double *dual_violation,     /* total violation of the generated columns, if any */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data */
{
    /* Get the pointer to the instance data */
    T_instance_data *instance = (T_instance_data*) instance_data;
    if( (instance == NULL) || (instance->SMPS_data == NULL) ) return 0;
    
    short flag_ray=0;
    
    int i, j, k, s, ret, 
        nS = instance->SMPS_data->nS,
        nr_p1 = instance->SMPS_data->SMPS_periods->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
        nc_p1 = instance->SMPS_data->SMPS_periods->first_variable_period[1],   /* number of variables in the first stage */
        nr_p2  = instance->SMPS_data->nr - nr_p1, /* number of constraints in the second stage */
        nr_linking = nc_p1, /* number of LINKING constraints in the master problem (DUAL FORMULATION) */
        nr_RMP = nr_linking + 2; /* total number constraints in the master problem (DUAL FORMULATION) */
         
    double aux_double, /* auxiliary */
           obj_value,  /* objective value returned by the subproblem */
           *x,         /* current dual solution (first stage variables) */
           *new_column;/* coefficients on the linking constraints of the generated column */
           
    double *z; /* solution of the subproblem (extreme point or extreme ray) */
    
    /* Coefficient matrix to store the generated columns */ 
    PDCGM_SMatrix_CW *M; 
    
    printf("\nHello from the ORACLE();\n");
    fflush(stdout);
    
    /* Get the current dual solution */
    x = PDCGM_get_dual_solution(PDCGM_env);
    
    /* Allocate memory for the solution of the subproblem */
    z = PDCGM_ALLOC(double, nr_p2);
    new_column = PDCGM_ALLOC(double, nr_RMP);
    for(i = 0; i < nr_RMP; i++) new_column[i] = 0.0;
    
    /* Allocate the auxiliary matrix in memory */
    M = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    if(M == NULL) return -1;
    PDCGM_set_SMatrix_CW(M, nr_RMP, nS, (nr_linking + 2) * nS);
    
    /* Set initial values */
    M->m = nr_RMP;
    M->n = 0;
    M->nz = 0;
    k = 0;
    ret = 0;
    *dual_violation = 0.0;
    
    for(s = 0; s < nS; s++)
    {
        /* Solve the Benders subproblem */
        obj_value = 0.0;
        if(PDCGM_env->outer == 0) ret = TSSP_solve_subproblem_with_box(s, &obj_value, z, x, instance);
        else ret = TSSP_solve_subproblem(s, &obj_value, z, x, instance);
        
        /* Multiply the subproblem objective by the scenario probability */
        obj_value *= instance->scenario_probability[s];
        
        if(ret <= 0) 
        {
            printf("\n### ERROR ###\n");
            return -1;
        }
        
        if(ret == 2)
        { 
            /* The subproblem is unbounded: Extreme ray! */
            if(flag_ray == 0)
            {
                for(i = 0; i < nr_RMP; i++) new_column[i] = 0.0;
                
                /* Discard any extreme point that may have been previously generated */
                M->n = 0;
                M->nz = 0;
                k = 0;
                *dual_violation = 0.0;
                flag_ray=1;
            }
            
            /* Set the dual violation (reduced cost) */
            *dual_violation += obj_value;
            
            /* Start a new column */
            M->clpnts[M->n] = k+1;
            (M->n)++;
            
            /* Set the coefficients of the linking rows */
            if((instance->T)->n_STO > 0)
            {
                for(i = 0; i < (instance->T)->n_STO; i++)
                    (instance->T)->R_value[ (instance->T)->STO_R_pos[i] ] = (instance->T)->STO_R_value[s][i];
            }
            for(i = 0; i < nr_p2; i++)
            {
                for(j = (instance->T)->R_row_ind[i]; j < (instance->T)->R_row_ind[i] + (instance->T)->row_len[i]; j++)
                {
                    /* Multiply each coefficient by the corresponding entry in z */
                    new_column[ (instance->T)->R_col_ind[j] ] += (instance->T)->R_value[j] * z[i];
                }
            }
            
            /* Set the column in the sparse representation */               
            for(i = 0; i < nr_linking; i++)
            {
                /* Check if the resulting coefficient is nonzero */
                if(fabs(new_column[i]) > TOL_ZERO)
                {
                    M->coeff[k] = instance->scenario_probability[s] * new_column[i];
                    M->rwnmbs[k] = i + 1; /* Fortran index! */
                    k++;
                }
                new_column[i] = 0.0;
            }
            
            /* Compute the coefficient cost */
            aux_double = 0.0;
            if((instance->h)->n_STO > 0)
            {
                for(i = 0; i < (instance->h)->n_STO; i++)
                    (instance->h)->value[ (instance->h)->STO_pos[i] ] = (instance->h)->STO_value[s][i];
            }
            for(i = 0; i < (instance->h)->nnz; i++)
                aux_double += (instance->h)->value[i] * z[ (instance->h)->index[i] ];
            
            M->coeff[k] = instance->scenario_probability[s] * aux_double;
            M->rwnmbs[k] = nr_RMP; /* Fortran index! */
            k++;
        }
        else if(flag_ray == 0)
        {
            
            /* Update the dual violation (reduced cost) */
            *dual_violation += obj_value;
            
            /* Set the coefficients of the linking rows */
            if((instance->T)->n_STO > 0)
            {
                for(i = 0; i < (instance->T)->n_STO; i++)
                    (instance->T)->R_value[ (instance->T)->STO_R_pos[i] ] = (instance->T)->STO_R_value[s][i];
            }
            for(i = 0; i < nr_p2; i++)
            {
                for(j = (instance->T)->R_row_ind[i]; j < (instance->T)->R_row_ind[i] + (instance->T)->row_len[i]; j++)
                {
                    /* Multiply each coefficient by the corresponding entry in z */
                    new_column[ (instance->T)->R_col_ind[j] ] += instance->scenario_probability[s] * (instance->T)->R_value[j] * z[i];
                }
            }
            
            /* Set the coefficient of the convexity row */
            new_column[nr_linking] = CNVXTY_VALUE;
            
            /* Add the coefficient cost */
            aux_double = 0.0;
            if((instance->h)->n_STO > 0)
            {
                for(i = 0; i < (instance->h)->n_STO; i++)
                    (instance->h)->value[ (instance->h)->STO_pos[i] ] = (instance->h)->STO_value[s][i];
            }
            for(i = 0; i < (instance->h)->nnz; i++)
                aux_double += (instance->h)->value[i] * z[ (instance->h)->index[i] ];
            new_column[nr_RMP - 1] += instance->scenario_probability[s] * aux_double;            
        }
    }
    
    /* End of columns! */
    M->clpnts[M->n] = k+1; /* -------> Fortran index */
    M->nz= k;
    
    /* Check if the column has been generated from extreme points */
    if(flag_ray == 0) 
    {
        /* add the value of dual component to the reduced cost */
        *dual_violation -= CNVXTY_VALUE * x[nr_linking];
        
        /* Add the generated column to the PDCGM coefficient matrix  */
        if(*dual_violation <= -TOL_ZERO || PDCGM_env->outer == 0)
        {
            /* Call the PDCGM function that add the new columns to the current RMP */
            ret = PDCGM_add_dense_columns(PDCGM_env, new_column, NULL, nr_RMP, 1);

            /* If the returning value is negative, report the ERROR */
            if(ret < 0) return ret;     
        }
        else PDCGM_env->ncols_last = 0;
    }
    else if(M->n > 0)
    {
        /* One or more columns have been generated from extreme rays */
        
        /* Call the PDCGM function that add the new columns to the current RMP */
        ret = PDCGM_add_columns(PDCGM_env, M, NULL);

        /* If the returning value is negative, report the ERROR */
        if(ret < 0) return -1;      
    }
    else PDCGM_env->ncols_last = 0; 
    
    if(*dual_violation > -TOL_ZERO) *dual_violation = 0.0;
    
    /* Clean memory */  
    PDCGM_free_SMatrix_CW(M);
    PDCGM_FREE(M);
    PDCGM_FREE(z);
    PDCGM_FREE(new_column);
    
    return 1;
}


/*------------------------------------------------------------------------------*/
/* AGGREGATED ORACLE FUNCTION 2: EXTREME POINTS AND EXTREME RAYS ARE AGGREGATED */
/*------------------------------------------------------------------------------*/
/* The oracle procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns, 2 -> one or more valid inequalities */
static short TSSP_agg2_oracle(
    double *primal_violation,   /* total violation of the generated constraints, if any */
    double *dual_violation,     /* total violation of the generated columns, if any */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data */
{
    /* Get the pointer to the instance data */
    T_instance_data *instance = (T_instance_data*) instance_data;
    if( (instance == NULL) || (instance->SMPS_data == NULL) ) return 0;
    
    short flag_ray=0;
    
    int i, j, s, ret, 
         nS = instance->SMPS_data->nS,
         nr_p1 = instance->SMPS_data->SMPS_periods->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
         nc_p1 = instance->SMPS_data->SMPS_periods->first_variable_period[1],   /* number of variables in the first stage */
         nr_p2  = instance->SMPS_data->nr - nr_p1, /* number of constraints in the second stage */
         nr_linking = nc_p1, /* number of LINKING constraints in the master problem (DUAL FORMULATION) */
         nr_RMP = nr_linking + 2; /* total number constraints in the master problem (DUAL FORMULATION) */
         
    double aux_double, /* auxiliary */
           obj_value,  /* objective value returned by the subproblem */
           *x,         /* current dual solution (first stage variables) */
           *new_column;/* generated column */
    
    double *z; /* solution of the subproblem (extreme point or extreme ray) */
    
    printf("\nHello from the ORACLE();\n");
    fflush(stdout);
    
    /* Get the current dual solution */
    x = PDCGM_get_dual_solution(PDCGM_env);
    
    /* Allocate memory for the solution of the subproblem and the generated column */
    z = PDCGM_ALLOC(double, nr_p2);
    new_column = PDCGM_ALLOC(double, nr_RMP);
    for(i = 0; i < nr_RMP; i++) new_column[i] = 0.0;
    
    /* Set initial values */
    ret = 0;
    *dual_violation = 0.0;
    
    for(s = 0; s < nS; s++)
    {
        /* Solve the Benders subproblem */
        obj_value = 0.0;
        
        if(PDCGM_env->outer == 0) ret = TSSP_solve_subproblem(s, &obj_value, z, x, instance);
        else ret = TSSP_solve_subproblem(s, &obj_value, z, x, instance);
        
        if(ret <= 0) 
        {
            printf("\n### ERROR ###\n");
            return -1;
        }
        
        if(ret == 2)
        { 
            /* The subproblem is unbounded: Extreme ray! */

            if(flag_ray == 0)
            {
                /* Discard any extreme point that may have been previously generated */
                for(i = 0; i < nr_RMP; i++) new_column[i] = 0.0;
                *dual_violation = 0.0;
                flag_ray=1;
            }

            /* Update the dual violation (reduced cost) */
            *dual_violation += instance->scenario_probability[s] * obj_value;
            
            /* Set the coefficients of the linking rows */
            if((instance->T)->n_STO > 0)
            {
                for(i = 0; i < (instance->T)->n_STO; i++)
                    (instance->T)->R_value[ (instance->T)->STO_R_pos[i] ] = (instance->T)->STO_R_value[s][i];
            }
            for(i = 0; i < nr_p2; i++)
            {
                for(j = (instance->T)->R_row_ind[i]; j < (instance->T)->R_row_ind[i] + (instance->T)->row_len[i]; j++)
                {
                    /* Multiply each coefficient by the corresponding entry in z */
                    new_column[ (instance->T)->R_col_ind[j] ] += instance->scenario_probability[s] * (instance->T)->R_value[j] * z[i];
                }
            }
            
            /* Compute the coefficient cost */
            aux_double = 0.0;
            if((instance->h)->n_STO > 0)
            {
                for(i = 0; i < (instance->h)->n_STO; i++)
                    (instance->h)->value[ (instance->h)->STO_pos[i] ] = (instance->h)->STO_value[s][i];
            }
            for(i = 0; i < (instance->h)->nnz; i++)
                aux_double += (instance->h)->value[i] * z[ (instance->h)->index[i] ];
            new_column[nr_RMP - 1] += instance->scenario_probability[s] * aux_double;
        }
        else if(flag_ray == 0)
        {
            /* Update the dual violation (reduced cost) */
            *dual_violation += instance->scenario_probability[s] * obj_value;
            
            /* Set the coefficients of the linking rows */
            if((instance->T)->n_STO > 0)
            {
                for(i = 0; i < (instance->T)->n_STO; i++)
                    (instance->T)->R_value[ (instance->T)->STO_R_pos[i] ] = (instance->T)->STO_R_value[s][i];
            }
            for(i = 0; i < nr_p2; i++)
            {
                for(j = (instance->T)->R_row_ind[i]; j < (instance->T)->R_row_ind[i] + (instance->T)->row_len[i]; j++)
                {
                    /* Multiply each coefficient by the corresponding entry in z */
                    new_column[ (instance->T)->R_col_ind[j] ] += instance->scenario_probability[s] * (instance->T)->R_value[j] * z[i];
                }
            }
            
            /* Set the coefficient of the convexity row */
            new_column[nr_linking] = CNVXTY_VALUE;
            
            /* Add the coefficient cost */
            aux_double = 0.0;
            if((instance->h)->n_STO > 0)
            {
                for(i = 0; i < (instance->h)->n_STO; i++)
                    (instance->h)->value[ (instance->h)->STO_pos[i] ] = (instance->h)->STO_value[s][i];
            }
            for(i = 0; i < (instance->h)->nnz; i++)
                aux_double += (instance->h)->value[i] * z[ (instance->h)->index[i] ];
            new_column[nr_RMP - 1] += instance->scenario_probability[s] * aux_double;
        }
    }
    
    /* If the columns is generated from extreme points, then add to the reduced cost the value of dual component */
    if(flag_ray == 0) *dual_violation -= CNVXTY_VALUE * x[nr_linking];
       
    /* Add the generated column to the PDCGM coefficient matrix  */
    if(*dual_violation <= -TOL_ZERO || PDCGM_env->outer == 0)
    {
        /* Call the PDCGM function that add the new columns to the current RMP */
        ret = PDCGM_add_dense_columns(PDCGM_env, new_column, NULL, nr_RMP, 1);

        /* If the returning value is negative, report the ERROR */
        if(ret < 0) return ret;     
    }
    else PDCGM_env->ncols_last = 0;
    
    if(*dual_violation > -TOL_ZERO) *dual_violation = 0.0;
    
    /* Clean memory */  
    PDCGM_FREE(z);
    PDCGM_FREE(new_column);
    
    return 1;
}

/*--------------------------------------------------------*/
/* SET INITIAL COLUMNS USING ARTIFICIAL VARIABLES         */
/*--------------------------------------------------------*/
int TSSP_set_initial_columns_with_artificial(
    int aggregated,
    PDCGM  *PDCGM_env,      /* PDCGM environment */
    T_SMPS_DATA* SMPS_data) /* instance data in the SMPS format */
{
    
    int i, j, k, cbeg, cend,
         nr_p1 = (SMPS_data->SMPS_periods)->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
         nc_p1 = (SMPS_data->SMPS_periods)->first_variable_period[1],   /* number of variables in the first stage */
         nr_linking = nc_p1, /* number of LINKING constraints in the master problem (DUAL FORMULATION) */
         nr_convexity, /* number of convexity constraints in the master problem (DUAL FORMULATION) */
         nr_RMP, /* total number constraints in the master problem (DUAL FORMULATION) */
         nc_RMP = nr_p1 - 1 + 1, /* number of variables in the master problem (DUAL FORMULATION) */
         nnz_RMP; /* number of nonzeroes in the master problem */
    
    PDCGM_SMatrix_CW *C = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    PDCGM_SMatrix_RW *R = PDCGM_ALLOC(PDCGM_SMatrix_RW, 1);
    
    /* Set the number of convexity constraints */
    if(aggregated == 1) nr_convexity = 1;
    else nr_convexity = SMPS_data->nS;
    nr_RMP = nr_linking + nr_convexity + 1;
    nnz_RMP = (SMPS_data->A)->nnz + nr_p1 + 2*nr_convexity;
        
    /* Set the indices of the coeff matrix according to the HOPDM style */
    if( PDCGM_set_SMatrix_RW(R, nr_RMP, nc_RMP, nnz_RMP) ) return -1;
    R->m = nr_RMP;
    R->n = nc_RMP;
    R->nz = nnz_RMP;
    
    /* Set auxiliary counters */
    k = 0;
    cbeg = 0;
    cend = 0;
    
    /* Set the first stage columns as the constraints of the RMP */
    for(i = 0; i < nr_linking; i++) 
    {
        /* Set the initial position of the new row */
        R->rwpnts[i] = k+1; /* Fortran index! */
        
        /* Get the initial position of the CORE column */
        cbeg = (SMPS_data->A)->C_col_ind[i];
        cend = cbeg + (SMPS_data->A)->col_len[i];
        
        /* Copy the first stage coefficients */
        for(j = cbeg; j < cend; j++)
        {
            if( ( (SMPS_data->A)->C_row_ind[j] > 0 ) && ( (SMPS_data->A)->C_row_ind[j] < nr_p1 ) )
            {
                /* It is an entry in a first stage row */
                R->clnmbs[k] = (SMPS_data->A)->C_row_ind[j]; /* Fortran index, BUT remove the first row (OBJ) */
                R->coeff[k]  = (SMPS_data->A)->C_value[j];
                k++;
            }
        }
    }
    
    /* Set the convexity rows as empty */    
    for(i = 0; i < nr_convexity; i++) 
    {
        R->rwpnts[nr_linking + i] = k+1; /* Fortran index! */
        
        /* Set the entry in the convexity row */
        R->clnmbs[k] = nr_p1; /* Fortran index, BUT remove the first row (OBJ) */
        R->coeff[k]  = CNVXTY_VALUE;
        k++;
    }
    
    /* Set the objective function row, using the RHS of the original formulation */
    R->rwpnts[nr_RMP-1] = k+1; /* Fortran index! */
    for(i = 1; i < nr_p1; i++) /* Starts at 1 to discard de OBJ row of the original formulation */
    {
        if(fabs(SMPS_data->RHS[i]) > TOL_ZERO)
        {
            /* It is an entry in a first stage row */
            R->clnmbs[k] = i; /* Fortran index, BUT remove the first row (OBJ) */
            R->coeff[k] = SMPS_data->RHS[i];
            k++;
        }
    }
    /* It is an entry in a first stage row */
    R->clnmbs[k] = nr_p1; /* Fortran index */
    R->coeff[k]  = 1000000.0; /* MINIMIZATION objective */
    k++;
    
    /* Set the last position of rwpnts */
    R->rwpnts[nr_RMP] = k+1; /* Fortran index! */
    
    /* Set the actual number of nonzeros in the new coeff matrix */
    R->nz = k;
    
    /* Set the columnwise representation from to the rowwise representation */
    PDCGM_set_SMatrix_CW_from_RW(C, R, nr_RMP, nc_RMP, R->nz);
    
    /* Free up the allocated memory */
    PDCGM_free_SMatrix_RW(R);
    free(R);
    
    /* Call the PDCGM function that add the new columns to the current RMP */
    PDCGM_add_auxiliary_columns(PDCGM_env, C);
    
    /* Free the allocated memory */
    PDCGM_free_SMatrix_CW(C);
    free(C);
    
    /* Set the variable bounds of the initial columns */
    for(i = 1; i < nr_p1; i++) 
    {
        if(SMPS_data->MPS_constraints[i].type == E) 
        {
            /* The dual variable must be FREE */
            PDCGM_env->low_bnd[PDCGM_env->auxiliary_begin + i-1] = -1.e30;
            PDCGM_env->upp_bnd[PDCGM_env->auxiliary_begin + i-1] = 1.e30;
        }
        else if(SMPS_data->MPS_constraints[i].type == G)
        {
            /* The dual variable must be NON-POSITIVE */
            PDCGM_env->low_bnd[PDCGM_env->auxiliary_begin + i-1] = -1.e30;
            PDCGM_env->upp_bnd[PDCGM_env->auxiliary_begin + i-1] = 0.0;
        }
        else if(SMPS_data->MPS_constraints[i].type == L) 
        {
            /* The dual variable must be NON-NEGATIVE */
            PDCGM_env->low_bnd[PDCGM_env->auxiliary_begin + i-1] = 0.0;
            PDCGM_env->upp_bnd[PDCGM_env->auxiliary_begin + i-1] = 1.e30;            
        }
        else
        {
            printf("\n### ERROR! \n");
            return -1;
            
            /* The dual variable must be FIXED */
            PDCGM_env->low_bnd[PDCGM_env->auxiliary_begin + i-1] = 0.0;
            PDCGM_env->upp_bnd[PDCGM_env->auxiliary_begin + i-1] = 0.0;
        }
    }
    
    return 1;
}

/*--------------------------------------------------------*/
/* SET INITIAL COLUMNS TO THE RMP                         */
/*--------------------------------------------------------*/
int TSSP_set_initial_columns(
    int aggregated,
    PDCGM  *PDCGM_env,      /* PDCGM environment */
    T_SMPS_DATA* SMPS_data) /* instance data in the SMPS format */
{
    
    int i, j, k, cbeg, cend,
         nr_p1 = (SMPS_data->SMPS_periods)->first_constraint_period[1], /* number of constraints in the first stage (includes OBJ) */
         nc_p1 = (SMPS_data->SMPS_periods)->first_variable_period[1],   /* number of variables in the first stage */
         nr_linking = nc_p1, /* number of LINKING constraints in the master problem (DUAL FORMULATION) */
         nr_convexity, /* number of convexity constraints in the master problem (DUAL FORMULATION) */
         nr_RMP, /* total number constraints in the master problem (DUAL FORMULATION) */
         nc_RMP = nr_p1 - 1 + 1, /* number of variables in the master problem (DUAL FORMULATION) */
         nnz_RMP; /* number of nonzeroes in the master problem */
    
    PDCGM_SMatrix_CW *C = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    PDCGM_SMatrix_RW *R = PDCGM_ALLOC(PDCGM_SMatrix_RW, 1);
    
    /* Set the number of convexity constraints */
    if(aggregated == 1) nr_convexity = 1;
    else nr_convexity = SMPS_data->nS;
    nr_RMP = nr_linking + nr_convexity + 1;
    nnz_RMP = (SMPS_data->A)->nnz + nr_p1;
        
    /* Set the indices of the coeff matrix according to the HOPDM style */
    if( PDCGM_set_SMatrix_RW(R, nr_RMP, nc_RMP, nnz_RMP) ) return -1;
    R->m = nr_RMP;
    R->n = nc_RMP;
    R->nz = nnz_RMP;
    
    /* Set auxiliary counters */
    k = 0;
    cbeg = 0;
    cend = 0;
    
    /* Set the first stage columns as the constraints of the RMP */
    for(i = 0; i < nr_linking; i++) 
    {
        /* Set the initial position of the new row */
        R->rwpnts[i] = k+1; /* Fortran index! */
        
        /* Get the initial position of the CORE column */
        cbeg = (SMPS_data->A)->C_col_ind[i];
        cend = cbeg + (SMPS_data->A)->col_len[i];
        
        /* Copy the first stage coefficients */
        for(j = cbeg; j < cend; j++)
        {
            if( ( (SMPS_data->A)->C_row_ind[j] > 0 ) && ( (SMPS_data->A)->C_row_ind[j] < nr_p1 ) )
            {
                /* It is an entry in a first stage row */
                R->clnmbs[k] = (SMPS_data->A)->C_row_ind[j]; /* Fortran index, BUT remove the first row (OBJ) */
                R->coeff[k]  = (SMPS_data->A)->C_value[j];
                k++;
            }
        }
    }
    
    /* Set the convexity rows as empty */    
    for(i = 0; i < nr_convexity; i++) 
    {
        R->rwpnts[nr_linking + i] = k+1; /* Fortran index! */
    }
    
    /* Set the objective function row, using the RHS of the original formulation */
    R->rwpnts[nr_RMP-1] = k+1; /* Fortran index! */
    for(i = 1; i < nr_p1; i++) /* Starts at 1 to discard de OBJ row of the original formulation */
    {
        if(fabs(SMPS_data->RHS[i]) > TOL_ZERO)
        {
            /* It is an entry in a first stage row */
            R->clnmbs[k] = i; /* Fortran index, BUT remove the first row (OBJ) */
            R->coeff[k] = SMPS_data->RHS[i];
            k++;
        }
    }
    
    /* Set the last position of rwpnts */
    R->rwpnts[nr_RMP] = k+1; /* Fortran index! */
    
    /* Set the actual number of nonzeros in the new coeff matrix */
    R->nz = k;
    
    /* Set the columnwise representation from to the rowwise representation */
    PDCGM_set_SMatrix_CW_from_RW(C, R, nr_RMP, nc_RMP, R->nz);
    
    /* Free up the allocated memory */
    PDCGM_free_SMatrix_RW(R);
    free(R);
    
    /* Call the PDCGM function that add the new columns to the current RMP */
    PDCGM_add_auxiliary_columns(PDCGM_env, C);
    
    /* Free the allocated memory */
    PDCGM_free_SMatrix_CW(C);
    free(C);
    
    /* Set the variable bounds of the initial columns */
    for(i = 1; i < nr_p1; i++) 
    {
        if(SMPS_data->MPS_constraints[i].type == E) 
        {
            /* The dual variable must be FREE */
            PDCGM_env->low_bnd[PDCGM_env->auxiliary_begin + i-1] = -1.e30;
            PDCGM_env->upp_bnd[PDCGM_env->auxiliary_begin + i-1] = 1.e30;
        }
        else if(SMPS_data->MPS_constraints[i].type == G)
        {
            /* The dual variable must be NON-POSITIVE */
            PDCGM_env->low_bnd[PDCGM_env->auxiliary_begin + i-1] = -1.e30;
            PDCGM_env->upp_bnd[PDCGM_env->auxiliary_begin + i-1] = 0.0;
        }
        else if(SMPS_data->MPS_constraints[i].type == L) 
        {
            /* The dual variable must be NON-NEGATIVE */
            PDCGM_env->low_bnd[PDCGM_env->auxiliary_begin + i-1] = 0.0;
            PDCGM_env->upp_bnd[PDCGM_env->auxiliary_begin + i-1] = 1.e30;            
        }
        else
        {
            printf("\n### ERROR! \n");
            return -1;
            
            /* The dual variable must be FIXED */
            PDCGM_env->low_bnd[PDCGM_env->auxiliary_begin + i-1] = 0.0;
            PDCGM_env->upp_bnd[PDCGM_env->auxiliary_begin + i-1] = 0.0;
        }
    }
    
    return 1;
}

/*--------------------------------------------------------*/
/* SOLVE THE DECOMPOSED PROBLEM USING PDCGM               */
/*--------------------------------------------------------*/
int TSSP_solve_decomposed(
    int aggregated,
    T_SMPS_DATA* SMPS_data)
{
    short *row_type = 0;  /* Type of constraints: EQUAL, GREATER_EQUAL, LESS_EQUAL, OBJECTIVE */
    
    double *x_0,  /* Initial solution obtained from the EVP */
           *RHS,   /* Right-hand side of the constraints in the RMP */
           *low_b, /* Lower bound of the first stage variables (dual variables of the RMP) */
           *upp_b; /* Upper bound of the first stage variables (dual variables of the RMP) */

    T_instance_data* instance; /* pointer to the instance data */
    PDCGM *TSSP;  /* PDCGM environment for the TSSP */

    int i, ret, max_outer = 1000,   
         nc_p1 = (SMPS_data->SMPS_periods)->first_variable_period[1],   /* number of variables in the first stage */
         nr_linking = nc_p1, /* number of LINKING constraints in the master problem (DUAL FORMULATION) */
         nr_convexity,       /* number of convexity constraints in the master problem (DUAL FORMULATION) */
         nr_RMP;             /* total number constraints in the master problem (DUAL FORMULATION) */
    
    /* Set the number of convexity constraints */
    if(aggregated > 0) nr_convexity = 1;
    else nr_convexity = SMPS_data->nS;
    nr_RMP = nr_linking + nr_convexity + 1;
    
    /* Check if the first stage constraints are ranged */
    if(SMPS_data->RANGES != NULL) 
    {
        /* THE CURRENT VERSION DOES NOT SUPPORT RANGES ON CONSTRAINTS */
        printf("\nSorry, the current version does not support ranges on constraints.\n");
        return -2;
    }
    
    /* Allocate and initialize the instance data structure */
    instance = (T_instance_data *) malloc( sizeof(T_instance_data) );
    if(instance == NULL) return -1;
    
    instance->SMPS_data = SMPS_data;

    instance->env_sub = (CPXENVptr *) malloc( sizeof(CPXENVptr) * SMPS_data->nS );
    for(i = 0; i < SMPS_data->nS; i++) instance->env_sub[i] = NULL;
    
    instance->lp_sub = (CPXLPptr *) malloc( sizeof(CPXLPptr) * SMPS_data->nS );
    for(i = 0; i < SMPS_data->nS; i++) instance->lp_sub[i] = NULL;
    
    instance->scenario_probability = (double *) malloc( sizeof(double) * SMPS_data->nS );
    if(instance->scenario_probability == NULL) return -1;
    
    instance->h = (STO_SVector_type *) malloc( sizeof(STO_SVector_type) );
    if(instance->h == NULL) return -1;
    
    instance->q = (STO_SVector_type *) malloc( sizeof(STO_SVector_type) );
    if(instance->q == NULL) return -1;
    
    instance->T = (STO_SMatrix_type *) malloc( sizeof(STO_SMatrix_type) );
    if(instance->T == NULL) return -1;
    
    instance->Wt = (STO_SMatrix_type *) malloc( sizeof(STO_SMatrix_type) );
    if(instance->Wt == NULL) return -1;
    
    instance->cost  = NULL;
    instance->RHS   = NULL;
    instance->low_b = NULL;
    instance->upp_b = NULL;
    instance->row_type = NULL;
    
    /* Scale to the original data */
    for(i = 0; i < SMPS_data->nc; i++) SMPS_data->c[i] /= OBJ_SCALING; 
    for(i = 0; i < SMPS_data->nr; i++) SMPS_data->RHS[i] /= RHS_SCALING; 

    /*------------------------------------------------------*/
    /* SET THE SUBPROBLEM                                   */
    /*------------------------------------------------------*/
    
    /* Create the subproblem using the instance data */
    ret = TSSP_set_subproblem(instance);
    if(ret < 0)
    {
        printf("\n### ERROR! The subproblem could not be set.\n");
        return -1;
    }

    /*------------------------------------------------------*/
    /* SET THE MASTER PROBLEM (dual formulation)            */
    /*------------------------------------------------------*/
    
    /* Allocate memory */
    x_0 = (double *) malloc( sizeof(double) * nr_RMP );
    if(x_0 == NULL) return -1;
    
    low_b = (double *) malloc( sizeof(double) * nr_linking );
    if(low_b == NULL) return -1;
    
    upp_b = (double *) malloc( sizeof(double) * nr_linking );
    if(upp_b == NULL) return -1;
    
    row_type = (short *) malloc( sizeof(short) * nr_RMP );
    if(row_type == NULL) return -1;
    
    RHS = (double *) malloc( sizeof(double) * nr_RMP );
    if(RHS == NULL) return -1;
    
    /* ATTENTION: the last row of the coefficient matrix must describe the objective function */
        
    /* Set the type of each linking constraint according to the first stage variables */
    for(i = 0; i < nr_linking; i++) 
    {
        if(SMPS_data->MPS_variables[i].type == PL)
        {
            /* NON-NEGATIVE VARIABLE */
            low_b[i] = 0.0;
            upp_b[i] = BOUND_DUAL;
            
            /* 'GREATER-THAN-OR-EQUAL-TO' CONSTRAINT */
            row_type[i] = GREATER_EQUAL; 
        }
        else if(SMPS_data->MPS_variables[i].type == MI)
        {
            /* NON-POSITIVE VARIABLE */
            low_b[i] = -BOUND_DUAL;
            upp_b[i] = 0.0;
            
            /* 'LESS-THAN-OR-EQUAL-TO' CONSTRAINT */
            row_type[i] = LESS_EQUAL; 
        }
        else if(SMPS_data->MPS_variables[i].type == FR)
        {
            /* FREE VARIABLE */
            low_b[i] = -BOUND_DUAL;
            upp_b[i] = BOUND_DUAL;
            
            /* EQUALITY CONSTRAINT */
            row_type[i] = EQUAL; 
        }
        else 
        {
            /* THE CURRENT VERSION DOES NOT SUPPORT BOUNDED VARIABLES */ 
            printf("\nSorry, the current version does not support bounded variables.\n");
            return -2;
        }
    }
    
    /* Set the type of the convexity constraints (EQUALITY) */
    for(i = nr_linking; i < nr_linking + nr_convexity; i++) row_type[i] = EQUAL;
    
    /* Set the type of the last row: OBJECTIVE FUNCTION */
    row_type[nr_RMP-1] = OBJECTIVE;
    
    /* Set the RHS of the linking constraints */
    for(i = 0; i < nr_linking; i++) RHS[i] = -SMPS_data->c[i];
    
    /* Set the RHS of the convexity constraints */
    for(i = nr_linking; i < nr_linking + nr_convexity; i++) RHS[i] = CNVXTY_VALUE;
    
    /* Set the RHS of the OBJECTIVE function row */
    RHS[nr_RMP-1] = 0.0;
    
    if( TSSP_solve_EVP(NULL, x_0, instance) != 1 )
    {
        /* The initial point could not be generated. Try another guess */
        for(i=0; i<nr_linking; i++) x_0[i] = 0.1;
        printf("\n### ARTIFICIAL INITIAL SOLUTION ###\n");
        fflush(stdout);
    }
    for(i=nr_linking; i<nr_RMP; i++) x_0[i] = 100.0;
    
    /* POPULATE THE INTERNAL DATA STRUCTURE OF PDCGM */
    TSSP = PDCGM_set_data( 
            nr_linking,     /* number of linking constraints in the RMP (DUAL) */
            nr_convexity,   /* number of convexity constraints in the RMP (DUAL) */
            nr_linking + 2, /* maximum number of nonzeros in a column */
            nr_convexity,   /* maximum number of columns generated by the oracle */
            max_outer,      /* maximum number of outer iterations */
            x_0,       /* initial guess of the dual solution (may be NULL) */
            RHS,        /* right-hand side of each constraint in the RMP (DUAL) */
            row_type,   /* type of each constraint (row) in the RMP (DUAL) */
            low_b,      /* lower bound vector of the variables in the RMP (DUAL) */
            upp_b,      /* upper bound vector of the variables in the RMP (DUAL) */
            instance);  /* instance data */
                
    /* Check if the PDCGM environment was created successfully */
    if(TSSP == NULL)   
    {
        printf("-- ERROR: PDCGM_set_data() returned NULL --\n");
        fflush(stdout);
        return 0;
    }
            
    /* Set the optimality tolerance for the column generation algorithm */
    PDCGM_set_delta(TSSP, TSSP_DELTA);
     
    /* Set the degree of optimality (parameter D > 1) */
    PDCGM_set_degree_of_optimality(TSSP, TSSP_DEGOPT);
    
    /* Set epsilon_max */
    PDCGM_set_max_opt_tol(TSSP, 0.1);
    
    /* Set the verbose mode (how much information is printed) */
    PDCGM_set_verbosity(TSSP, 1);
    
    /* Enable column elimination on insertion */
    PDCGM_set_column_elimination(TSSP, 1);
    
    /* Preloaded columns */
    //TSSP_set_initial_columns((aggregated>0), TSSP, SMPS_data);
    TSSP_set_initial_columns_with_artificial((aggregated>0), TSSP, SMPS_data);
    
    //PDCGM_set_max_number_columns(TSSP, 50000);
    //TSSP->use_difficult_iters = 1;

    PDCGM_set_centrality_value(TSSP, 0); /* RELAX THE CENTRALITY REQUIREMENT */
    TSSP->WS_strategy = 0; /* TURN OFF THE WARMSTARTING STRATEGY */
    
    /*------------------------------------------------------*/
    /* CALL THE COLUMN GENERATION PROCEDURE                 */
    /*------------------------------------------------------*/
    if(aggregated == 2) ret = PDCGM_solve_MP(TSSP, TSSP_agg2_oracle);
    else if(aggregated == 1) ret = PDCGM_solve_MP(TSSP, TSSP_agg1_oracle);
    else ret = PDCGM_solve_MP(TSSP, TSSP_disagg_oracle);
    
    #ifdef SAVE_TO_FILE
    /* Save to an output file */
    FILE *out_file = fopen("output-tssp-pdcgm.txt", "a");
    fprintf(out_file, "\t%lf \t%lf \t%1.10E \t%d \t%lf \t%lf \t%lf ", 
        -OBJ_SCALING * RHS_SCALING * TSSP->upperBnd, -OBJ_SCALING * RHS_SCALING * TSSP->lowerBnd, TSSP->rel_gap, 
        TSSP->outer, TSSP->cputime_RMP, TSSP->cputime_oracle, TSSP->cputime_CG);
    fclose(out_file); 
    #endif
    
    /* Free the allocated memory */
    PDCGM_free(TSSP);
    free(TSSP);
    free(x_0);
    free(RHS);
    free(row_type);
    free(low_b);
    free(upp_b);
    
    return 1;
}


/* C Main function */
int main(int arg_c, char *arg_v[])
{ 
    /* Statement of variables */
    char file_name_1[1000], file_name_2[1000], file_name_3[1000];
    int ret, aggregated;
    
    T_SMPS_DATA* SMPS_data = (T_SMPS_DATA*) malloc( sizeof(T_SMPS_DATA) ); /* Instance data (from SMPS files) */
    
    /* Checks if the parameters were typed correctly */
    if(arg_c < 4) 
    {
        printf("\nPlease type: ./SP_solver <COR_file> <TIM_file> <STO_file> \n\n");
        return 0;
    }
    
    if(arg_c < 5) aggregated = 2;
    else aggregated = fabs(atoi(arg_v[4])); 
    
    /* Copies the parameters */
    strcpy(file_name_1, arg_v[1]);
    strcpy(file_name_2, arg_v[2]);
    strcpy(file_name_3, arg_v[3]);
    
    /*----------------------------------------------------------*/
    /* READ THE INSTANCE IN THE SMPS FORMAT                     */
    /*----------------------------------------------------------*/
    
    /* Read the CORE and the STOCHASTIC files */
    if( (ret = readSMPS(file_name_1, file_name_2, file_name_3, SMPS_data) ) >= 0 )
    {
        /* MPS file not found or an error was found in the file... STOP the program! */
        printf("#### ERROR in the MPS file at line %d.\n", ret);
        return 0;
    }

    //printf("\nNumber of rows, columns, nonzeros = %d, %d, %d ", SMPS_data->nr, SMPS_data->nc, SMPS_data->A->nnz);
    printf("\nNumber of INDEP: %d ", SMPS_data->nI);
    printf("\nNumber of BLOCKS: %d ", SMPS_data->nB);
    printf("\n=> Number of scenarios: %d \n\n", SMPS_data->nS);
    
    #ifdef SAVE_TO_FILE
        FILE *out_file = fopen("output-tssp-pdcgm.txt", "a");
        //fprintf(out_file, "\n%s \t%s \t%s \t%d \t%d \t%d \t%d \t%d ", 
        //    file_name_1, file_name_2, file_name_3, SMPS_data->nc, SMPS_data->nr, SMPS_data->nI, SMPS_data->nB, SMPS_data->nS);
        fprintf(out_file, "\n%s \t%d ", file_name_1, SMPS_data->nS);
        fflush(out_file);
        fclose(out_file); 
    #endif
    
    /* Solve the Decomposed Problem */
    TSSP_solve_decomposed(aggregated, SMPS_data);
    
    /* Free up the allocated memory */   
    SMatrix_freeMatrix(SMPS_data->A);
    free(SMPS_data->c);
    free(SMPS_data->RHS);
    free(SMPS_data->RANGES);    
    free(SMPS_data->MPS_constraints);
    free(SMPS_data->MPS_variables);
      
    return 0;
}

