/* ---------------------------------------------------------------------------
 * rcespp (Resource contrained elementary shortest path problem)
 * ---------------------------------------------------------------------------
 * Author: Pedro Munari (pedro.munari@gmail.com)
 * University of Sao Paulo, Brazil
 * Last update: Oct/2012
 * ---------------------------------------------------------------------------
 */

#ifndef MYLIB_RCESPP
#define MYLIB_RCESPP

/* Bring in the auxiliary libraries */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <omp.h>  

#include "priority_queue.h"
#include "vrptw_SRC.h"
#include "rcespp_states.h"
#include "rcespp_priority_queue.h"

typedef struct
{
	RCESPP_state_type *states; /* a buffer of states */
	int n_states;    	/* current number of states already used */
	int n_states_max;   /* total number of states allocated in memory */

	RCESPP_state_type **old_buffers; /* full buffers */
	int n_old_buffers;    	/* current number of old buffers */
	int n_old_buffers_max;  /* maximum number of old buffers */
	
	RCESPP_state_type **deleted_states; /* a buffer of states that can be discarded */
	int n_deleted_states;    	/* current number of deleted states */
	int n_deleted_states_max;   /* maximum number of deleted states */
	
	int nN;
	int SRC_n;
}
RCESPP_states_buffer_type;


/* Data structure to store the list of the best paths found */
/* A priority queue is used to keep them sorted by the obj value */
typedef struct
{
	/* Priority queue of the best paths. It stores the objective value
	   and the first position of the path in the array 'ind_paths' */
    PriorityQueue_type *queue;
	
	/* Indices of the best paths. The first entry of each path is its cost.
	   The first entry after the indices in the path is either 0 (in case it is
	   an elementary path) or -1 (in case there is at least a cycle in the path). */
    int *ind_paths;
    
    /* Total number of paths */
    int n_paths;
    
    /* Number of indices in 'ind_paths' (and its maximum) */
    int n_indices;
    int n_indices_max;    
}
RCESPP_paths_list_type;

/* Data to structure to store the instance data and auxiliary information */
typedef struct
{
	/* Pointers to the instance data */
	int **distance;		/* distance between any two customers */
	int *demand;		/* demand at each customer */
	int *window_open;   /* instant when the time window opens at each customer */
	int *window_close;	/* instant when the time window closes at each customer  */
	int *service_time;	/* time spent in a customer */
	int q;	 /* maximum capacity of each vehicle */
	int nN;	 /* total number of nodes in the problem (including source and sink) */
	double T_max;  /* maximum feasible arrival time at the sink */
	double T_avg;  /* maximum feasible arrival time at the sink */

	int *dmin;
    int *rem_prizes_ind;
	
	/* Adjacency list (graph): they are used to express the 
	   forward and backward connection among the nodes */	   
	int *adjacency_index_fw;   /* pointer to the adjacent vertices of each vertex (nN+1 entries), 
								  to be used in the forward extension */
   	int *adjacency_vertex_fw;  /* adjacent vertices (the number of entries for each index is computed 
   						          according to the difference between two consecutive indices) */
	int *adjacency_index_bw;   /* pointer to the adjacent vertices of each vertex (nN+1 entries), 
								  to be used in the backward extension */
   	int *adjacency_vertex_bw;  /* adjacent vertices (the number of entries for each index is computed 
   						          according to the difference between two consecutive indices) */
   	
   	/* Data structure for handling SRC (subset row cuts)  */
   	VRPTW_SRC_CS_data_type *cuts_SRC;
   	
   	/* Data structure that stores the resulting routes */
   	RCESPP_paths_list_type *routes;
   	
   	/* Priority queue of the best routes that are stored in 'routes'. It stores the 
	   objective value and the first position of the route in the array 'routes->ind_paths' */
    PriorityQueue_type *routes_queue;
   	
	/* A copy of the last used prizes vector */
   	double *old_prizes;
	
	/* ordered sets of states to be extended (sorted by reduced cost) */
    RCESPP_PriorityQueue_type *fresh_states_fw;
    RCESPP_PriorityQueue_type *fresh_states_bw;
    
    /* Buffer of states allocated in memory */
    RCESPP_states_buffer_type *states_buffer_fw;
    RCESPP_states_buffer_type *states_buffer_bw;
    
	/* Data structure that stores the parameters of the method */
	int method;
	int phase;
	int presolver;
	int A_MAX;
	int R_MAX;
	int S_MAX;
	int H_MAX;
	int RCH_DP;
	int SRC_MAX;
	int DP_USE_BOUNDS;
	int DP_INIT_MAX;
	int DP_INIT_SUGG;
	double DP_INIT_FACTOR;
	
	int nS_fw;
	int nS_bw;
	
	int n_calls_method_2;
	int n_calls_method_3;
	int n_calls_method_6;
}
RCESPP_data_type;
 
/* Data structure for the route construction heuristics (RCH) */
typedef struct
{
	int n_routes; 	     /* number of routes in the solution */
	int *first_customer; /* first_customer[i] = first customer of route i */
	int *capacity;		 /* capacity[i] = amount of vehicle capacity used in route i */
	int *length;    	 /* length[i] = number of customers in route i */
	double *cost;		 /* cost[i] = total cost of route i */
	
	double *time;		  /* time[i] = instant that the vehicle arrives at customer i */
	int *route_customers; /* sequence of customers in a route, starting from each 'first_customer' */
						  /* route_customers[i]: the next customer to be visited after customer i */
	
	int *SRC_set; /* number of nodes of a route that belongs to a set S_q of the SRC cuts */
	
	/* REMARK: customers are indexed from 1 to nN-2 */
}
RCESPP_RCH_routes_type;

/* Memory allocation for the data structure of states lists */
int RCESPP_allocate_states_lists(
    RCESPP_states_lists_type *lists,  /* pointer to the states lists */
    int n,  /* total number of nodes to be visited */
    double objective_threshold); /* initial values for the best costs */

RCESPP_state_type* RCESPP_allocate_state(
	int n, 		/* number of nodes in the problem */
	int SRC_n);  /* number of SRC sets */

RCESPP_paths_list_type* RCESPP_allocate_paths_list(
    int n); /* number of paths in the structure (a guess) */

int RCESPP_reset_paths_list(RCESPP_data_type* RCESPP_env);

/* Free up memory */
int RCESPP_free_states_lists(
    RCESPP_states_lists_type *lists); /* pointer to the states lists */

/* Free up memory */
int RCESPP_free_paths_list(RCESPP_data_type* RCESPP_env);

/* Memory allocation */
RCESPP_data_type* RCESPP_set_data(
	int **distance,		/* distance between any two customers */
	int *demand,		/* demand at each customer */
	int *window_open,   /* instant when the time window opens at each customer */
	int *window_close,	/* instant when the time window closes at each customer  */
	int *service_time,	/* time spent in a customer */
    int q,	 			/* maximum capacity of each vehicle */
    int nN,			/* total number of nodes in the problem (including source and sink) */
    double T_max,
    double T_avg);

/* Free up memory */
int RCESPP_free_data(RCESPP_data_type *pointer);

/* Set the adjacency lists of the vertices */
int RCESPP_set_adjacency_list(
    RCESPP_data_type *data,  /* instance data */
    double *prizes,
    int *deleted_arcs,
    int deleted_arcs_n);

/* Set the arc (i,j) as deleted */
int RCESPP_delete_arc(
    RCESPP_data_type *data,  /* instance data */
    int vertex_from,         /* i: vertex from which the arc departs */
    int vertex_to,           /* j: vertex in which the arc arrives */
    int **deleted_arcs,
    int *deleted_arcs_n,
    int *deleted_arcs_n_max);

/* Convert the next stored path in a solution of the problem. The solution is given in 
   the format of a column, to be used by the oracle in the column generation procedure. */
int RCESPP_get_next_solution(
    double *objective,             /* value of the solution */
    int    *solution,   		   /* solution vector (already allocated in memory) */
    double *path_cost,			   /* real cost of the path (without prizes) */
    int *RMP_paths,				   /* Array of paths used to create a column of the RMP */
    int *RMP_paths_n,		   	   /* Number of paths in the array 'RMP_paths' */
    int **RMP_paths_indices,	   /* Pointer to the array of indices of the nodes in the paths */
    int *RMP_paths_indices_n_max,  /* Maximum number of indices in the array 'RMP_paths_indices' */
    int discard_similar,
    double MAX_objective,
    double *prizes,	    /* prizes at nodes (dual solution of the current RMP) */
    double *SRC_duals,
    RCESPP_data_type *RCESPP_env);     /* instance data */

int RCESPP_RCH_solve(
    double *objective,  /* value of the optimal solution */
    double *prizes,	    /* prizes at nodes (dual solution of the current RMP) */
    double *old_prizes,	/* prizes at nodes in the last calling (dual solution of the last RMP) */
    double *SRC_duals,	    	 /* dual solution of the RMP that is associated to the SR constraints */
    RCESPP_data_type *RCESPP_env,   /* instance data */
    double objective_threshold,
    int max_n_paths);

int RCESPP_multiphases_solver(
	RCESPP_data_type *RCESPP_env, /* environment with data and parameters */
	double rel_gap,	    /* relative gap in the current RMP */
	double delta,	    /* optimality tolerance of the CG  */
	int ncols_last,
	int n_cols,
	int n_cuts, 
    double *objective, 	/* value of the optimal solution */
    double objective_threshold,  
    double *path_cost,	/* real cost of the optimal path (without prizes) */
    double *prizes);    /* prizes at nodes (dual solution of the current RMP) */

int RCESPP_multiphases_clean(
	RCESPP_data_type *RCESPP_env); /* environment with data and parameters */
	
int RCESPP_nonelementary_solver(
	RCESPP_data_type *RCESPP_env, /* environment with data and parameters */
	double rel_gap,	    /* relative gap in the current RMP */
	double delta,	    /* optimality tolerance of the CG  */
	int outer,
	int n_cols,
	int n_cuts, 
    double *objective, 	/* value of the optimal solution */
    double objective_threshold,
    double *path_cost,	/* real cost of the optimal path (without prizes) */
    double *prizes);	    /* prizes at nodes (dual solution of the current RMP) */

#endif

