/* ---------------------------------------------------------------------------
 * File: rcespp_priority_queue.h (Priority queue)
 * ---------------------------------------------------------------------------
 * Author: Pedro Munari (pedro.munari@gmail.com)
 * Last update: Nov/2011
 * ---------------------------------------------------------------------------
 */

/* Data structure:

		       0 1                     (n)                    (n_max)
		      ----------------------------------------------------
	 values   | |**********************|                         |   
		      ----------------------------------------------------
	 pointers | |**********************|                         |
		      ----------------------------------------------------

   HEAP CONDITION: The key in each node should be larger than (or equal to)
                   the keys in its children.
                   - Parent of node j: j/2 (rounded down)
                   - Children of node j: 2j and 2j + 1.
 
   The zero position of the arrays is empty, in order to make easier 
   the relationship between parent/children indices

*/

/* Bring in the auxiliary libraries */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Bring in the header */
#include "rcespp_states.h"

//#ifndef MYLIB_RCESPP_PriorityQueue
//#define MYLIB_RCESPP_PriorityQueue

/* Data structure for the priority queue */
typedef struct
{
    int n_max;       /* maximum number of entries in the queue */
    int n;           /* number of entries in the queue */

	/* array with the priority values */
    double *values;
    
    /* array with information about the entry */
    int *info;
    
    /* array with the pointers to the nodes associated to the values */
    RCESPP_state_type **pointers;    
    
} RCESPP_PriorityQueue_type;

/* Allocate the priority queue */
int RCESPP_PriorityQueue_allocate(
    RCESPP_PriorityQueue_type *queue,   /* Pointer to the queue */
    int n_max);   /* Maximum number of entries in the queue */

/* Reallocate the priority queue */
int RCESPP_PriorityQueue_reallocate(
    RCESPP_PriorityQueue_type *queue,   /* Pointer to the queue */
    int n_max);   /* New maximum number of entries in the queue */

/* Insert a new entry in the priority queue */
int RCESPP_PriorityQueue_insert(
    RCESPP_PriorityQueue_type *queue,  /* Pointer to the queue */
    double value,  /* Value of the entry to be inserted */
    int info,
    RCESPP_state_type *pointer);  /* Pointer associated to the entry to be inserted */

/* Remove the top of the priority queue */
int RCESPP_PriorityQueue_getNext(
    RCESPP_PriorityQueue_type *queue, /* Pointer to the queue */
    double *value, /* Value of the entry to be inserted */
    int *info,
    RCESPP_state_type **pointer); /* Pointer associated to the entry to be inserted */

/* Peak at the top of the priority queue */
double RCESPP_PriorityQueue_peakTop(RCESPP_PriorityQueue_type *queue);   /* Pointer to the queue */

/* Free the memory of the priority queue */
int RCESPP_PriorityQueue_free(RCESPP_PriorityQueue_type *queue);

//#endif

