/* ---------------------------------------------------------------------------
 * File: vrptw_SRC.h (Subset row cuts for the VRPTW)
 * ---------------------------------------------------------------------------
 * Author: Pedro Munari (pedro.munari@gmail.com)
 * University of Sao Paulo, Brazil
 * Last update: Mar/2011
 * ---------------------------------------------------------------------------
 */
 
#ifndef MYLIB_VRPTW_SRC
#define MYLIB_VRPTW_SRC

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> 

#include "priority_queue.h"

/* Maximum number of sets to be generated by a call to SRC */
#define VRPTW_SRC_max_S_n 10000

/* Data to structure to store information about the sets */
typedef struct
{	
	/* Sets corresponding to generated cuts that were inserted in the matrix */
	/* The order of the sets follow the order of insertion in the matrix */
	int *S_C;   /* Lists containing three customers that belong to the same set */
	int S_C_n;	 /* Number of sets (smaller than or equal to the maximum number of cuts) */	
}
VRPTW_SRC_data_type; 

/* Data to structure to store information about customers in the sets */
typedef struct
{	
	/* Inverse of S_C: For each customer, it stores the list of sets containing the customer */
	/* We need one more array, as the number of times a customer belongs to a set is not fixed */
	int *C_ptr; /* entry i: pointer to the list of sets that contain the customer i */
	int *C_S;   /* Lists of sets that contain a given customer */
	int C_S_n;	 /* Number of entries in C_S */	
	int C_S_n_max;	 /* Maximum number of entries in C_S */	
	int S_n;	 /* Number of sets considered in the generation of list C_S */
}
VRPTW_SRC_CS_data_type;

/* Data to structure to store internal information */
typedef struct
{	
	/* Sets of (three) customers and associated paths which are used to generate the cuts */
	int *S; 		/* entry i: Pointer to a list of customers and associated paths of a set S_i = {i,j,k} */
	int *S_C_P;    /* lists of customers and paths corresponding to a given set. The first three positions 
	  				   corresponds to the indices of the set. The remaining entries are the paths the visit
	  				   at least two customers in the set. */
	int S_n;		/* Number of generated sets */
	int S_n_max;	/* Maximum number of generated sets */
	int S_C_P_max;	/* Maximum number of entries in the array of lists */

	/* Priority queue of the best sets S={i,j,k}. It stores the violation of the 
	   corresponding cut and the index of the set */
    PriorityQueue_type *best_sets;
}
VRPTW_SRC_internal_data_type;

#ifndef PDCGM_MATRIX_H
#define PDCGM_MATRIX_H
typedef struct PDCGM_SMatrix_CW_struct {
     int      m;
     int      n;
     int      nz;
     int      max_m;
     int      max_n;  
     int      max_nz;
     double    *coeff;
     long      *rwnmbs;
     long      *clpnts;
     char      *rwname;
     char      *clname;
} PDCGM_SMatrix_CW;

typedef struct PDCGM_SMatrix_RW_struct {
     int      m;
     int      n; 
     int      nz;
     int      max_m;
     int      max_n;
     int      max_nz;
     double    *coeff;
     long      *rwpnts;
     long      *clnmbs;
     char      *rwname;
     char      *clname;
} PDCGM_SMatrix_RW;
#endif

/*---------------------------------------------------------------------------*/
/* VRPTW_SRC_separate()														 */
/* Find and store all the subset row cuts (SRC) corresponding to the current */
/* solution (received as a parameter) and sort them all according to the	 */
/* violation of the associated inequality									 */
/*---------------------------------------------------------------------------*/
int VRPTW_SRC_separate(
	VRPTW_SRC_internal_data_type *cuts,	/* ( ->) Solver data */
	PDCGM_SMatrix_RW *M_RW, 	/* (<->) Coefficient matrix of the MP in the rowwise format */
	double *solution,			/* ( ->) Solution of the master problem (in terms of master variables) */
	int nC,						/* ( ->) Number of customers in the instance (without sink and source) */
	double violation_threshold);

/* Function: VRPTW_SRC_set_C_S() */
/* Create the inverse of S_C list: The relationship is indexed by the customers */
/* For each customer, there is a pointer to a list of sets that contain the customer */
/* It is useful in the RCESPP solver, for instance. */
int VRPTW_SRC_set_C_S(
	VRPTW_SRC_CS_data_type *list, /* (<- ) Information about which customer belongs to each set */
	int *indices,				  /* ( ->) Indices of the sets that must be used to generate the list */
	VRPTW_SRC_data_type *sets,	  /* ( ->) Solver data containing the generated sets */
	int nS,						  /* ( ->) Number of sets in 'indices' */
	int nC);				      /* ( ->) Number of customers in the instance (without sink and source) */

/* Given a fractionary solution of the master problem, this function */
/* finds and stores the subset row cuts (SRC) corresponding to this solution */
int VRPTW_SRC_set_cuts(
	int max_number_of_cuts,		/* ( ->) Maximum number of cuts to be generated */
	VRPTW_SRC_data_type *sets,	/* ( ->) Solver data */
	PDCGM_SMatrix_RW *M_RW, 	/* (<->) Coefficient matrix of the MP in the rowwise format */
	PDCGM_SMatrix_RW *M_cuts, 	/* (<->) Coefficient matrix of the MP in the rowwise format */
	double *solution,			/* ( ->) Solution of the master problem (in terms of master variables) */
	double **RHS, 				/* (<- ) Right-hand side of the system of equations */
	long  **rw_type, 			/* (<- ) Type of each row in the coefficient matrix */
	double **lo_box,			/* (<- ) Lower bound of the dual variables associated to the each constraint */
	double **up_box,			/* (<- ) Upper bound of the dual variables associated to the each constraint */
	int **rows_links,			/* (<- ) Links associated to each generated cut */
	double *primal_violation,   /* (<- ) Total violation in the generated cuts */
	int nC,						/* ( ->) Number of customers in the instance (without sink and source) */
	int max_nCinS,
	int max_nCall,
	double violation_threshold);

int VRPTW_SRC_free(
	VRPTW_SRC_data_type *sets);	/* ( ->) Solver data */

int VRPTW_SRC_CS_free(
	VRPTW_SRC_CS_data_type *CS);	/* ( ->) Solver data */

#endif

