/*
 *  COLUMN GENERATION PROCEDURE FOR THE 
 *  [CAPACITATED] VEHICLE ROUTING PROBLEM WITH TIME WINDOWS (VRPTW)
 *  [Aggregated, k-best columns]
 *
 *  Author: Pedro Munari (pedro.munari@gmail.com)
 *  University of Sao Paulo, Brazil
 *  Last modified: July, 2012
 *
 *
 *  ON TOP OF HOPDM/PDCGM
 *  Written by:    Jacek Gondzio
 *                 Department of Mathematics and Statistics,
 *                 The University of Edinburgh,
 *                 Mayfield Road, King's Buildings.
 *                 EH9 3JZ Edinburgh, Scotland.
 *  on leave from: Systems Research Institute, 
 *                 Polish Academy of Sciences, 
 *                 Newelska 6, 01-447 Warsaw, Poland.
 */
 
/* Bring in the standard C library files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <omp.h>  

/* Bring in the PDCGM/HOPDM library files */
#include "../../pdcgm/pdcgm.h"  

/* Bring in the elementary shortest path problem solver (RCESPP) */
#include "include/rcespp.h" 

/* Flag: Save to a file when active */ 
#define SAVE_TO_FILE

/* Optimality tolerance */
#define VRPTW_DELTA 1.E-6

/* Tolerance to zero */
#define VRPTW_TOL_ZERO 1.E-6

/* Maximum number of columns that can be generated at a time */
int MAX_NUMBER_COLS_IN_A_CALL = 100;

/* Data structure to store the routes in a solution */
typedef struct
{
    /* Save all the routes that were used to create the columns in the RMP */
    /* We need to store them in order to write the optimal solution in terms of the original variables  */
    int *route; /* Array of pointers to the routes described in 'route_indices' */
    int *route_indices; /* Indices of the nodes in a route */   
        
    int nr; /* Number of routes in the array 'route' */
    int nr_max; /* Maximum number of routes in the array 'route' */
    int ni_max; /* Maximum number of indices in the array 'route_indices' */
}
T_VRPTW_routes;

/* Data structure for the instance data */
typedef struct
{
    /* Instance data read from a solomon's instance */
    int **distance;     /* distance between any two customers */
    int *demand;        /* demand at each customer */
    int *window_open;   /* instant when the time window opens at each customer */
    int *window_close;  /* instant when the time window closes at each customer  */
    int *service_time;  /* time spent in a customer */
    int q;   /* maximum capacity of each vehicle */
    int nC;  /* number of costumers */
    int nV;  /* number of vehicles */
    int nN;  /* total number of nodes in the problem (including source and sink) */
    double T_max;  /* maximum feasible arrival time at the sink */
    double T_avg;  
    
    /* Data structure of the subproblem solver */
    /* This variable is put here because we use it inside the oracle */
    RCESPP_data_type* RCESPP_data; 

    /* Routes that were used to create the columns in the RMP */
    /* We need to store them in order to write solution in the original variables  */
    T_VRPTW_routes *routes;
    
    /* auxiliar array of integers to be used inside the oracle */  
    int *buffer_int; 
    
}T_instance_data;


/* Reduce time windows in order to reduce the number of states */
/* Apply the reductions proposed by Desrochers et al. (1992) */
int VRPTW_reduce_time_windows(T_instance_data *instance) /* instance data */
{
    /* Statement of auxiliar variables */
    int i, k, aux_1, it, changed;
    
    /* Apply the reductions while changes have been applied */
    it = 0;
    changed = 1;
    while( (changed == 1) && (it < 5) )
    {
        /* No change was made yet */
        changed = 0;
        
        /* Minimum arrival time from predecessors */
        for(k=1; k < instance->nN - 1; k++)
        {
            /* Find the minimum between b_k and min_i{a_i + t_ik} */
            aux_1 = instance->window_close[k];
            for(i=0; i < instance->nN; i++)
            {
                if( (i != k) && (aux_1 > instance->window_open[i] + instance->distance[i][k] + instance->service_time[i]) )
                    aux_1 = instance->window_open[i] + instance->distance[i][k] + instance->service_time[i];
            }
            /* Get the maximum between a_k and aux_1 */
            if( aux_1 > instance->window_open[k] ) 
            {
                /* Change the window */
                instance->window_open[k] = aux_1;
                changed = 1;
            }
        }
        
        /* Minimum arrival time to successors */
        for(k=1; k < instance->nN - 1; k++)
        {
            /* Find the minimum between b_k and min_i{a_i - t_ik} */
            aux_1 = instance->window_close[k];
            for(i=0; i < instance->nN; i++)
            {
                if( (i != k) && (aux_1 > instance->window_open[i] - instance->distance[k][i] - instance->service_time[i]) )
                    aux_1 = instance->window_open[i] - instance->distance[k][i] - instance->service_time[i];
            }
            /* Get the maximum between a_k and aux_1 */
            if( aux_1 > instance->window_open[k] ) 
            {
                /* Change the window */
                instance->window_open[k] = aux_1;
                changed = 1;
            }
        }
        
        /* Maximal departure time from predecessors */
        for(k=1; k < instance->nN - 1; k++)
        {
            /* Find the maximum between a_k and max_i{b_i + t_ik} */
            aux_1 = instance->window_open[k];
            for(i=0; i < instance->nN; i++)
            {
                if( (i != k) && (aux_1 < instance->window_close[i] + instance->distance[i][k] + instance->service_time[i]) )
                    aux_1 = instance->window_close[i] + instance->distance[i][k] + instance->service_time[i];
            }
            /* Get the minimum between b_k and aux_1 */
            if( aux_1 < instance->window_close[k] ) 
            {
                /* Change the window */
                instance->window_close[k] = aux_1;
                changed = 1;
            }
        }

        /* Maximal departure time to successors */
        for(k=1; k < instance->nN - 1; k++)
        {
            /* Find the maximum between a_k and max_i{b_i + t_ik} */
            aux_1 = instance->window_open[k];
            for(i=0; i < instance->nN; i++)
            {
                if( (i != k) && (aux_1 < instance->window_close[i] - instance->distance[k][i] - instance->service_time[i]) )
                    aux_1 = instance->window_close[i] - instance->distance[k][i] - instance->service_time[i];
            }
            /* Get the minimum between b_k and aux_1 */
            if( aux_1 < instance->window_close[k] ) 
            {
                /* Change the window */
                instance->window_close[k] = aux_1;
                changed = 1;
            }
        }
        
        /* Update the number of iterations */
        it++;
    }
    return 1;
}


/* READ AN INSTANCE FROM THE FILE SPECIFIED AS PARAMETER */
int VRPTW_read_instance(   
    char file_name[],  /* name of the file that describes the instance */
    T_instance_data *instance)   /* pointer to the instance data */
{

    /* Instances from the problem sets:
       - http://www.fernuni-hagen.de/WINF/touren/inhalte/probinst.htm
       - http://w.cba.neu.edu/~msolomon/problems.htm (http://www2.imm.dtu.dk/~jla/solomon.html)
       
       "Associated with each arc (i,j) is a cost c_ij and a duration t_ij. 
        We assume that the service time at the costumer i is included
        in the duration of the arc (i,j).[Desrochers et al., 1992]
       
       "Travel time between customers are equal to the corresponding distances 
        truncated to one decimal place. All values are then multiplied by 10 to
        obtain integers." [Desrochers et al., 1992]
       
     */

    /* Statement of variables */
    FILE *myfile;
    char *ret, current_line[101];
    int i, j, aux, x, y,
        *xcoord, *ycoord; /* coordinates of the location of each customer */
    
    /* Open the file */
    myfile = fopen(file_name, "r");
    if(myfile == NULL) return 0;
    
    /* Read (and discard) the first four lines of the file (labels) */
    for(i=0;i<4;i++) ret = fgets(current_line, 100, myfile);
    
    /* Read the values nV and q (max number of vehicles and capacity) */
    aux = fscanf(myfile, "%d %d", &(instance->nV), &(instance->q));
    
    /* Guess the number of nodes (it is not explictly provided in the input file) */
    instance->nN = 102;

    /* Memory allocation */
    xcoord = (int *) malloc(sizeof(int) * (instance->nN));
    if(xcoord == NULL) return -1;
    ycoord = (int *) malloc(sizeof(int) * (instance->nN));
    if(ycoord == NULL) return -1;
    instance->demand = (int *) malloc(sizeof(int) * (instance->nN));
    if(instance->demand == NULL) return -1;
    instance->window_open = (int *) malloc(sizeof(int) * (instance->nN));
    if(instance->window_open == NULL) return -1;
    instance->window_close = (int *) malloc(sizeof(int) * (instance->nN));
    if(instance->window_close == NULL) return -1;
    instance->service_time = (int *) malloc(sizeof(int) * (instance->nN));
    if(instance->service_time == NULL) return -1;
    
    /* Read (and discard) the next four lines of the file (labels) */
    for(i=0;i<5;i++) ret = fgets(current_line, 100, myfile);
    
    /* Read the length of the items */
    i=0;
    while( !feof(myfile) )
    {
        /* Read a line from the file */
        if( fscanf(myfile, "%d", &j) > 0 )
        {
            aux = fscanf(myfile, "%d %d %d %d %d %d", &(xcoord[i]), 
                &(ycoord[i]), &(instance->demand[i]), &(instance->window_open[i]), 
                &(instance->window_close[i]), &(instance->service_time[i]));
            
            /* Increment the number of read customers */
            i++;
        
            /* Check if the allocated memory is enough */
            if( i >= instance->nN )
            {
                /* Reallocated arrays */
                instance->nN *= 2;
                xcoord = (int *) realloc(xcoord, sizeof(int) * instance->nN);
                if(xcoord == NULL) return -1;
                ycoord = (int *) realloc(ycoord, sizeof(int) * instance->nN);
                if(ycoord == NULL) return -1;
                instance->demand = (int *) realloc(instance->demand, sizeof(int) * instance->nN);
                if(instance->demand == NULL) return -1;
                instance->window_open  = (int *) realloc(instance->window_open, sizeof(int) * instance->nN);
                if(instance->window_open == NULL) return -1;
                instance->window_close = (int *) realloc(instance->window_close, sizeof(int) * instance->nN);
                if(instance->window_close == NULL) return -1;
                instance->service_time = (int *) realloc(instance->service_time, sizeof(int) * instance->nN);
                if(instance->service_time == NULL) return -1;
            }
        }
    }
    
    /* Set the number of nodes and customers */
    instance->nN = i+1;
    instance->nC = i-1;
    
    /* Create a dummy copy of the depot in the last position */
    xcoord[instance->nN-1] = xcoord[0];
    ycoord[instance->nN-1] = ycoord[0];
    instance->demand[instance->nN-1] = instance->demand[0];
    instance->window_open[instance->nN-1] = instance->window_open[0];
    instance->window_close[instance->nN-1] = instance->window_close[0];
    instance->service_time[instance->nN-1] = instance->service_time[0];
    
    /* Reallocate arrays in order to free unused memory */
    xcoord = (int *) realloc(xcoord, sizeof(int) * instance->nN);
    ycoord = (int *) realloc(ycoord, sizeof(int) * instance->nN);
    instance->demand = (int *) realloc(instance->demand, sizeof(int) * instance->nN);
    instance->window_open  = (int *) realloc(instance->window_open, sizeof(int) * instance->nN);
    instance->window_close = (int *) realloc(instance->window_close, sizeof(int) * instance->nN);
    instance->service_time = (int *) realloc(instance->service_time, sizeof(int) * instance->nN);
    
    /* Close the file */
    fclose(myfile);
    
    /* Create the matrix of distances */
    instance->distance = (int **) malloc(sizeof(int *) * instance->nN);
    if(instance->distance == NULL) return -1;
    for(i=0; i<instance->nN; i++) instance->distance[i] = (int *) malloc(sizeof(int) * instance->nN);

    /* Fill the matrix of distances */
    for(i=0; i<instance->nN; i++) 
    {
        for(j=0; j<instance->nN; j++) 
        {
            x = xcoord[i] - xcoord[j];
            y = ycoord[i] - ycoord[j];
            instance->distance[i][j] = (int) ( 10 * sqrt( x*x + y*y ) );            
        }
        instance->window_open[i] *= 10; 
        instance->window_close[i] *= 10;
        instance->service_time[i] *= 10;
    }

    /* Free auxiliar arrays */
    free(xcoord);
    free(ycoord);
    
    /* Pre-solve for the time windows */
    VRPTW_reduce_time_windows(instance);
    
    /* Computes T_max: the maximum feasible arrival time at the sink */
    instance->T_max = instance->window_close[0] + instance->service_time[0]; /* from source */
    instance->T_avg = 0.0;
    for(i=1; i < instance->nN-1; i++)
    {
        instance->T_avg += instance->window_close[i] - instance->window_open[i];
        if( instance->T_max < instance->window_close[i] + instance->service_time[i] + instance->distance[i][instance->nN-1] )
            instance->T_max = instance->window_close[i] + instance->service_time[i] + instance->distance[i][instance->nN-1];
    }
    instance->T_avg /= 1.0 *(instance->nC);
    
    /* Return successfully */
    return 1;
}

/* Set an auxiliary data structure to store information about routes, cuts, and temporary buffers */
int VRPTW_set_auxiliary_data(
    T_instance_data *instance,  /* (<- ) Pointer to the instance */
    int nr_max,                 /* ( ->) Maximum number of routes in the list of generated routes (may be a guess) */
    int ni_max)                 /* ( ->) Maximum number of indices (nodes) in the list of generated routes (may be a guess) */
{
    
    /* Create the environment of the subproblem solver: RCESPP */
    instance->RCESPP_data = RCESPP_set_data(instance->distance, instance->demand, instance->window_open, 
        instance->window_close, instance->service_time, instance->q, instance->nN, instance->T_max, instance->T_avg);
    
    /* Allocate memory for the array of generated routes  */
    instance->routes = (T_VRPTW_routes *) malloc( sizeof(T_VRPTW_routes) );
    instance->routes->nr = 0;
    instance->routes->nr_max = nr_max; /* a guess */
    instance->routes->ni_max = ni_max;  /* another guess */
    
    instance->routes->route = (int *) malloc( sizeof(int) * (nr_max + 1) );
    if( instance->routes->route == NULL ) return -1;
    instance->routes->route[0] = 0;
    
    instance->routes->route_indices = (int *) malloc( sizeof(int) * (ni_max) );
    if( instance->routes->route_indices == NULL ) return -1;
    instance->routes->route_indices[0] = 0;
    
    /* Allocate the auxiliary arrays (buffers) */
    instance->buffer_int = (int *) calloc((instance->nC + MAX_NUMBER_COLS_IN_A_CALL) , sizeof(int));
    if(instance->buffer_int == NULL) return -1;
    
    return 1;
}

/* Free up the memory allocated to the instance data */
int VRPTW_free_data(
    T_instance_data *instance)   /* pointer to the instance data */
{
    int i;
   
    if(instance == NULL) return 0;
    if(instance->demand != NULL) free(instance->demand);
    if(instance->window_open != NULL) free(instance->window_open);
    if(instance->window_close != NULL) free(instance->window_close);
    if(instance->service_time != NULL) free(instance->service_time);
    if(instance->distance != NULL) 
    {
        for(i=0; i<instance->nN; i++) free(instance->distance[i]);
        free(instance->distance);
    }
    if(instance->routes != NULL) 
    {
        if(instance->routes->route != NULL) free(instance->routes->route);
        instance->routes->route = NULL;
        
        if(instance->routes->route_indices != NULL) free(instance->routes->route_indices);
        instance->routes->route_indices = NULL;
        
        free(instance->routes);
        instance->routes = NULL;
    }
    if(instance->RCESPP_data != NULL) 
    {
        RCESPP_free_data(instance->RCESPP_data);
        free(instance->RCESPP_data);
        instance->RCESPP_data = NULL;
    }
    if(instance->buffer_int != NULL) 
    {
        free(instance->buffer_int);
        instance->buffer_int = NULL;
    }
    
    /* Return successfully */
    return 1;
}

/* Set initial columns of the RMP using *single-customer* routes */
int VRPTW_set_initial_columns(
    PDCGM  *PDCGM_env,         /* PDCGM environment */
    T_instance_data *instance) /* Pointer to the instance data */
{
    int i, j, k, t, nc, max_m, max_n, max_nz, ret, 
        count, route_len, *columns_links=NULL, *solution;
    PDCGM_SMatrix_CW *M;  /* Matrix with the generated column(s) */
    double obj_value, route_cost, route_red_cost;

    /* Allocate the auxiliary matrix in memory */
    M = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    if(M == NULL) return -1;
    max_m = instance->nC + 2; /* number of rows in the RMP:
            (number of customers + convexity + obj function) */ 
    max_n = instance->nC;
    max_nz = max_n * max_m;
    PDCGM_set_SMatrix_CW(M, max_m, max_n, max_nz);
    
    /* Allocate the array of columns links */
    columns_links = PDCGM_ALLOC(int, max_n);
    
    /* Set initial dimensions */
    k=0;
    nc = 0;
    M->m = instance->nC + 2; /* number of rows in the RMP (number of customers + convexity + obj function) */
    M->n = 0; /* number of columns in the RMP (number of customers) */
    M->nz= 0; /* number of nonzeros */
    
    /* Set the initial columns */
    for(i=0; i<instance->nC; i++)
    {
        /* Set a new column */
        M->clpnts[nc] = k + 1; /* -------> Fortran index */
        
        /* Set the entries of the column */
        M->coeff[k] = 1.0;
        M->rwnmbs[k] = i + 1; /* -------> Fortran index */
        k++;
        
        /* Set the coefficient in the convexity constraints */
        M->coeff[k] = 1.0;
        M->rwnmbs[k] = instance->nC + 1; /* -------> Fortran index */
        k++;
        
        /* Set the cost corresponding to the generated column */
        /* Cost: distance from the source to the customer in {1, ..., nC} */
        M->coeff[k] = 2 * instance->distance[0][i+1];
        M->rwnmbs[k] = instance->nC + 2; /* -------> Fortran index */
        k++;
        
        /* Set the homogeneous route */
        instance->routes->route[nc] = i;
        instance->routes->route_indices[nc] = i+1;
        
        /* Associate the route to the column */
        columns_links[nc] = i;
        nc++;
    }
        
    /* End of columns! */
    M->clpnts[nc] = k + 1; /* -------> Fortran index */
    instance->routes->route[instance->nC] = instance->nC;
    instance->routes->nr = instance->nC;
    
    /* Set the number of columns and nonzeros */
    M->n = nc;
    M->nz = k;
    
    /* Call the PDCGM function that add the new columns to the current RMP */
    nc = PDCGM_add_columns(PDCGM_env, M, columns_links);

    /* Clean the memory */
    PDCGM_free_SMatrix_CW(M);
    PDCGM_FREE(M);
    PDCGM_FREE(columns_links);
    
    /* If the answer is negative, report the ERROR */
    if(nc < 0) return -1;
    
    /* Return successfully */
    return 0;
}

/* ORACLE FUNCTION */
/* The oracle() procedure must return which of the following elements was generated:
   0 -> nothing was generated, 1 -> one or more columns*/
static short VRPTW_oracle(
    double *primal_violation,   /* total violation of the generated constraints, if any */
    double *dual_violation,     /* total violation of the generated columns , if any */
    PDCGM  *PDCGM_env,          /* PDCGM environment */
    void   *instance_data)      /* instance data */
{    
    /* Convert the instance_data to the local type */
    T_instance_data *instance = (T_instance_data *) instance_data;

    int i, j, k, t, q, ret, nc, *solution, 
        route_len, count, max_m, max_n, max_nz,
        *columns_links=NULL;
        
    double *u, route_cost, route_red_cost;
        
    PDCGM_SMatrix_CW *M; 
    
    PDCGM_set_change_bounds(PDCGM_env, 0);
    PDCGM_set_reliable_answer(PDCGM_env, 1);
    
    /* Set initial values */
    PDCGM_env->ncols_last = 0;
    *dual_violation = 0.0;
    nc = 0;
    ret = 0;
    
    /* Get the current dual solution */
    u = PDCGM_get_dual_solution(PDCGM_env);
    
    /* Allocate the auxiliary matrix in memory */
    M = PDCGM_ALLOC(PDCGM_SMatrix_CW, 1);
    if(M == NULL) return -1;
    max_m = instance->nC + 2; /* number of rows in the RMP:
            (number of customers + number of cuts (SRC) + convexity + obj function) */ 
    max_n = MAX_NUMBER_COLS_IN_A_CALL;
    max_nz = max_n * max_m;
    PDCGM_set_SMatrix_CW(M, max_m, max_n, max_nz);
    M->m = max_m;
    M->n = M->nz = 0;
    
    /* Solution array: (x_{1}, ..., x_{nC}) and columns links */
    solution = instance->buffer_int;
    columns_links = &(instance->buffer_int[instance->nC]);
    
    /* Set the adjacency list (graph) */
    ret = RCESPP_set_adjacency_list(instance->RCESPP_data, u, NULL, 0);
    if(ret < 0) 
    {
        printf("\n-- Memory allocation ERROR at oracle --");
        return -1;
    }
    
    /* Call the subproblem solver until new columns are found (if possible) */  
    do
    {
        /* Reset the vector of routes */
        RCESPP_reset_paths_list(instance->RCESPP_data);
        
        /* Solve the subproblem (RCESPP) */     
        ret = RCESPP_nonelementary_solver(instance->RCESPP_data, PDCGM_env->rel_gap, PDCGM_env->delta, PDCGM_env->outer, PDCGM_env->G->n - PDCGM_env->columns_begin, 0, dual_violation, -VRPTW_TOL_ZERO, &route_cost, u);
        if(ret < 0) 
        {
            printf("\n-- Memory allocation ERROR at oracle --");
            return -1;
        }
        
        /* Check if new columns were generated */
        k=0;
        nc = 0;
        M->n = M->nz = 0;
        if( (ret != -1) && ((instance->RCESPP_data)->routes->queue->n > 0) )
        {
            /* Clean the solution array */
            for(i=0; i < instance->nC; i++) solution[i] = 0;
        
            /* Scan all the routes */
            for(j=0; (j < max_n) && (j < (instance->RCESPP_data)->routes->queue->n); j++)   
            {   
                /* Associate the route to the column */
                columns_links[nc] = instance->routes->nr;
                
                /* Get the column corresponding to the next stored route */
                route_len = RCESPP_get_next_solution(&route_red_cost, solution, &route_cost, instance->routes->route, 
                    &(instance->routes->nr), &(instance->routes->route_indices), &(instance->routes->ni_max), 0, -VRPTW_TOL_ZERO, u, &(u[instance->nC]), instance->RCESPP_data); 
                
                if(route_len < 1) break;
                if(route_red_cost < *dual_violation) *dual_violation = route_red_cost;
                
                /* Check the memory available to store the routes */
                if(instance->routes->nr >= instance->routes->nr_max-1)
                {
                    /* Reallocate RMP routes */
                    instance->routes->nr_max = (int) ( instance->routes->nr_max * 1.5 );
                    instance->routes->route = (int *) realloc(instance->routes->route, sizeof(int) * instance->routes->nr_max );
                    printf("\n* [ realloc from oracle(), %d, %d, %p ]\n", instance->routes->nr, instance->routes->nr_max, instance->routes->route);
                    fflush(stdout);
                    if(instance->routes->route == NULL) return -1;
                }
                
                /* Set in the sparse representation */
                M->clpnts[nc] = k + 1; /* -------> Fortran index */
                
                /* Set the entries that correspond to the routes */
                for(i=0; i < instance->nC; i++)
                {
                    /* Check if it is a nonzero entry */
                    if(solution[i] > 0.5) 
                    {
                        /* Set a new entry in the column */
                        M->coeff[k] = (double) solution[i];
                        M->rwnmbs[k] = i + 1; /* -------> Fortran index */
                        k++;
                        /* Reset the solution vector */             
                        solution[i] = 0;
                    }
                }
        
                /* Set the coefficient of the convexity constraint */
                M->coeff[k] = 1.;
                M->rwnmbs[k] = instance->nC + 1; /* -------> Fortran index */
                k++;

                /* Set the cost corresponding to the generated column */
                M->coeff[k] = route_cost;
                M->rwnmbs[k] = instance->nC + 2; /* -------> Fortran index */
                k++;
                
                /* Increase the number of columns */
                nc++;
                
            }
    
            /* End of columns! */
            M->clpnts[nc] = k + 1; /* -------> Fortran index */

            /* Set the number of columns and nonzeros */
            M->n = nc;
            M->nz = k; 
        
            /* Update the objective value of the oracle */
            *dual_violation = PDCGM_env->rhs[PDCGM_env->dim] * (*dual_violation); /* == N * (c*x) */
        
            /* Add the generated columns to the coefficient matrix  */
            if(M->n > 0)
            {
                /* Call the PDCGM function that add the new columns to the current RMP */
                nc = PDCGM_add_columns(PDCGM_env, M, columns_links);
        
                /* If the answer is negative, report the ERROR */
                if( nc < 0 ) return -1;     
            }
        }       
        
        /* If nc > 0, then new columns were generated. */
        /* Otherwise, check if a new run is possible */
        if(nc == 0) *dual_violation = 0.0;
    }
    while( (nc == 0) && (ret == 0) );
    
    /* Clean the memory */
    PDCGM_free_SMatrix_CW(M);
    PDCGM_FREE(M);
    
    /* Return 1: COLUMN GENERATION */
    return 1;
}


/* MAIN FUNCTION */
int main(int arg_c, char *arg_v[])
{
    FILE *out_file;

    char file_name[1000];
    int i, j, k, p;
    short ret; 

    /* Variable to store the instance data */
    T_instance_data VRPTW_instance;
    
    int dim,                 /* number of kept constraints in the RMP */
        nb_cvxty_constr;    /* number of convexity constraints in the RMP */

    PDCGM *VRPTW;
    PDCGM_SMatrix_CW M;
    
    double *u_initial, *b, *lo_box, *up_box, aux_double;
    short  *rw_type;
    
    int    max_outer = 1000;

    /* Check if the file was specified */
    if (arg_c > 1) 
    {
        /* Copy the parameters */
        strcpy(file_name, arg_v[1]);
        if (arg_c > 2 && atoi(arg_v[2]) > 0) MAX_NUMBER_COLS_IN_A_CALL = atoi(arg_v[2]);
    }
    else 
    {
        printf("\n-- ERROR: File name not specified --\n");
        return 0;
    } 
    
    /* Read the problem */
    if(VRPTW_read_instance(file_name, &VRPTW_instance) < 1)
    {
        printf("\nThe file '%s' could not be read.\n", file_name);
        return 0;
    }
    
    #ifdef SAVE_TO_FILE
        /* Save to an output file */
        out_file = fopen("output-vrptw-pdcgm.txt", "a");
        fprintf(out_file, "\n%s ", file_name);
        fflush(out_file); 
        fclose(out_file);
    #endif
    
    /* Set the dimensions of the instance */
    dim = VRPTW_instance.nC;
    nb_cvxty_constr = 1;
    
    /* Allocate the auxiliary data structures */
    VRPTW_set_auxiliary_data(&VRPTW_instance, max_outer, max_outer * 10);
    
    /* initial guess u, b, rw_type and boxes: */
    b = PDCGM_ALLOC(double, dim + nb_cvxty_constr + 1);
    rw_type = PDCGM_ALLOC(short, dim + nb_cvxty_constr + 1);
    lo_box = PDCGM_ALLOC(double, dim);
    up_box = PDCGM_ALLOC(double, dim);
    u_initial = PDCGM_ALLOC(double, dim + nb_cvxty_constr + 1);
    for (i=0; i<dim; i++) 
    {
        /* Type of the constraint */
        rw_type[i] = GREATER_EQUAL;
        
        /* RHS of the constraint */
        b[i] = 1.0;
        
        /* Lower and upper bound of the associated dual variable */
        lo_box[i] = -1.e+6;
        up_box[i] = 1.e+6;
        
        /* An initial guess for the associated dual variable */
        u_initial[i] = 100.0;
    }
    for (i=dim; i<dim+nb_cvxty_constr; i++) 
    {
        /* Type of the constraint */
        rw_type[i] = LESS_EQUAL;
        
        /* RHS of the constraint */
        b[i] = dim;
        
        /* An initial guess for the associated dual variable */
        u_initial[i] = 0.0;
    }
    u_initial[dim+nb_cvxty_constr] = 0.0;
    b[dim+nb_cvxty_constr] = 0.0;
    rw_type[dim+nb_cvxty_constr] = OBJECTIVE;

    /* POPULATE THE INTERNAL DATA STRUCTURE OF PDCGM */ 
    VRPTW = PDCGM_set_data( 
                dim,  /* number of kept constraints in the RMP */
                nb_cvxty_constr,  /* number of convexity constraints in the RMP */
                (dim + nb_cvxty_constr + 1),  /* maximum number of nonzeros in a column */
                MAX_NUMBER_COLS_IN_A_CALL,    /* maximum number of columns generated by the oracle */
                max_outer,     /* maximum number of outer iterations */
                NULL, //u_initial, /* initial guess of the dual solution (may be NULL) */
                b,    /* RHS of each constraint in the RMP */
                rw_type,  /* type of each constraint (row) in the RMP */
                lo_box,   /* lower bound vector of the DUAL variables in the RMP */
                up_box,   /* upper bound vector of the DUAL variables in the RMP */
                &VRPTW_instance);   /* instance data */
    
    if(VRPTW == NULL)   
    {
        printf("-- ERROR: PDCGM_set_data() returned NULL --\n");
        fflush(stdout);
        return 0;
    }
    
    /*--------------------------------------------------------*/
    /* SOLVE THE PROBLEM USING COLUMN GENERATION              */
    /*--------------------------------------------------------*/
     
    /* Set the optimality tolerance for the column generation algorithm */
    PDCGM_set_delta(VRPTW, VRPTW_DELTA);   
     
    /* Set the degree of optimality (parameter D > 1) */
    PDCGM_set_degree_of_optimality(VRPTW, 10.0);
    
    /* Set the verbose mode (how much information is printed) */
    PDCGM_set_verbosity(VRPTW, 1);
    
    /* Disable cut elimination on insertion */
    PDCGM_set_column_elimination(VRPTW, 1);
    
    /* The columns corresponding to homogeneous routes are never removed */
    VRPTW->start_from_reduce_matrix = VRPTW_instance.nN-2;
    
    /* Set epsilon_max */
    PDCGM_set_max_opt_tol(VRPTW, 0.5);
    
    /* Preloaded columns */
    VRPTW_set_initial_columns(VRPTW, &VRPTW_instance);
    
    /* CALL THE COLUMN GENERATION PROCEDURE */
    VRPTW_instance.RCESPP_data->presolver = 0; 
    ret = PDCGM_solve_MP(VRPTW, VRPTW_oracle);
    
    /* Analyse the answer obtained from PDCGM */
    printf("PDCGM has finished with code = %d\n\n", ret);
    fflush(stdout); 
    
    #ifdef SAVE_TO_FILE 
        /* Save to an output file */
        out_file = fopen("output-vrptw-pdcgm.txt", "a");
        //fprintf(out_file, "\t%d \t%d \t%d \t%d ", VRPTW_instance.nC, VRPTW_instance.q, (VRPTW->G)->n, (VRPTW->G)->m-dim-2);
        fprintf(out_file, "\t%lf \t%lf \t%1.6E \t%d \t%lf \t%lf \t%lf ",  
            VRPTW->lowerBnd * 0.1, VRPTW->upperBnd * 0.1, VRPTW->rel_gap, VRPTW->outer,
            VRPTW->cputime_RMP, VRPTW->cputime_oracle, VRPTW->cputime_CG);
        fflush(out_file);
        fclose(out_file);
    #endif
    
    /* Clean the memory before returning */
    PDCGM_free(VRPTW);
    free(VRPTW);
    free(u_initial);
    free(b);
    free(rw_type);
    free(lo_box);
    free(up_box);
    VRPTW_free_data(&VRPTW_instance);
    
    return 0; 
}

