/*
 *  PDCGM v.2.0 (Primal-Dual Column Generation Method)
 *  Copyright: Jacek Gondzio, 1996, 2010.
 *
 *  Written by:  
 *
 *  -> Version 1.0: Jacek Gondzio & Robert Sarkissian,
 *                  LogiLab, HEC Geneva, University of Geneva,
 *                  Bd. Carl-Vogt 102, 1205 Geneva,
 *                  Switzerland. 
 *
 *  -> Version 2.0: Pedro Munari 
 *			  	    University of Sao Paulo, Brazil
 *		            pedro.munari@gmail.com
 *
 *  Last modified: Oct/2012
 *
 *  On top of HOPDM  (Higher Order Primal-Dual Method)
 *  Copyright: Jacek Gondzio, 1990, 2010.
 */

#ifdef _cplusplus
extern "C" {
#endif

#include "../interface/Pb.h"
#include "priority_queue.h"

/* PDCGM_TOL_ZERO: Tolerance to zero */
#define PDCGM_TOL_ZERO 1.e-10

/* PDCGM_TOL_INFEAS: Tolerance to infeasibility */
#define PDCGM_TOL_INFEAS 1.e-4

/* PDCGM_BIG: Big number (infinity) */ 
#define PDCGM_BIG 1.e+30

/*--------------------------------------------------------------*/
/* DATA STRUCTURE							 	   				*/
/*--------------------------------------------------------------*/

#ifndef PDCGM_MATRIX_H
#define PDCGM_MATRIX_H
typedef struct PDCGM_SMatrix_CW_struct {
     int      m;
     int      n;
     int      nz;
     int      max_m;
     int      max_n;  
     int      max_nz;
     double    *coeff;
     INTS      *rwnmbs;
     long      *clpnts;
     char      *rwname;
     char      *clname;
} PDCGM_SMatrix_CW;

typedef struct PDCGM_SMatrix_RW_struct {
     int      m;
     int      n; 
     int      nz;
     int      max_m;
     int      max_n;
     int      max_nz;
     double    *coeff;
     long      *rwpnts;
     INTS      *clnmbs;
     char      *rwname;
     char      *clname;
} PDCGM_SMatrix_RW;
#endif

typedef struct struct_pdcgm 
{
	/* Information about the problem */
	void *instance_data;  /* pointer to the instance data */
   
	int    n_vars,			/* number of primal variables in the RMP */
		   n_art_vars,	    /* number of artificial variables in the RMP */
		   dim, 		    /* number of (coupling) constraints in the RMP */	
           nb_cvxty_constr, /* number of convexity constraints in the RMP */
           n_cuts;			/* number of valid inequalities in the RMP */

	Pb      *RMP; 		/* RMP data in the internal structure of HOPDM */
	Pb_save *L_save;    /* Pointer to the solution vectors of the RMP */
	                          
	PDCGM_SMatrix_CW *G;    /* coefficient matrix of the RMP */
   
	double *lambda,	    /* primal solution of the RMP */
		   *dual_slack, /* dual slack solution of the RMP */
		   lambda_LB,   /* lower bound of EVERY primal variable (lambda_LB <= lambda[i] <= lambda_UB) */
		   lambda_UB,   /* upper bound of EVERY primal variable (lambda_LB <= lambda[i] <= lambda_UB) */
		   objective_LB,  /* lower bound of the optimal objective value */
		   objective_UB,  /* upper bound of the optimal objective value */
		   *u,   	  /* dual solution of the RMP */
		   *low_bnd,  /* lower bound vector of the PRIMAL variables in the RMP (low_bnd[i] = lambda_LB) */
           *upp_bnd,  /* upper bound vector of the PRIMAL variables in the RMP (upp_bnd[i] = lambda_UB) */
           *rhs,      /* RHS of each constraint in the RMP */
   		   *range;    /* Range of each constraint in the RMP */
   
    INTS    *rwstat;  /* Type of the constraints (rows) in the RMP 
   					    (EQUAL, LESS_EQUAL, GREATER_EQUAL, OBJECTIVE) */
    
    int auxiliary_begin; /* Beginning of the auxiliary columns inserted by the user */
    
    int columns_begin;   /* Beginning of the columns inserted by the oracle in the coefficient
	       			        matrix. They start after the slack columns (given by the two
	        			    identity submatrices) and auxiliary columns inserted by the user. */
	        			      
	int *columns_links; /* In column generation, some type of information is usually associated to a column
						    (e.g., in the VRPTW, a route should be associated to a column it generates).
						    This information should be kept until the end of the algorithm and, hence, a link
						    to it is associated to each column in the coefficient matrix. Since columns may be
						    eliminated in the course of the algorith, a dynamic link should be ketp. */
	int *rows_links;    /* Similar to columns links, but for rows now */
   
	/* Information about the column generation */
	double delta,    	/* maximum gap tolerance in the column generation (required to the user) */
		   opt_tol,		/* optimality tolerance used to solve the current RMP */
		   opt_tol_MAX,
           upperBnd,    /* upper bound in the column generation 
            			   (here, it is provided by the oracle) */
           lowerBnd;    /* lower bound in the column generation 
            			   (here, it is provided by the RMP) */

	double last_upperBnd, last_lowerBnd, last_rel_gap_cut;

	/* Bounds of the DUAL variables in the RMP */
	double *y_lo_bnd, /* lower bounds */
 		   *y_up_bnd, /* upper bounds */
 		   *y_fxd_lo, /* minimum lower bounds */
 		   *y_fxd_up; /* maximum upper bounds */
 				
	/* Degree of optimality (parameter D > 1) */
	double degree_of_optimality; /* It controls the distance to optimality used to obtain a suboptimal solution.
	   								See function 'PDCGM_solve_RMP()' for more information */

	/* IPDCGM arrays: Data structure of the Integer Primal Dual Column Generation */
	double IP_obj_value;/* objective value of integer master solution */
	double *IP_values;  /* values of the nonzero integer master solution */
	int   *IP_indices,  /* indices of the nonzero integer master solution */
		   IP_n, 		/* number of nonzero nonzero integer master variables (number of entries in IP_indices) */
		   IP_n_max,	/* maximum number of nonzero nonzero integer master variables (allocated in memory) */
		  *IP_seeds,	/* indices of variables used to initialized the rounding heuristic in the last calls */
		   IP_n_seeds,   /* number of indices in IP_seeds */
		   IP_n_seeds_max;   /* maximum number of indices allocated in IP_seeds */
	
	/* Limits */
	int MAX_n_columns;
	int MAX_n_cuts;
	int MAX_n_columns_oracle;   /* maximum number of columns generated by the oracle */
	int MAX_n_cuts_oracle;      /*  */
	int MAX_nnz_col;         	/* maximum number of nonzeros in a column */
	int MAX_outer;              /* maximum number of (outer) iterations in the column generation */
	double MAX_cputime;         /* maximum cpu time spent in the column generation (in seconds) */
	
	double ratio_reduce_matrix; 
	double tolerance_reduce_matrix;
	int last_outer_reduce_matrix;
	int start_from_reduce_matrix;
	
	short  WS_strategy,	 /* Flag: Select which WS strategy should be used (0 -> cold start) */
		   WS_last,		 /* Flag: WS Strategy used in the last outer iteration (0 -> cold start) */  
	       bounds_all_expanded,  /* Flag: Set if all the bounds on u were already expanded */
		   change_bounds,
		   answer_type,
		   reliable_answer,
		   centrality_value,
		   use_difficult_iters;
    
	int   ncols_last;	   /* Number of new columns added to the RMP in the last call to the oracle */
	int   extra_call;

    double rel_gap;		   /* relative distance between bounds in the column generation */
	double rel_gap_real;
	double rel_gap_last;   /* relative gap in the previous iteration */
	
	double abs_diff;	   /* absolute distance between the objective values of the current RMP and oracle */
	double abs_diff_last;  /* absolute distance between the objective values of the previous RMP and oracle */
	double abs_diff_ratio; /* ratio: abs_diff / abs_diff_last */
	
	int    outer,		   /* current number of (outer) iterations in the column generation */
           inner_WS,       /* number of iterations to find the WS point (sum over all outer iterations) */
           inner_query;    /* number iterations to find the query point (sum over all outer iterations) */
            
	double cputime_CG,     /* total cpu time of the column generation procedure */
   		   cputime_oracle, /* total cpu time spent on the oracle */
   		   cputime_WS,     /* total cpu time spent on finding a WS */
   		   cputime_RMP;    /* total cpu time to solve all the RMPs */
   
   /* verbosity: Flag that says what should be printted on screen */
   short verbosity;    /* 0 -> do not print anything; (default)
						  1 -> basic information;
						  2 -> basic information and columns added to the RMP; 
						  3 -> basic information, columns added to the RMP and the RMP */   
	
	int RMP_modified;  /* flag: 1 -> RMP was modified in the last call to the oracle; 0 -> o.w. */
	
	short column_elimination; /* column_elimination WHEN INSERTING a column in the RMP: 
	                             flag: 0 -> off; 1 -> on. */
	                             
	INTS   *go_to_zero;       /* array of flags: 1 -> the corresponding primal variable is vanishing */
    						  
} PDCGM;

typedef short (*Oracle_function) (double *, double *, PDCGM*, void*);

/*--------------------------------------------------------------*/
/* MACROS 									 	   				*/
/*--------------------------------------------------------------*/

/* auxiliary functions */
#define PDCGM_ALLOC(T,n)  (T*) calloc( (n), sizeof(T))
#define PDCGM_REALLOC(V,T,n)  (T*) realloc(V, (  (n)) * ( sizeof(T)))
#define PDCGM_FREE(v)     if(v != NULL) free(v)

#define PDCGM_MAX(A,B)   ((A)>(B)? (A):(B))
#define PDCGM_MIN(A,B)   ((A)<(B)? (A):(B))

/* Returning messages */
#define PDCGM_COMPLETE        0 /* The problem was solved successfully */
#define PDCGM_MEMORY_ERROR    1 /* Error when ALLOCATING MEMORY */
#define PDCGM_ORACLE_ERROR    2 /* Error when calling the oracle function */
#define PDCGM_SOLVERMP_ERROR  3 /* Error when calling the solve_RMP function */
#define PDCGM_NEGATIVE_GAP    4 /* PDCGM has detected an upper bound smaller than the lower bound */
#define PDCGM_INFEASIBLE	  5 /* PDCGM has detected the problem is infeasible */
#define PDCGM_INCOMPLETE	  6 /* PDCGM has stopped before solving the problem successfully */
#define PDCGM_BOUND_CONFLICT  7 /* PDCGM has stopped prematurely because the last lower (upper) bound is 
								   greater (less) than the upper (lower) bound provided as a parameter */
#define PDCGM_INTEGER_EARLY   8 /* An integer solution was found by IPDCGM rounding heuristic and it is optimal */
 
/* Macros used to measure time */
#define PDCGM_TIME_Statement(TOKEN) struct timeval startTime##_##TOKEN, endTime##_##TOKEN; long time_mtime##_##TOKEN, time_seconds##_##TOKEN, time_useconds##_##TOKEN
#define PDCGM_TIME_Start(TOKEN) gettimeofday(&startTime##_##TOKEN, NULL)
#define PDCGM_TIME_Stop(TOKEN) gettimeofday(&endTime##_##TOKEN, NULL); time_seconds##_##TOKEN = endTime##_##TOKEN.tv_sec - startTime##_##TOKEN.tv_sec; time_useconds##_##TOKEN = endTime##_##TOKEN.tv_usec - startTime##_##TOKEN.tv_usec
#define PDCGM_TIME_Output(TOKEN) 0.001 * round( ((time_seconds##_##TOKEN) * 1000 + time_useconds##_##TOKEN/1000.0) )

/*--------------------------------------------------------------*/
/* HEADERS OF PUBLIC FUNCTIONS 			   						*/
/*--------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_data													*/
/* Populate the PDCGM data structure and return a pointer to the data 		*/
/*--------------------------------------------------------------------------*/
PDCGM *PDCGM_set_data (
    int dim,  	            /* number of kept constraints in the RMP */
    int nb_cvxty_constr,   /* number of convexity constraints in the RMP */
    int max_nnz_col,        /* max_nnz_col: maximum number of nonzeros in a column */
	int max_ncols_oracle,   /* max_ncols_oracle: maximum number of columns generated by the oracle */
    int max_outer,		    /* maximum number of outer iterations */
    double *u_initial,		/* initial guess of the dual solution (may be a pointer to NULL)  */
    double *b,       		/* RHS of each constraint in the RMP */
    short  *rw_status,      /* type of each constraint (row) in the RMP */
    double *lo_box,  		/* lower bound vector of the dual variables in the RMP */
    double *up_box,  		/* upper bound vector of the dual variables in the RMP */
    void   *instance_data); /* instance data */


/*--------------------------------------------------------------------------*/
/* Function: PDCGM_solve_MP													*/
/* Core function of PDCGM: Column generation procedure 						*/
/*--------------------------------------------------------------------------*/
short PDCGM_solve_MP (
	PDCGM *P,  /* Internal data of PDCGM solver */
	Oracle_function ORACLE);  /* pointer to the oracle function */
	
/*---------------------------------------------------------------------------------------*/
/* Function: IPDCGM_solve_MP														     */
/* Setup variables and call the core function of PDCGM 									 */
/* If flag 'rounding' is on, it tries to find an integer solution of the master problem  */
/* using a residual rounding strategy 													 */
/* Return: 
	PDCGM_COMPLETE (0) -> The problem was solved successfully 
	PDCGM_MEMORY_ERROR (1) -> Error when ALLOCATING MEMORY 
	PDCGM_ORACLE_ERROR (2) -> Error when calling the oracle function 
	PDCGM_SOLVERMP_ERROR (3) -> Error when calling the solve_RMP function 
	PDCGM_NEGATIVE_GAP (4) -> Upper bound GREATER THAN lower bound 
	PDCGM_INFEASIBLE (5) -> PDCGM has detected the problem is infeasible    
	PDCGM_INCOMPLETE (6) -> PDCGM has stopped before finishing solving the problem 
	PDCGM_BOUND_CONFLICT (7) -> PDCGM has stopped prematurely because the last lower 
	(upper) bound is greater (less) than the upper (lower) bound provided as a parameter */
/*---------------------------------------------------------------------------------------*/
short IPDCGM_solve_MP (
	PDCGM* P,  				/* ( ->) pointer to a PDCGM environment */
	Oracle_function ORACLE, /* ( ->) pointer to the oracle function */
	int rounding); 		    /* flag: 1, try to find an integer solution; 0, o.w. */

/*--------------------------------------------------------------------------*/
/* Function: IPDCGM_residual_rounding										*/
/* Residual rounding procedure for a BINARY master problem. It selects a	*/
/* subset of variables to be ROUNDED UP and then added to the solution.     */
/* The RHS if modified and the feasibility of the constraints is checked. 	*/
/* REMARK: It assumes that the master problem was already solved up to a    */
/* given tolerance (not necessarily small) */
/*--------------------------------------------------------------------------*/
int IPDCGM_residual_rounding(
	PDCGM* P,  				/* ( ->) pointer to a PDCGM environment */
	Oracle_function ORACLE, /* ( ->) pointer to the oracle function */
	double threshold_value,	/* ( ->) Threshold value for selecting variables (all variables with 
	                                 values larger than it must be selected) */
	int T_MAX);				/* ( ->) Maximum number of variables with value below the threshold that can be selected */

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_add_columns												*/
/*--------------------------------------------------------------------------*/
int PDCGM_add_columns (
	PDCGM *P,  /* Internal data of PDCGM solver */
    PDCGM_SMatrix_CW* G_new,  /* matrix with columns to be added to G */
	int *columns_links);   /* external links to be associated to the new columns */

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_add_dense_columns									    */
/* Add to the current coefficient matrix the dense columns received as      */
/* parameter. One or more columns can be added in a call and they must have */
/* the same number of coefficients (zeroes+nonzeroes).                      */
/*--------------------------------------------------------------------------*/
int PDCGM_add_dense_columns (
	PDCGM  *P,  /* Internal data of PDCGM solver */
    double *columns,  /* matrix with columns to be added to G */
	int    *columns_links,  /* external links to be associated to the new columns */
	int    n_coeff_column,  /* number of coefficients in a column */
	int    n_columns);   /* total number of columns to be added */
	
/*--------------------------------------------------------------------------*/
/* Function: PDCGM_add_rows													*/
/* ATTENTION: The columns of R_new must be indexed without consider the     */
/* articial columns															*/
/*--------------------------------------------------------------------------*/
int PDCGM_add_rows(
	PDCGM *P,		/* Internal data of PDCGM solver */
    PDCGM_SMatrix_RW *R_new,  /* RW matrix with columns to be inserted in G */
    double *RHS,     /* RHS of the bew rows */
	INTS   *rwstat,  /* type of the new rows */
	double *lo_box,  /* lower bound on the dual variables associated to the new rows */
	double *up_box,  /* upper bound on the dual variables associated to the new rows */
	int *rows_links);  /* external links to be associated to the new rows */
	
/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_dual_box												*/
/* Set artificial lower and upper bounds for the dual variables				*/
/*--------------------------------------------------------------------------*/
double PDCGM_set_dual_box (
	PDCGM *P,	  /* Internal data of PDCGM solver */
    double u_LB,  /* lower bound: u_LB <= u_i */
    double u_UB); /* upper bound: u_i <= u_UB */

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_remove_columns										    */
/*--------------------------------------------------------------------------*/
int PDCGM_remove_columns(
	PDCGM *P,		/* Internal data of PDCGM solver */
    int  *column_indices,  /* Indices of columns in P->G that must be KEPT in the matrix */
    int  n_column_indices);  /* Number of columns in P->G that must be KEPT in the matrix */

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_verbosity											*/
/* Set a flag that is used to print information about the solution process  */
/*--------------------------------------------------------------------------*/
void PDCGM_set_verbosity (PDCGM *P, short value);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_column_elimination 									*/
/* Reset a flag to apply or not elimination when adding columns to RMP		*/
/*--------------------------------------------------------------------------*/
void PDCGM_set_column_elimination (PDCGM *P, int flag);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_max_number_columns 									*/
/* Reset the maximum number of columns allowed to be in the RMP             */
/*--------------------------------------------------------------------------*/
void PDCGM_set_max_number_columns (PDCGM *P, int max_n);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_degree_of_optimality									*/
/* Reset the degree of optimality (parameter D > 1) 						*/
/*--------------------------------------------------------------------------*/
void PDCGM_set_degree_of_optimality (PDCGM *P, double D);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_max_opt_tol      									*/
/* Reset the maximum value of the optimality tolerance (epsilon_max)		*/
/*--------------------------------------------------------------------------*/
void PDCGM_set_max_opt_tol (PDCGM *P, double max);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_change_bounds       	    							*/
/* Set the way the lower and upper bounds are updated:              		*/
/*   0 -> update bounds in the usual way (if they are better),              */
/*   1 -> update bounds in the usual way but avoid big changes,             */
/*   2 -> replace the current bounds by the new ones.                       */
/*--------------------------------------------------------------------------*/
void PDCGM_set_change_bounds (PDCGM *P, int value);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_reliable_answer       								*/
/* Set if the answer given by the oracle is reliable or not:           		*/
/*   0 -> heuristic/non-reliable solution used to generate columns/rows,    */
/*   1 -> an optimal solution has been used to generate columns/rows.       */
/*--------------------------------------------------------------------------*/
void PDCGM_set_reliable_answer (PDCGM *P, int value);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_centrality_value 									*/
/* Reset the level of centrality required from HOPDM                        */
/*--------------------------------------------------------------------------*/
void PDCGM_set_centrality_value (PDCGM *P, short value);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_max_cputime												*/
/* Reset the maximum CPU time allowed in the CG  [default: 1.0e+6] 		    */
/*--------------------------------------------------------------------------*/
void PDCGM_set_max_cputime (PDCGM * P, double value);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_delta												*/
/* Reset the optimality tolerance of the CG scheme [default: 1.0e-6] 		*/
/*--------------------------------------------------------------------------*/
void PDCGM_set_delta (PDCGM * P, double value);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_get_master_solution										*/
/* Get the vector containing the master (primal) solution					*/
/*--------------------------------------------------------------------------*/
void PDCGM_get_pointer_to_master_solution (
	PDCGM  *P,		 /* ( ->) Internal data of PDCGM solver */
	double **lambda, /* (<- ) pointer to the vector of primal variables in the (restricted) master problem */
	int    *n);      /* (<- ) returns the number of primal variables in the (restricted) master problem */

void PDCGM_get_pointer_to_full_master_solution (
	PDCGM  *P,		 /* ( ->) Internal data of PDCGM solver */
	double **lambda, /* (<- ) pointer to the vector of primal variables in the (restricted) master problem */
	int    *n);      /* (<- ) returns the number of primal variables in the (restricted) master problem */

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_get_dual_solution	    						        */
/* Get the pointer to the array having the dual solution of the RMP		    */
/*--------------------------------------------------------------------------*/
double* PDCGM_get_dual_solution (PDCGM  *P);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_get_master_columns										*/
/* Get a pointer to the sparse representation of the coefficient matrix of  */
/* the current restricted master problem	                                */
/*--------------------------------------------------------------------------*/
void PDCGM_get_pointer_to_master_columns (
	PDCGM  *P,  /* ( ->) Internal data of PDCGM solver */
	PDCGM_SMatrix_CW *M);	/* (<- ) matrix of generated column(s) */

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_get_master_columns										*/
/* Get a pointer to the dense representation of the slack dual solutions of */
/* the current restricted master problem	                                */
/*--------------------------------------------------------------------------*/
void PDCGM_get_pointer_to_dual_slacks (
	PDCGM  *P,		/* ( ->) Internal data of PDCGM solver */
	double **s,     /* (<- ) pointer to the vector of primal variables in the (restricted) master problem */
	int    *n);      /* (<- ) returns the number of primal variables in the (restricted) master problem */

/*--------------------------------------------------------------------------*/
/* Function: print_gap_pdcgm 												*/
/* Print the gap obtained at the end of the column generation               */
/*--------------------------------------------------------------------------*/
void PDCGM_print_gap_pdcgm (PDCGM * P);

/*--------------------------------------------------------------------------*/
/* Function: print_upbnd_pdcgm 												*/
/* Print the best upper bound obtained at the end of the column generation  */
/*--------------------------------------------------------------------------*/
void PDCGM_print_upbnd_pdcgm (PDCGM * P);

/*--------------------------------------------------------------------------*/
/* Function: print_iterations 											    */
/* Print the number of iterations in the column generation procedure, the   */
/* sum of the number of iterations to solve all the RMPs and the sum of the */
/* number of iterations to reach the analytic center when required.         */
/*--------------------------------------------------------------------------*/
void PDCGM_print_iterations (PDCGM * P);

/*--------------------------------------------------------------------------*/
/* Function: print_cputime 												    */
/* Print the CPU time to solve the problem 									*/
/*--------------------------------------------------------------------------*/
void PDCGM_print_cputime (PDCGM * P);

/*--------------------------------------------------------------------------*/
/* Function: free_pdcgm 													*/
/* Clean the memory used by the solver									    */
/*--------------------------------------------------------------------------*/
void PDCGM_free (PDCGM * P);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_copy_full_master_columns									*/
/* Allocate memory and make a copy of all the columns in the current 	    */
/* coefficient matrix, including artificial columns                         */
/*--------------------------------------------------------------------------*/
void PDCGM_copy_full_master_columns (
	PDCGM  *P,  /* ( ->) Internal data of PDCGM solver */
	PDCGM_SMatrix_CW *M);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_set_SMatrix_CW											*/
/*--------------------------------------------------------------------------*/
int PDCGM_set_SMatrix_CW(
	PDCGM_SMatrix_CW *M, 
	int max_m, 
	int max_n, 
	int max_nz);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_reset_SMatrix_CW											*/
/* Reallocate the memory of the arrays 										*/
/*--------------------------------------------------------------------------*/
int PDCGM_reset_SMatrix_CW(
	PDCGM_SMatrix_CW *M, 
	int max_m, 
	int max_n, 
	int max_nz);
	
/*--------------------------------------------------------------------------*/
/* Function: PDCGM_reset_SMatrix_RW											*/
/* Reallocate the memory of the arrays 										*/
/*--------------------------------------------------------------------------*/
int PDCGM_reset_SMatrix_RW(
	PDCGM_SMatrix_RW *M, 
	int max_m, 
	int max_n, 
	int max_nz);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_free_SMatrix_CW											*/
/*--------------------------------------------------------------------------*/
void PDCGM_free_SMatrix_CW(PDCGM_SMatrix_CW *M);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_free_SMatrix_RW											*/
/*--------------------------------------------------------------------------*/
void PDCGM_free_SMatrix_RW(PDCGM_SMatrix_RW *M);

/*--------------------------------------------------------------------------*/
/* Print a matrix in the format of a (dual) LP 								*/
/*--------------------------------------------------------------------------*/
void PDCGM_print_Matrix(PDCGM_SMatrix_CW *M, int dim);

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_add_auxiliary_columns_dense							    */
/*--------------------------------------------------------------------------*/
int PDCGM_add_auxiliary_columns_dense (
	PDCGM  *P,  /* Internal data of PDCGM solver */
    double *columns,  /* matrix with columns to be added to G */
	int    n_coeff_column,  /* number of coefficients in a column */
	int    n_columns);   /* total number of columns to be added */

/*--------------------------------------------------------------------------*/
/* Function: PDCGM_add_auxiliary_columns									*/
/*--------------------------------------------------------------------------*/
int PDCGM_add_auxiliary_columns (
	PDCGM *P,  /* Internal data of PDCGM solver */
    PDCGM_SMatrix_CW* G_new);  /* matrix with columns to be added to G */
    	
#ifdef _cplusplus
}
#endif

