/* ---------------------------------------------------------------------------
 * File: priority_queue.c (Priority queue)
 * ---------------------------------------------------------------------------
 * Author: Pedro Munari (pedro.munari@gmail.com)
 * Last update: 08/2009
 * ---------------------------------------------------------------------------
 */

/* Data structure:

		       0 1                     (n)                    (n_max)
		      ----------------------------------------------------
	 values   | |**********************|                         |   
		      ----------------------------------------------------
	 indices  | |**********************|                         |
		      ----------------------------------------------------

   HEAP CONDITION: The key in each node should be larger than (or equal to)
                   the keys in its children.
                   - Parent of node j: j/2 (rounded down)
                   - Children of node j: 2j and 2j + 1.
 
   The zero position of the arrays is empty, in order to make easier 
   the relationship between parent/children indices

*/

#ifndef MYLIB_PriorityQueue_h
#define MYLIB_PriorityQueue_h

#include <stdio.h>
#include <stdlib.h>  
#include <string.h>

/* Data structure for the priority queue */
typedef struct
{
    int n_max;       /* maximum number of entries in the queue */
    int n;           /* number of entries in the queue */

    double *values;     /* array with the priority values */
    int    *indices;    /* array with the indices of the priority values */
    
} PriorityQueue_type;

PriorityQueue_type *PriorityQueue_allocate_pointer(void);

/* Allocate the priority queue */
int PriorityQueue_allocate(
    PriorityQueue_type *queue,   /* Pointer to the queue */
    int n_max);                  /* Maximum number of entries in the queue */

/* Reallocate the priority queue */
int PriorityQueue_reallocate(
    PriorityQueue_type *queue,   /* Pointer to the queue */
    int n_max);                  /* New maximum number of entries in the queue */

/* Insert a new entry in the priority queue */
int PriorityQueue_insert(
    PriorityQueue_type *queue,   /* Pointer to the queue */
    double value,                /* Value of the entry to be inserted */
    int index);                   /* Index of the entry to be inserted */

/* Insert a new entry in the priority queue, but replace an entry 
   if the maximum number of entries was reached */
int PriorityQueue_insert_with_replace(
    PriorityQueue_type *queue,   /* Pointer to the queue */
    double value,                /* Value of the entry to be inserted */
    int index);                  /* Index of the entry to be inserted */

/* Remove the top of the priority queue */
int PriorityQueue_getNext(
    PriorityQueue_type *queue,   /* Pointer to the queue */
    double *value,               /* Value of the entry to be inserted */
    int *index);                 /* Index of the entry to be inserted */

/* Peak at the top of the priority queue */
double PriorityQueue_peakTop(PriorityQueue_type *queue);

/* Free the memory of the priority queue */
int PriorityQueue_free(PriorityQueue_type *queue);

#endif

